/*

nhPhyml :  a program that builds phylogenetic trees from DNA alignments using the non-homogeneous, non-stationary model of evolution conceived by Galtier and Gouy (1998) and the algorithmic structure of phyml (Guindon and Gascuel, 2003).

Copyright (C) Bastien Boussau. jul. 2005 onward

All parts of  the source except where indicated  are distributed under
the GNU public licence.  See http://www.opensource.org for details.

*/

#ifndef UTILITIES_H
#define UTILITIES_H


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <errno.h>

#define For(i,n)                     for(i=0; i<n; i++)
#define Fors(i,n,s)                  for(i=0; i<n; i+=s)
#define SHFT2(a,b,c)                 (a)=(b);(b)=(c);
#define SHFT3(a,b,c,d)               (a)=(b);(b)=(c);(c)=(d);
#define MAX(a,b)                     ((a)>(b)?(a):(b))
#define MIN(a,b)                     ((a)<(b)?(a):(b))
#define SHFT(a,b,c,d)                (a)=(b);(b)=(c);(c)=(d);
#define PointGamma(prob,alpha,beta)  PointChi2(prob,2.0*(alpha))/(2.0*(beta))
#define SIGN(a,b)                    ((b) > 0.0 ? fabs(a) : -fabs(a))
#define MMAX(a,b)                    ((a)>(b)?(a):(b))
#define MMIN(a,b)                    ((a)<(b)?(a):(b))
#define lmot (8*sizeof(int))
#define lmin 1.e-6
#define lmax 9999.
#define FREE_ARG char*
//#define double long double



/***************************************************/
#define VERSION "v0.1"

extern int    BRENT_ITMAX;
extern double BRENT_CGOLD;
extern double BRENT_ZEPS;
extern double MNBRAK_GOLD;
extern double MNBRAK_GLIMIT;
extern double MNBRAK_TINY;
extern double ALPHA_MIN;
extern double ALPHA_MAX;
extern double BL_MIN;
extern double BL_START;
extern double BL_MAX;
extern double MIN_DIFF_LK;
extern double GOLDEN_R;
extern double GOLDEN_C;
extern int    T_MAX_FILE;
extern int    T_MAX_LINE;
extern int    T_MAX_NAME;
extern int    T_MAX_SEQ;
extern int    N_MAX_INSERT;
extern int    N_MAX_OTU;
extern double UNLIKELY;
extern double NJ_SEUIL;
extern int    MAX_TOPO_DIST;
extern double ROUND_MAX;
extern double DIST_MAX;
extern int    LIM_SCALE;
extern double LIM_SCALE_VAL;
extern double AROUND_LK;
extern double PROP_STEP;
extern int    T_MAX_ALPHABET;
extern double MDBL_MAX;
extern double MDBL_MIN;
extern int    POWELL_ITMAX;
extern double LINMIN_TOL;
extern int    MEM_LIM;

/*********************************************************/

typedef struct __Node {
  struct __Node     **v;
  double             *l;
  struct __Edge     **b;

  /* To account for Non-homogeneity */
  /* A=0; C=1, G=2, T=3 */
  double **pi;/* base composition at the given node */ 
  double *theta; //Equilibium base compositions in parent branch: as many as there are different site positions
  double *theta_old; //Former equilibium base composition in parent branch (for use when moving back)
  double** qtheta;
  
  //Codon specific
  double* cpi; //Equilibrium codon composition

  int               num;
  char            *name;
  int               tax; //=0 if not a taxon, =1 otherwise
  int     pull_e,pull_i;
  int          ni,agglo;
  int      check_branch;
  double         *score;
  int         is_attach;
  int           is_free;
  int       *n_ex_nodes;
  int          numNodes; // Number of underlying nodes
  int numNodesNormalized;
  int        MaxSubtree; //Maximal size of the node's subtrees
  struct __Node     *BigSubtree;
  struct __Node     *SmallSubtree;
  int            height;
  int             depth;
  int             rearr; //This boolean tells whether we can rearrange (make NNIs around) this node upper branch 


}node;


/*********************************************************/

typedef struct __Edge {
  struct __Node               *left,*rght;
  int         l_r,r_l,l_v1,l_v2,r_v1,r_v2;//lv1,lv2  : numbers to get to the other neighbors of the left node
  //rv1, rv2 : numbers to get to the other neighbors of the right node
  double                            score;
  int                                 num;
  double                                l; /*Branch length: as many as there are different site positions, with the l_coef from the model*/
  double                           best_l;
  double                            l_old;


  //Three variables used during the optimization of Thetas, 
  //in the limited set case ; best_l is used too, as diff_lk

  double                       best_theta;
  
  int                      check_this_one;
  int                             in_both;
  int                              n_quad;
  int                            n_bipart;
  double                         nj_score;
  double                          diff_lk;

  int                       get_p_lk_left;
  int                       get_p_lk_rght;
  int                        ud_p_lk_left;
  int                        ud_p_lk_rght;
  double                         site_dlk;
  double                        site_d2lk;

  double                     *site_dlk_rr;
  double                    *site_d2lk_rr;
  double        ****p_lk_left,****p_lk_rght;//Conditional superior and inferior likelihoods for one site, one rate, one character
  int         p_lk_left_num,p_lk_rght_num;
  /*  double                 **site_p_lk_left; 
      double                 **site_p_lk_rght; */
  double              ****dp_lk_kappa_rght; //Derivative of conditional inferior likelihood in respect to kappa (Ts/Tv)for one position, one site, one rate, one character
  double             ****d2p_lk_kappa_rght; //Second derivative of conditional inferior likelihood in respect to kappa (Ts/Tv) for one position, one site, one rate, one character
  double ****Pij_rr,****dPij_rr,****d2Pij_rr,****dPij_rr_theta,****dPij_rr_frac,****d2Pij_rr_frac, ****dPij_rr_tot, ****d2Pij_rr_tot,****dPij_rr_kappa, ****d2Pij_rr_kappa  ;

  int                         qmat_indice; //indicates which matrix corresponds to the edge

  double                              *ql;
  int                      up_2_date_left;
  int                      up_2_date_rght;
  int                           best_conf;
  int                         num_st_left;
  int                         num_st_rght;
  int                          scale_left;
  int                          scale_rght;
  double            site_sum_scale_f_left;
  double            site_sum_scale_f_rght;
  double                site_scale_f_left;
  double                site_scale_f_rght;
  double                *sum_scale_f_left;
  double                *sum_scale_f_rght;

  int                         *rght_nodes; //which nodes are on the right of current edge
  int                         *left_nodes; //which nodes are on the left of current edge
  int                             numrght; //number of nodes which are on the right of current edge
  int                             numleft; //number of nodes which are on the left of current edge
  int                               rearr; //Do we allow rearrangements (NNIs) around this edge ? This boolean is initialized from the same boolean in the noeud structure.

}edge;

/*********************************************************/

typedef struct __Elk {
  double                            *lk_q;
  double                        *site_dlk;
  double                       *site_d2lk;
  double                     *site_dlk_rr;
  double                    *site_d2lk_rr;
  double        ***p_lk_left,***p_lk_rght;
  double ***Pij_rr,***dPij_rr,***d2Pij_rr;
  double                         *lk_swap;

}elk;


/*********************************************************/

typedef struct __Arbre {
  struct __Node                         *root;
  struct __Node                       **noeud;
  struct __Edge                     **t_edges;
  struct __Arbre                    *old_tree;
  struct __Arbre                   *best_tree;

  double                              T_t,T_b;
  double                          min_diff_lk;

  int                              both_sides;
  int                                   n_otu;
  double                    d_alpha, d2_alpha;
  int                                 n_alpha;
  int                               curr_site;
  int                               curr_catg;
  int                                   nswap;
  double                           best_loglk;
  double                            *t_best_l;

  double                            tot_loglk;
  double                    *tot_loglk_sorted; /* used to compute tot_loglk by adding sorted terms to minimize CPU errors */
  double            *tot_dloglk_domega_sorted; /* used to compute tot_dloglk_domega by adding sorted terms to minimize CPU errors */
  double          *tot_d2loglk_domega2_sorted; /* used to compute tot_d2loglk_domega2 by adding sorted terms to minimize CPU errors */
  double            *tot_dloglk_dtheta_sorted; /* used to compute tot_dloglk_dtheta by adding sorted terms to minimize CPU errors */
  double          *tot_d2loglk_dtheta2_sorted; /* used to compute tot_d2loglk_dtheta2 by adding sorted terms to minimize CPU errors */
 double                  *tot_d2loglk_dtheta2; /* used to compute tot_d2loglk_dtheta2 by adding sorted terms to minimize CPU errors */
  double            *tot_dloglk_dtot_sorted; /* used to compute tot_dloglk_dtot by adding sorted terms to minimize CPU errors */
  double          *tot_d2loglk_dtot2_sorted; /* used to compute tot_d2loglk_dtot2 by adding sorted terms to minimize CPU errors */
  double            *tot_dloglk_dfrac_sorted; /* used to compute tot_dloglk_dfrac by adding sorted terms to minimize CPU errors */
  double          *tot_d2loglk_dfrac2_sorted; /* used to compute tot_d2loglk_dfrac2 by adding sorted terms to minimize CPU errors */
  double                    tot_dloglk_dkappa; 
  double                  tot_d2loglk_dkappa2; 
  double            *tot_dloglk_dkappa_sorted; /* used to compute tot_dloglk_dkappa by adding sorted terms to minimize CPU errors */
  double          *tot_d2loglk_dkappa2_sorted; /* used to compute tot_d2loglk_dkappa2 by adding sorted terms to minimize CPU errors */
  double                    tot_dloglk_dl_coef; 
  double                  tot_d2loglk_dl_coef2; 
  double            *tot_dloglk_dl_coef_sorted; /* used to compute tot_dloglk_dkappa by adding sorted terms to minimize CPU errors */
  double          *tot_d2loglk_dl_coef2_sorted; /* used to compute tot_d2loglk_dkappa2 by adding sorted terms to minimize CPU errors */
  



  double                          *tot_dloglk; /*likelihood derivative in respect to branch length*/
  double                   *tot_dloglk_dtheta; /*likelihood derivative in respect to branch length*/
  double                    tot_dloglk_domega;
  double                         *tot_d2loglk;
  double                  tot_d2loglk_domega2;
  double                             *site_lk;
  double                        *true_site_lk; //site_lk is in reality log_site_lk
  double                      unconstraint_lk;
  double                      tot_dloglk_dtot;
  double                    tot_d2loglk_dtot2;
  double                     tot_dloglk_dfrac;
  double                   tot_d2loglk_dfrac2;
  double                            prop_stab;
  double                                p_exp;
  double                            frac_site;
  int                                  n_swap;
  int                            n_p_lk_alloc;
  struct __Model                         *mod;
  struct __AllSeq                       *data;
  int                              n_iter_opt;

  struct __P_Lk                         *p_lk;

  int                               n_pattern;
}arbre;


/*********************************************************/

typedef struct __Seq {
  char    *name;
  int       len;
  char    *state; //nucleotide
}seq;

/*********************************************************/


typedef struct __AllSeq {
  seq         **c_seq;             /* crunched sequences      */
  // int          *invar;             /* 1 -> states are identical, 0 states vary */
  int           *wght;             /* # of each site in c_seq */
  int            *pos;             /* position (in codon...) of each site in c_seq*/
  double       *factw;
  int           n_otu;
  int       clean_len;             /* uncrunched sequences lengths without gaps */
  int      crunch_len;             /* crunched sequences lengths */
  double       *b_frq;             /* Observed base frequencies */
  double       *c_frq;             /* Codon frequencies */
  int        init_len;
  int         *ambigu;
  int       **pospatt;
  double  *score_site;
  struct __Model *mod;
  int      *whole_seq;
}allseq;

/*********************************************************/
typedef struct __Model {
  struct __Code       *c_code; /* genetic code structure (see below) */
  struct __Qmat **qmat_struct; /* pointer to a table of instantaneous rate matrice structures (see below) */
  struct __Optimiz *s_opt;


  int              ns; /* number of states : 4 for DNA */
  double          R,Y; /* R = A,G frequencies ; Y = C,T frequencies */
  double       *kappa; /* transition/transversion rates: as many as there are different positions for sites */
  double       *omega; /* root GC content: as many as there are different positions for sites */
  double        alpha; /* Gamma distribution parameter */
  double        unity; /* Values 1.0*/
  double           mr; /*Mean substitution rate */
  // double       pinvar;
  // int           invar;
  double    alpha_old;
  double   *omega_old;
  double   *kappa_old;
  int   omega_changed;
  // double   pinvar_old;
  double      *l_coef;
  
  

  double     *r_proba; /* Probability to observe a given substitution rate according to the discretized gamma distribution */
  double          *rr; /*  */
  int          n_catg; /* Number of categories for gamma distribution  */
  int           n_pos; /* Number of different positions for sites (3 for codons) */
  int         seq_len;
  int        stepsize; /*1 for nucleotides, 3 for codons*/

  double          *pi; /* states frequencies as observed in the data, not to confuse with the base distribution at the root*/
  int    update_eigen; /* update_eigen=1 -> the instantaneous rate matrix must be inverted */

  int        ns_codon;
  int      codon_freq; /* 0 -> 1/61  ; 
			  1 -> Freq(XYZ) = Freq(X on first position)
			       x Freq(Y on second position)
			       x Freq(Z on third position) */
  int          EqFreq; /*0 means each branch has its own optimized equilibrium frequency, 1 means we use a limited set of equilibrium frequencies, among which each branch choses the one it "prefers" (maximizes the likelihood) */
  int       NumEqFreq; /* In case EqFreq==1, gives the number of equilibrium frequencies each branch can choose from */
  double *ThetaValues; /* Values of the equilibrium frequencies */

  }model;

/*********************************************************/
typedef struct __Matrix {
  double    **P,**Q,**dist;
  arbre              *tree;
  int              *on_off;
  int                n_otu;
  char              **name;
  int                    r;
  struct __Node **tip_node;
  int     curr_int,curr_br;
}matrix;

/*********************************************************/
typedef struct __Option {
  struct __Model      *mod;   /* pointer to the model structure */
  double              titv;
  double             alpha;
  double            lambda;
  int            positions;  // 0 : 123, 1 : 1, 2 : 12, 3 : 13, 4 : 2, 5 : 23, 6 : 3
  // double            pinvar;
  // int                invar;
  char            *seqfile;
  int                ncatg;
  int             datatype;
  char          *modelname;
  int                model;
  int          interleaved;
  int              end_opt;
  struct __Optimiz  *s_opt;
  int            inputtree;
  struct __Arbre     *tree;
  char      *inputtreefile;
  FILE             *fp_seq;
  FILE            *fp_tree;
  char           *out_file;
  char          *tree_file;
  char       *tree_file_Eq;
  char            *lk_file;
  char       *seq_out_file; //To output sequences if wanted
  char       *lnf_out_file; //To output site likelihoods
  char       *dist_out_file; //To output site distributions for interior nodes
  char       *comp_out_file; //To output sequence compositions for interior nodes
  int   out_file_open_mode;
  int  tree_file_open_mode;
  int          n_data_sets;
  int              seq_len;
  int     n_data_set_asked;
  int           changetree;
  int    internalsequences;
  struct __Seq      **data;
  struct __AllSeq *alldata;
  int             nt_or_cd; /*Use DNA (0) or Codon (1) models ? */
  int             stepsize;
  int                quick; /*With (0) or without (1) final optimization*/
  int    estimate_variance; /*If 1, runs the program on a range of root GC content, so that one can estimate
			      how abrupt is the peak in the likelihood surface*/
  double    lower_GC_bound; /*Lower root GC to try*/
  double    upper_GC_bound; /*Upper root GC to try*/
  double         precision; /*integer giving when the optimization should stop : 10^(-precision)*/
 
}option;

/*********************************************************/

typedef struct __Optimiz {
  int           print;
  int first_opt_alpha;
  int first_opt_kappa;
  int       opt_alpha; /*Optimize gamma distribution shape parameter*/
  int       opt_kappa; /*Optimize ts/tv parameter*/
  int      opt_lambda; /*Optimize Churchill approximate Newton-Raphson  parameter*/ 
  int       opt_omega; /*Optimize root GC parameter*/
  int       opt_theta; /*Optimize equilibrium G+C parameters*/
  int       opt_RootPos; /*Optimize root position parameter*/
  //  int      opt_pinvar;
  int          opt_bl; /*Optimize branch length parameters*/
  double      init_lk;
  int        n_it_max;
}optimiz;

/*********************************************************/

typedef struct __Code {
  char               *aa; /* table of amino acid symbols */
  char             *name; /* name of that code */ 
  
  int           **genetc; /* map that relates codons to amino acids characters */
  int *n_diff_b_2_codons; /* number of differences between two codons */
  int         *tstvtable; /* tstvtable[i*61+j]=2 if at least one transversional 
                               change occured between codons i & j, 
                               =1 if transition, =0 otherwise */
  int      num_curr_code; /* genetic code number */
  int           *sense_c; /* array of sense codons */
  int          n_sense_c; /* number of sense codons */
  int      *from_64_2_61; /* this table is used to get rid of non-sense codons */
  double   **gg_98_codon; /* table used for the Galtier and Gouy 1998 version of codon models */
}code;

/*********************************************************/
typedef struct __Qmat{
  double           *u_mat; /* right eigen vectors */
  double           *v_mat; /* left eigen vectors = inv(u_mat) */
  double        *root_vct; /* eigen values */  
  double     *expD_mr_vct; /* exp(root_vect) */
  double            *qmat; /* instantaneous rate matrix */
  
  double      *qmat_proba; /* probabilities matrix */
  
  
  
  double             dnds; /* dn/ds ratio values */
  double           t_dnds; /* dn/ds ratio values used to fill the instantaneous rate matrices */
  
  
}qmat;

/*********************************************************/


void      Init_Constant();
FILE     *Openfile(char *filename, int mode);
FILE     *OpenTree(option *opt, int mode);
FILE     *OpenSeqs(option *opt, int mode);
void     *mCalloc(int nb, size_t size);
void      Getstring_Stdin(char *file_name);
void      Dynamic_P_Lk_Free(edge *b, arbre *tree);
void      Exit(char *message);
optimiz  *Alloc_Optimiz();
void      Init_Optimiz(optimiz *s_opt);
void      Uppercase(char *ch);
int       Filexists(char *filename);
seq     **Get_Seq(FILE *fp_seq, int interleaved, int *n_otu, int rw, int datatype);
seq     **Read_Seq_Sequential(FILE *in, int *n_otu);
seq     **Read_Seq_Interleaved(FILE *in, int *n_otu);
int       Read_One_Line_Seq(seq ***data, int num_otu, FILE *in);
allseq   *Compact_Seq(seq **data, int n_otu, option *input);
void      Get_Base_Freqs(allseq *data, int position, int step);
allseq   *Compact_CSeq(allseq *data, model *mod);
arbre    *Read_Tree_File(FILE *fp_tree,allseq *data);
char     *Get_Tree_Char(FILE *fp_tree);
void      Check_Ambiguities(allseq *data, int datatype, int stepsize);
allseq   *Make_Seq(int n_otu, int len, char **sp_names);
int       Is_Ambigu(char *state, int datatype, int stepsize);
void      Copy_One_State(char *from, char *to, int state_size);
int       Assign_State(char *c, int stepsize);
void     *mRealloc(void *p,int nb, size_t size);
arbre    *Read_Tree(char *s_tree,allseq *data);
void      Make_Node_Light(node *n,allseq *data);
void      Init_Node_Light(node *n,allseq *data);
void      Init_Tree(arbre *tree);
char    **Sub_Trees_Root(char *tree);
char    **Sub_Trees(char *tree);
void      R_rtree(char *s_tree, node *pere, arbre *tree, int *n_int, int *n_ext,allseq *data);
void      Make_All_Edges_Light(node *a, node *d, int n_otu);
int       Next_Par(char *s, int pos);
void      Test_Tree(char *s);
void      Init_Tree_Edges(node *a, node *d, arbre *tree, int *cur);
void      Make_Edge_Light(node *a, node *d, int n_otu);
void      Init_Edge_Light(edge *b);
void      Computes_Substitution_Probabilities (edge *b, arbre *tree, int catg, int pos);
void      Computes_Derivative_Substitution_Probabilities (edge *b, arbre *tree, int catg, int pos);
void      Computes_Second_Derivative_Substitution_Probabilities (edge *b, arbre *tree, int catg, int pos);
void      Make_Edge_Dirs(edge *b, node *a, node *d, int n_otu);
void      Order_Tree_CSeq(arbre *tree, allseq *data);
void      Make_Edge_Lk(node *a, node *d, arbre *tree);
void      Make_Node_Lk(node *n);
void      Alloc_All_P_Lk(arbre *tree);
int       DiscreteGamma (double freqK[], double rK[], 
			 double alfa, double beta, int K, int median);
double    LnGamma (double alpha);
double    IncompleteGamma (double x, double alpha, double ln_gamma_alpha);
double    PointChi2 (double prob, double v);
double    PointNormal (double prob);
void      qksort(double* A, int ilo, int ihi);
void      Swap(node *a, node *b, node *c, node *d, arbre *tree);
void      NNI(arbre *tree, edge *b_fcus, int do_swap);
double    Update_Lk_At_Given_Edge(edge *b_fcus, arbre *tree);
int       Sort_Edges_Diff_Lk(arbre *tree, edge **sorted_edges, int n_elem);
void      Update_BrLen_Invar(arbre *tree);
char      *Write_Tree(arbre *tree);
void      R_wtree(node *pere, node *fils, char *s_tree,arbre *tree);
void      Print_Fp_Out(FILE *fp_out, time_t t_beg, time_t t_end, arbre *tree, option *input, int n_data_set);
void      Print_Fp_Out_Lines(FILE *fp_out, time_t t_beg, time_t t_end, arbre *tree, option *input, int n_data_set);
void      Update_Pis(arbre *tree);
void      Update_Pis_Subtree (edge *b, arbre *tree);
void      Compute_Pis (edge *b, arbre *tree);
char      *Write_Tree_Equilibrium(arbre *tree);
void      R_wtree_Equilibrium(node *pere, node *fils, char *s_tree, arbre *tree);
void      Print_Tree_Text (arbre *tree, FILE *fp_seqs_out);
void      Count_Nodes (edge *b, arbre *tree);
void      Fill_Table (node *n, int h, int d, char *map, int sizeh, int sized);
char**      Estimate_Anc_Seqs(arbre *tree, char **seqs);
void      Estimate_Seqs (arbre *tree,double **parp, char *align);
void      Estimate_Seqs_Subtree (edge *b, arbre *tree,double **parp, char *align);
double**  Compute_Seqs (edge *b, arbre *tree, double **parp, char *align, double **p);
void      Output_datas  (arbre *tree);
void      Set_values (arbre *tree);
double    true_length(node *nd, double *kappa, arbre *tree);
void      Adapt_Length(arbre *tree);
double    gg95(seq* seq1, seq* seq2, allseq *data, double a);
int       freq_obs(seq* seq1, seq* seq2, int *seqw, double* freq, allseq *data);
void      init_bl(arbre* tree, double** dist);
int       unroot_c(char* carbre);
int       ctob(char *carbre, arbre *tree, int *barbre[]);
void      btltostl(int** btree, arbre *tree, double* treelg);
int       congruent(node* nd, node* from, int** btree, arbre *tree, int nbseq, int j);
int       testbit(int *plist, int num);
void      aj(int *liste, int nb);
double*   lslgbr(int** btree, int n, double** dist, double* sce, double* totallength);
int       matvect(double** m1, int nbl1, int nbc1, double* v2, int nbl2, double* vprod);
int       matmat(double** m1, int nbl1, int nbc1, double** m2, int nbl2, int nbc2, double** prod);
double*   vectmat(double* v1, int nbc1, double** m2, int nbl2, int nbc2);
int       invmat(double** mat, int n, double** invmat);
int       retder(int *liste);
int       samename(char* name1, char* name2);
void      bit1(int *plist, int num);
void      bit0(int *plist, int num);
int       ludcmp(double **a, int n, int *indx, double *d);
void      lubksb(double **a, int n, int *indx, double b[]);
double   *vector(int nl, int nh);
void      free_vector(double *v, int nl, int nh);
void      nrerror(char error_text[]);
double    alsurbet_kim(seq* seq1, seq* seq2, double* freq, allseq* data);
void      Compute_Derivatives_Domega(arbre *tree, int pos);
void      Site_Derivatives_Domega(arbre *tree, int pos);
void      Compute_Derivatives_Dtheta(arbre *tree, edge *b_fcus, int pos);
void      Site_Derivatives_Dtheta(arbre *tree, edge *b_fcus, int pos) ;
void      Computes_Dtheta_Substitution_Probabilities (edge *b, arbre *tree, int catg, int pos);
//void      Compute_Derivatives_Substitution_Probabilities_DtotRootBl(arbre* tree, int rb1, int rb2);
void      Compute_Derivatives_DtotRootBl(arbre* tree);
void      Computes_Derivative_Substitution_Probabilities_Tot (edge *b, arbre *tree, double frac, int catg);
void      Computes_Second_Derivative_Substitution_Probabilities_Tot (edge *b, arbre *tree, double frac, int catg);
void      Site_Derivatives_Dtot(arbre* tree);
void      Compute_Derivatives_DfracRootBl(arbre* tree);
void      Site_Derivatives_Dfrac(arbre* tree);
void      Compute_Derivatives_Substitution_Probabilities_DfracRootBl(arbre* tree, int rb1, int rb2);
void      Computes_Derivative_Substitution_Probabilities_Frac (edge *b, arbre *tree, double tot, int catg);
void      Computes_Second_Derivative_Substitution_Probabilities_Frac (edge *b, arbre *tree, double tot,  int catg);
void      Make_Bit_Table(double** table, int numracine, arbre *tree);
void      Fill_Bit_Table (node *pere, node *fils, int *dejavu, int *pasvu, int **table, int numracine, arbre *tree);
void      Post_Order_Bit (node *pere, node *fils,arbre *tree);
void      Pre_Order_Bit (node *pere, node *fils,arbre *tree);
void      Fill_Nodes_List_Rght(edge *b0,edge *b1, edge *b2);
void      Fill_Nodes_List_Left(edge *b0,edge *b1, edge *b2);
void      NJ_Tree_Fit (node *pere, node *fils, arbre *tree, double **dist, int *n_Non_Agglo);
void      Computes_Derivative_Substitution_Probabilities_Kappa (edge *b, arbre *tree, int catg, int pos);
void      Computes_Second_Derivative_Substitution_Probabilities_Kappa (edge *b, arbre *tree, int catg, int pos);
void      Compute_Derivatives_Dkappa(arbre *tree, int pos);
void      Post_Order_Dkappa(node *pere, node *fils, arbre *tree, int pos);
void      Site_Derivatives_Dkappa(arbre *tree, int pos);
int       Compare_Two_States(char *state1, char *state2, int state_size);
void      Get_Codon_Freqs(allseq *data, model *mod);
model    *Make_Model_Basic();
code     *Make_Code(int num_code);
void      Init_Codons(code *c_code);
void      Init_Nb_Diff_Between_Two_Codons(code *c_code, int ns);
void      Init_Tstv_Table(code *c_code, int ns);
void      Init_GG98_Codon_Table(model *mod);
void      Make_All_Qmat_Struct(model *mod, int n_otu);
qmat     *Make_Qmat_Struct(int ns,  model *mod);
void      Compute_Codon_Equilibrium_Frequencies (double* cpi, double gc,model *mod);
int       Get_AA(code *c_code, int num_codon);
void      Fill_lnf_File (arbre *tree, FILE *fp_lnf_out);
int       Make_Thetas (model *mod) ;
double    Try_Thetas (arbre *tree, edge *b, int ChangeValues);
void      Select_Edges_Theta_To_Change(arbre *tree, edge **sorted_b, int *n_neg);
void      Make_Theta_Changes(arbre *tree,edge **b,int beg,int end);
void      UnMake_Theta_Changes(arbre *tree,edge **b,int beg,int end);
void      Fill_lnf_File_iter (arbre *tree, FILE *fp_lnf_out, int num);
void      Reconstruct_Ancestral_Distributions(arbre *tree, FILE *fout, FILE *fout2);
#endif


