/*

nhPhyml :  a program that builds phylogenetic trees from DNA alignments using the non-homogeneous, non-stationary model of evolution conceived by Galtier and Gouy (1998) and the algorithmic structure of phyml (Guindon and Gascuel, 2003).

Copyright (C) Bastien Boussau. jul. 2005 onward

All parts of  the source except where indicated  are distributed under
the GNU public licence.  See http://www.opensource.org for details.

*/

#include <assert.h>
#include "utilities.h"
#include "free.h"
#include "optimiz.h"
#include "ml.h"

double ROUND_MAX;
double PROP_STEP;


int    BRENT_ITMAX;
double BRENT_CGOLD;
double BRENT_ZEPS;
double MNBRAK_GOLD;
double MNBRAK_GLIMIT;
double MNBRAK_TINY;
double ALPHA_MIN;
double ALPHA_MAX;
double BL_MIN;
double BL_START;
double BL_MAX;
double MIN_DIFF_LK;
double GOLDEN_R;
double GOLDEN_C;
int    T_MAX_FILE;
int    T_MAX_LINE;
int    T_MAX_NAME;
int    T_MAX_SEQ;
int    N_MAX_INSERT;
int    N_MAX_OTU;
int    T_MAX_ALPHABET;
double UNLIKELY;
double NJ_SEUIL;
int    MAX_TOPO_DIST;
double DIST_MAX;
int    LIM_SCALE;
double LIM_SCALE_VAL;
double AROUND_LK;
double MDBL_MAX;
double MDBL_MIN;
int    POWELL_ITMAX;
double LINMIN_TOL;
int    MEM_LIM;
int    MAX_NPOS; //easily modifiable
extern int errno;




/*********************************************************/

void Init_Constant()
{

  BRENT_ITMAX =         100;
  BRENT_CGOLD =   0.3819660;
  BRENT_ZEPS  =      1.e-10;
  MNBRAK_GOLD =    1.618034;
  MNBRAK_GLIMIT =     100.0;
  MNBRAK_TINY =      1.e-20;
  ALPHA_MIN =          0.04;
  ALPHA_MAX =           100;
  BL_MIN =           1.e-06;
  BL_START =         1.e-03;
  BL_MAX =           1.e+05;
  MIN_DIFF_LK =      1.e-06;
  GOLDEN_R =     0.61803399;
  GOLDEN_C = (1.0-GOLDEN_R);
  T_MAX_FILE =        10000;
  T_MAX_LINE =       100000;
  T_MAX_NAME =          200;
  T_MAX_SEQ  =      2000000;
  N_MAX_INSERT  =        20;
  N_MAX_OTU  =         2000;
  UNLIKELY =         -1.e10;
  NJ_SEUIL =            0.1;
  ROUND_MAX =           100;
  DIST_MAX =             4.;
  LIM_SCALE =            30;
  AROUND_LK =          50.0;
  PROP_STEP =           1.0;
  T_MAX_ALPHABET =       64;
  MDBL_MIN =  2.225074E-308;
  MDBL_MAX =  1.797693E+308;
  POWELL_ITMAX =        200;
  LINMIN_TOL =      2.0E-04;
  LIM_SCALE_VAL   =  1.E-50;
  MEM_LIM   =        100000;
  MAX_NPOS  =             3;
}

/*********************************************************/
FILE *Openfile(char *filename, int mode)
{
  /* mode = 0 -> read */
  /* mode = 1 -> write */
  /* mode = 2 -> append */

  FILE *fp;
  char *s;
  int open_test=0;

  s = (char *)mCalloc(T_MAX_FILE,sizeof(char));

  strcpy(s,filename);

  fp = NULL;

  switch(mode)
    {
    case 0 :
      {
        while(!(fp = (FILE *)fopen(s,"r")) && ++open_test<10)
          {
            printf("\nCan't open file %s, enter a new name : ",s);
            Getstring_Stdin(s);
            fflush(stdout);
          }
        break;
      }
    case 1 :
      {
        fp = (FILE *)fopen(s,"w");
        break;
      }
    case 2 :
      {
        fp = (FILE *)fopen(s,"a");
        break;
      }

    default : break;

    }

  Free(s);

  return fp;
}


/*********************************************************/
FILE *OpenTree(option *opt, int mode)
{
  /* mode = 0 -> read */
  /* mode = 1 -> write */
  /* mode = 2 -> append */

  FILE *fp;
  char *s;
  int open_test=0;

  s = (char *)mCalloc(T_MAX_FILE,sizeof(char));

  strcpy(s,opt->inputtreefile);

  fp = NULL;

  switch(mode)
    {
    case 0 :
      {
        while(!(fp = (FILE *)fopen(s,"r")) && ++open_test<10)
          {
            printf("\nCan't open file %s, enter a new name : ",s);
            Getstring_Stdin(s);
            fflush(stdout);
          }
        break;
      }
    case 1 :
      {
        fp = (FILE *)fopen(s,"w");
        break;
      }
    case 2 :
      {
        fp = (FILE *)fopen(s,"a");
        break;
      }

    default : break;

    }
  strcpy(opt->inputtreefile,s);
  Free(s);

  return fp;
}

/*********************************************************/

FILE *OpenSeqs(option *opt, int mode)
{
  /* mode = 0 -> read */
  /* mode = 1 -> write */
  /* mode = 2 -> append */

  FILE *fp;
  char *s;
  int open_test=0;

  s = (char *)mCalloc(T_MAX_FILE,sizeof(char));

  strcpy(s,opt->seqfile);

  fp = NULL;

  switch(mode)
    {
    case 0 :
      {
        while(!(fp = (FILE *)fopen(s,"r")) && ++open_test<10)
          {
            printf("\nCan't open file %s, enter a new name : ",s);
            Getstring_Stdin(s);
            fflush(stdout);
          }
        break;
      } //The following cases are probably useless
    case 1 :
      {
        fp = (FILE *)fopen(s,"w");
        break;
      }
    case 2 :
      {
        fp = (FILE *)fopen(s,"a");
        break;
      }

    default : break;

    }
  strcpy(opt->seqfile,s);
  strcpy(opt->out_file,opt->seqfile);
  strcat(opt->out_file,"_nhPhyml.out");

  strcpy(opt->tree_file,opt->seqfile);
  strcat(opt->tree_file,"_nhPhymlGC.tree");

  strcpy(opt->tree_file_Eq,opt->seqfile);
  strcat(opt->tree_file_Eq,"_nhPhymlEq.tree");
  
  strcpy(opt->lk_file,opt->seqfile);
  strcat(opt->lk_file,"_nhPhyml.lk");

  if (opt->internalsequences) {
    strcpy(opt->seq_out_file,opt->seqfile);
    strcat(opt->seq_out_file,"_nhPhyml.seq");
  }
  
  strcpy(opt->lnf_out_file,opt->seqfile);
  strcat(opt->lnf_out_file,"_nhPhyml.lnf");

  strcpy(opt->dist_out_file,opt->seqfile);
  strcat(opt->dist_out_file,"_nhPhyml.distout");

  strcpy(opt->comp_out_file,opt->seqfile);
  strcat(opt->comp_out_file,"_nhPhyml.compout");


  Free(s);

  return fp;
}

/*********************************************************/

void *mCalloc(int nb, size_t size)
{
  void *allocated;

  if((allocated = calloc((size_t)nb,(size_t)size)) != NULL)
    {
      return allocated;
    }
  else
    Exit("\n. Err: low memory\n");

  return NULL;
}

/*********************************************************/

void Getstring_Stdin(char *file_name)
{
  fgets(file_name,T_MAX_LINE,stdin);
  if (strchr(file_name, '\n') != NULL)
    *strchr(file_name, '\n') = '\0';
}

/*********************************************************/

void Exit(char *message)
{
  fprintf(stderr,"%s",message);
  exit(1);
}

/*********************************************************/

optimiz *Alloc_Optimiz()
{
  optimiz *s_opt;
  s_opt = (optimiz *)mCalloc(1,sizeof(optimiz));
  return s_opt;
}

/*********************************************************/

void Init_Optimiz(optimiz *s_opt)
{
  s_opt->print           = 1;
  s_opt->first_opt_kappa = 0;
  s_opt->opt_kappa       = 4;
  s_opt->opt_bl          = 0;
  //  s_opt->opt_pinvar      = 0;
  s_opt->init_lk         = UNLIKELY;
  s_opt->n_it_max        = 1000;
}

/*********************************************************/


void Uppercase(char *ch)
{
  /* convert ch to upper case -- either ASCII or EBCDIC */
   *ch = isupper((int)*ch) ? *ch : toupper((int)*ch);
}

/*********************************************************/

int Filexists(char *filename)
{ 
  FILE *fp;
  fp =fopen(filename,"r");
  if (fp) {
    fclose(fp);
    return 1;
  } else
    return 0;
}

/*********************************************************/


seq **Get_Seq(FILE *fp_seq, int interleaved, int *n_otu, int rw, int datatype)
{
  seq **data;
  int i,j, gap;
  char **buff;
  int n_unkn,n_removed,pos;
  int *remove;


/*   rewind(fp_seq); */

  if(interleaved) data = Read_Seq_Interleaved(fp_seq,n_otu);
  else            data = Read_Seq_Sequential(fp_seq,n_otu);
  
  if(data)
    {
      buff = (char **)mCalloc(*n_otu,sizeof(char *));
      For(i,*n_otu) buff[i] = (char *)mCalloc(data[0]->len,sizeof(char));
      remove = (int *)mCalloc(data[0]->len,sizeof(int));
  
      n_removed = 0;

      For(i,data[0]->len)
	{
	  gap=0;
	  For(j,*n_otu)
	    {
	      if (!gap){
		switch(data[j]->state[i]) {
		case '?': case '-': case 'N':
		  data[j]->state[i] = 'X';
		  gap=1;
		  break;
		case 'A': case 'T': case 'G': case 'C': case 'U':
		  break;
		default:
		  printf("An input sequence contains an unknown symbol (aka different from A,T,C,G,U)\n Interpreting as gap\n");
		  data[j]->state[i] = 'X';
		  gap=1;
		  break;
		  //		assert(0);
		}
	      }
	    }
	  //The following lines would ignore the whole site even if there is only one gap in one sequence
	  /*
	  if (gap) {
	    For(j,*n_otu)
	      {
		data[j]->state[i] = 'X';
	      }
	  }
	  */
	  n_unkn = 0;
	  For(j,*n_otu) if(data[j]->state[i] == 'X') n_unkn++; 
	  
	  if(n_unkn == *n_otu)
	    {
	      remove[i] = 1;
	      n_removed++;
	    }
	  
	  For(j,*n_otu) buff[j][i] = data[j]->state[i];
	}
      
      if(n_removed > 0) 
	{
	  if(!datatype)
	    printf("\n. %d sites are made from at totally undetermined states ('X', '-', '?' or 'N')...\n",n_removed);
	  else
	    printf("\n. %d sites are made from at totally undetermined states ('X', '-', '?')...\n",n_removed);
	}

      pos = 0;
      For(i,data[0]->len)
	{
	      For(j,*n_otu) data[j]->state[pos] = buff[j][i];
	      pos++;
	}

      For(i,*n_otu) data[i]->len = pos;
      For(i,*n_otu) Free(buff[i]);
      Free(buff);
      Free(remove);
    }
  return data;
}

/*********************************************************/
seq **Read_Seq_Sequential(FILE *in, int *n_otu)
{
  int i;
  char *line;
  int len,readok;
  seq **data;
  char c;

  line = (char *)mCalloc(T_MAX_LINE,sizeof(char));

  readok = len = 0;
  do
    {
      if(fscanf(in,"%s",line) == EOF)
	{
	  Free(line); return NULL;
	}
      else
	{
	  if(strcmp(line,"\n") && strcmp(line,"\n") && strcmp(line,"\t"))
	    {
	      *n_otu = atoi(line);
	      if(*n_otu <= 0) Exit("\n. Problem with sequence format\n");
	      data = (seq **)mCalloc(*n_otu,sizeof(seq *));
	      fscanf(in,"%s",line);
	      len = atoi(line);
	      if(len <= 0) Exit("\n. Problem with sequence format\n");
	      else readok = 1;
	    }
	}
    }while(!readok);

  
  while(((c=fgetc(in))!='\n') && (c != ' ') && (c != '\r') && (c != '\t'));

  For(i,*n_otu)
    {
      data[i] = (seq *)mCalloc(1,sizeof(seq));
      data[i]->len = 0;
      data[i]->name = (char *)mCalloc(T_MAX_NAME,sizeof(char));
      data[i]->state = (char *)mCalloc(T_MAX_SEQ,sizeof(char));
      fscanf(in,"%s",data[i]->name);
      
      while(data[i]->len < len)
	Read_One_Line_Seq(&data,i,in);
	
      if(data[i]->len != len) 
	{
	  printf("\n. Err: Problem with species %s's sequence (check the format)\n",data[i]->name);
	  Exit("");
	}
    }

  
  Free(line);
  return data;
}

/*********************************************************/

seq **Read_Seq_Interleaved(FILE *in, int *n_otu)
{
  int i,end,num_block;
  char *line;
  int len,readok;
  seq **data;
  char c;

  line = (char *)mCalloc(T_MAX_LINE,sizeof(char));

  readok = len = 0;
  do
    {
      if(fscanf(in,"%s",line) == EOF)
	{
	  Free(line); return NULL;
	}
      else
	{
	  if(strcmp(line,"\n") && strcmp(line,"\r") && strcmp(line,"\t"))
	    {
	      *n_otu = atoi(line);
	      if(*n_otu <= 0) Exit("\n. Problem with sequence format\n");
	      data = (seq **)mCalloc(*n_otu,sizeof(seq *));
	      fscanf(in,"%s",line);
	      len = atoi(line);
	      if(len <= 0) Exit("\n. Problem with sequence format\n");
	      else readok = 1;
	    }
	}
    }while(!readok);


  while(((c=fgetc(in))!='\n') && (c != ' ') && (c != '\r') && (c != '\t'));

  end = 0;
  For(i,*n_otu)
    {
      data[i] = (seq *)mCalloc(1,sizeof(seq));      
      data[i]->len = 0;
      data[i]->name = (char *)mCalloc(T_MAX_NAME,sizeof(char));
      data[i]->state = (char *)mCalloc(T_MAX_SEQ,sizeof(char));
      fscanf(in,"%s",data[i]->name);
      if(!Read_One_Line_Seq(&data,i,in)) 
	{
	  end = 1;
	  if(i != *n_otu) 
	    {
	      printf("\n. Err: Problem with species %s's sequence\n",data[i]->name);
	      Exit("");
	    }
	  break;
	}
    }

  if(data[0]->len == len) end = 1;

  if(!end)
    {
      end = 0;

      num_block = 1;
      do
	{
	  num_block++;
	  
	  /* interblock */
	  if(!fgets(line,T_MAX_LINE,in)) break;
	  
	  if(strcmp(line,"\n")) 
	    {
	      printf("\n. One or more missing sequences in block %d\n",num_block-1);
	      Exit("");
	    }

	  For(i,*n_otu)
	    if(data[i]->len != len)
	      break;

	  if(i == *n_otu) break;

	  
	  For(i,*n_otu)
	    {
	      if(data[i]->len > len) 
		{
		  printf("\n. Err: Problem with species %s's sequence\n",data[i]->name);
		  Exit("");
		}
	      else if(!Read_One_Line_Seq(&data,i,in)) 
		{
		  end = 1;
		  if(i != *n_otu) 
		    {
		      printf("\n. Err: Problem with species %s's sequence\n",data[i]->name);
		      Exit("");
		    }
		  break;
		}
	    }
	}while(!end);
    }

  For(i,*n_otu)
    if(data[i]->len != len)
      {
	printf("\n. Declared sequence length at the top of the phylip file is not in agreement with sequence %d length...\n",i);
	Exit("");
      }


  Free(line);
  return data;
}


/*********************************************************/

int Read_One_Line_Seq(seq ***data, int num_otu, FILE *in)
{
  char c;

  c=' ';
  while(1)
    {
      if((c == EOF) || (c == '\n')) break;
      else if((c==' ') || (c=='\t') || (c=='\r')) {c=(char)fgetc(in); continue;}
      Uppercase(&c);
      
      /*if(strchr("ACGTUMRWSYKBDHVNXO?-.",c) == NULL)*/
      if (strchr("ABCDEFGHIKLMNOPQRSTUVWXYZ?-.", c) == NULL)
	{
	  printf("\n. Err: bad symbol: \"%c\" at position %d of species %s\n",
		 c,(*data)[num_otu]->len,(*data)[num_otu]->name);
	  Exit("");
	}

      if(c == '.')
	{
	  c = (*data)[0]->state[(*data)[num_otu]->len];
	  if(!num_otu) 
	    Exit("\n. Err: Symbol \".\" should not appear in the first sequence\n");
	}
      (*data)[num_otu]->state[(*data)[num_otu]->len]=c;
      (*data)[num_otu]->len++;
/*       if(c=='U') c='T'; */
      c = (char)fgetc(in);
    }
  if(c == EOF) return 0;
  else return 1;
}
   
/*********************************************************/

allseq *Compact_Seq(seq **data, int n_otu, option *input)
{
  allseq *alldata;
  int i,j,k,/*diff,*/site,num, skip;
  int n_patt,which_patt, true_len;//,n_invar;
  char **sp_names;

  sp_names = (char **)mCalloc(n_otu,sizeof(char *));
  For(i,n_otu) 
    {
      sp_names[i] = (char *)mCalloc(T_MAX_NAME,sizeof(char));
      strcpy(sp_names[i],data[i]->name);
    
    }

  alldata = Make_Seq(n_otu,data[0]->len,sp_names);

  For(i,n_otu) Free(sp_names[i]);
  Free(sp_names);

  n_patt = which_patt = 0;
  /*  diff = -1;*/

  num=0;
  


  if (data[0]->len%input->stepsize) 
    {
      printf("\n. Sequence length is not a multiple of %d\n",input->stepsize);
      Exit("");
    }


  //If we use a model that deals with nucleotide positions in codons, sequences need to be 3*n long
  if (input->model==1) {
    if (data[0]->len%3) {
      printf("\n. Sequence length is not a multiple of %d\n",3);
      Exit("");
    }
  }
  
  true_len=0;
  Fors(site,data[0]->len,input->stepsize) // from the beginning of the alignment to its end, using steps of the right size (1 or 3)
    {  
      //These ifs permit to know if the site is to be used or not based on its position in the codon.
      skip=0;
      if (input->nt_or_cd) {} //Do not enter the loop if we consider codons
      else{
	if (input->positions==1) {
	  if ((site % 3==1)||(site % 3==2)) {skip=1;} 
	}
	if (input->positions==2) {
	  if (site % 3==2) {skip=1;} 
	}
	if (input->positions==3) {
	if (site % 3==1) {skip=1;} 
	}
	if (input->positions==4) {
	  if ((site % 3==0)||(site % 3==2)) {skip=1;} 
	}
	if (input->positions==5) {
	  if (site % 3==0) {skip=1;} 
	}
	if (input->positions==6) {
	  if ((site % 3==0)||(site % 3==1)) {skip=1;} 
	}
      }
      if (skip==1) {}
      else {
	true_len++;
	Fors(k,n_patt,input->stepsize) // test all the  patterns already found 
	  {
	    For(j,n_otu) // test all the  sequences
	      {
		//if the pattern we study is not at the same position as the stored pattern 
		if ((input->model)&& (alldata->pos[k] != site % 3)) break;
		// If two states are different (two bases in two sequences), try the next pattern
		if(!Compare_Two_States(alldata->c_seq[j]->state+k,
				       data[j]->state+site,
				       input->stepsize)) break;
			    /*
		if(strncmp(alldata->c_seq[j]->state+k,
			   data[j]->state+site,
			   input->stepsize))
		  
			   break;*/
	      }
	    //if we haven't found any difference : we found the right pattern, this is the knth one.
	    
	    if(j == n_otu)
	      {
		which_patt = k;
		break;
	      }
	  }

	/* sites are not compressed if we use codons*/
	if (input->nt_or_cd) {
	  k = n_patt; 
	}
	
	//If the right pattern had never been seen before
	if(k == n_patt)
	  {
	    For(j,n_otu) 
	      Copy_One_State(data[j]->state+site,
			     alldata->c_seq[j]->state+n_patt,
			     input->stepsize);

	    //we record the position of the site
	    if (input->model) {
	      alldata->pos[n_patt]=site % 3;
	    }
	    else {
	      alldata->pos[n_patt]=0;
	    }
	    
	    for(j=0;j<n_otu;j++) 
	      {
		if((Is_Ambigu(alldata->c_seq[j]->state+n_patt,input->datatype,input->stepsize) || 
		    (/*!Compare_Two_States*/strncmp(alldata->c_seq[j]->state+n_patt,
						    alldata->c_seq[0]->state+n_patt,
						    input->stepsize))))
		  break;
	      }
	    
	    if(j==n_otu) 
	      {
		For(j,n_otu)
		  {
		    /*		  alldata->invar[n_patt] =Assign_State(alldata->c_seq[j]->state+n_patt,
				  input->stepsize);*/
		    break;
		  }
	      }
	    else {}    
	    //alldata->invar[n_patt] = -1;
	    
	    alldata->pospatt[n_patt] = (int *)mRealloc(alldata->pospatt[n_patt],
						       (int)alldata->wght[n_patt]+1,sizeof(int));
	    alldata->pospatt[n_patt][(int)alldata->wght[n_patt]] = site;
	    alldata->wght[n_patt] += 1.;
	    alldata->whole_seq[site]=n_patt;
	    n_patt+=input->stepsize;
	    
	    
	  }
	else 
	  {
	    alldata->pospatt[which_patt] = (int *)mRealloc(alldata->pospatt[which_patt],
							   (int)alldata->wght[which_patt]+1,sizeof(int));
	    alldata->pospatt[which_patt][(int)alldata->wght[which_patt]] = site;
	    //	    alldata->wght[which_patt] += 1.;
	    alldata->whole_seq[site]=which_patt;

	    For(j,input->stepsize) alldata->wght[which_patt+j] += 1.;
	    
	  }
      }
    }


  alldata->init_len = true_len;//n_patt;//data[0]->len;
  alldata->crunch_len = n_patt;
  printf("Sequence length : %d,  Number of sites used : %d, Number of patterns : %d\n",data[0]->len, true_len, n_patt/input->stepsize);
                               //alldata->init_len, true_len, n_patt);
  For(i,n_otu) alldata->c_seq[i]->len = n_patt;
  /*
  for (i=0;i<n_patt;i++) {
    printf("alldata->pos[i] : %d alldata->wght[i] : %d\n", alldata->pos[i],alldata->wght[i] );
  }
  */
  //  n_invar=0;
  // For(i,alldata->crunch_len) if(alldata->invar[i]>-1) n_invar+=(int)alldata->wght[i];


  if(!input->datatype)
    Get_Base_Freqs(alldata,0,1);
  else
    // Get_AA_Freqs(alldata);
    printf("Wrong number assignment datatype");
/*   fprintf(stderr,"Average nucleotides frequencies : \n"); */
/*   fprintf(stderr,"%f %f %f %f\n", */
/* 	 alldata->b_frq[0], */
/* 	 alldata->b_frq[1], */
/* 	 alldata->b_frq[2], */
/* 	 alldata->b_frq[3]); */
  return alldata;
}




/*********************************************************/

void Get_Base_Freqs(allseq *data, int position, int step)
{
  /* Compute observed base frequencies */

  int otu,site;
  double A,C,G,T;
  double fA,fC,fG,fT;
  double w;

  fA = fC = fG = fT = .25;
  
  // For(k,8) 
/* this loop allows to takes into account ambiguous characters in the computation of base frequencies */
  //   {
      A = C = G = T = .0;
      For(otu,data->n_otu)
	{
	  for(site=position;site < data->crunch_len;site+=step)
	    {
	      w = data->wght[site];
	      if(w)
		{
		  switch(data->c_seq[otu]->state[site]){
		  case 'A' : A+=w;//printf("A+%f\n", w);
		    break;
		  case 'C' : C+=w;//printf("C+%f\n", w);
		    break;
		  case 'G' : G+=w;//printf("G+%f\n", w);
		    break;
		  case 'T' : T+=w;//printf("T+%f\n", w);
		    break;
		  case 'U' : T+=w;
		    break;
		  case 'M' : C+=w*fC/(fC+fA); A+=w*fA/(fA+fC);
		    break;
		  case 'R' : G+=w*fG/(fA+fG); A+=w*fA/(fA+fG);
		    break;
		  case 'W' : T+=w*fT/(fA+fT); A+=w*fA/(fA+fT);
		    break;
		  case 'S' : C+=w*fC/(fC+fG); G+=w*fG/(fC+fG);
		    break;
		  case 'Y' : C+=w*fC/(fC+fT); T+=w*fT/(fT+fC);
		    break;
		  case 'K' : G+=w*fG/(fG+fT); T+=w*fT/(fT+fG);
		    break;
		  case 'B' : C+=w*fC/(fC+fG+fT); G+=w*fG/(fC+fG+fT); T+=w*fT/(fC+fG+fT);
		    break;
		  case 'D' : A+=w*fA/(fA+fG+fT); G+=w*fG/(fA+fG+fT); T+=w*fT/(fA+fG+fT);
		    break;
		  case 'H' : A+=w*fA/(fA+fC+fT); C+=w*fC/(fA+fC+fT); T+=w*fT/(fA+fC+fT);
		    break;
		  case 'V' : A+=w*fA/(fA+fC+fG); C+=w*fC/(fA+fC+fG); G+=w*fG/(fA+fC+fG);
		    break;
		  case 'N' : case 'X' : case '?' : case 'O' : case '-' : 
		    A+=w*fA; C+=w*fC; G+=w*fG; T+=w*fT; break;
		  default : break;
		  }
		}
	    }  
	}
      //     printf("A :%f, C : %f, G : %f, T :%f\n",A,C, G, T);
      fA = A/(A+C+G+T);
      fC = C/(A+C+G+T);
      fG = G/(A+C+G+T);
      fT = T/(A+C+G+T);
      //  }

  data->b_frq[0] = fA;
  data->b_frq[1] = fC;
  data->b_frq[2] = fG;
  data->b_frq[3] = fT;

}

/*********************************************************/

void Get_Codon_Freqs(allseq *data, model *mod)
{
  /* Compute codon frequencies using base frequencies at the three coding 
     positions (F3X4 in CODEML) */

  int i;

  if(!mod->codon_freq) 
    {
      For(i,mod->c_code->n_sense_c) 
	data->c_frq[i] = 1./mod->c_code->n_sense_c;
    }
  else
    {
      int j,k,num_codon;
      double **base_freqs_pos;
      double sum;

      base_freqs_pos = (double **)mCalloc(4,sizeof(double *));
      For(i,4) base_freqs_pos[i] = (double *)mCalloc(3,sizeof(double));

      For(i,3) 
	{
	  Get_Base_Freqs(data,i,3); // Computes base frequencies in 1st, 2nd and eventually 3rd position
	  For(j,4) 
	    {
	      base_freqs_pos[j][i] = data->b_frq[j];
	    }
	}
      sum = .0;
      For(i,4)
	{
	  For(j,4)
	    {
	      For(k,4)
		{
		  num_codon = i*16+j*4+k;
		  if(mod->c_code->from_64_2_61[num_codon] > -1) 
		    {
		      /*  printf("base_freqs_pos[i][0] : %f\n",base_freqs_pos[i][0]);
		      printf("base_freqs_pos[j][1] : %f\n",base_freqs_pos[j][1]);
		      printf("base_freqs_pos[k][2] : %f\n",base_freqs_pos[k][2]);*/
		      data->c_frq[mod->c_code->from_64_2_61[num_codon]] = 
			base_freqs_pos[i][0] *
			base_freqs_pos[j][1] *
			base_freqs_pos[k][2];
		      sum += data->c_frq[mod->c_code->from_64_2_61[num_codon]];
		    }
		}
	    }
	}

      For(i,mod->c_code->n_sense_c) data->c_frq[i] /= sum;

      For(i,4) Free(base_freqs_pos[i]);
      Free(base_freqs_pos);
    }
}

/*********************************************************/


/* Reads the tree and loads it into a char* "line", and eventually builds a arbre structure*/
arbre *Read_Tree_File(FILE *fp_tree,allseq *data )
{
  char *line;
  arbre *tree;
  int i;
  char c;

  line = (char *)mCalloc(T_MAX_LINE,sizeof(char));

  //Discarding useless (e.g. phylo_win header) stuff
  do{
    c=fgetc(fp_tree);
  if(c=='[') {
    do   c=fgetc(fp_tree);
    while((c != ']') && (c != EOF));
  }
  }
  while((c != '(') && (c != EOF));

  if(c==EOF) return NULL;

  i=0;
  for(;;)
    {
      if((c==' ') || (c=='\n'))
	{
	  //c=fgetc(fp_tree); 
	  //if(c==EOF) break;
	  do   c=fgetc(fp_tree);
	  while((c != '}') && (c != '{') && (c != ')') && (c != '(') &&(c != EOF)&&(c != ',')&& (isalpha(c)==0));//To avoid being troubled by bootstrap trees
	  if(c==EOF) break;
	}
      
      line[i]=c;
      i++;
      c=fgetc(fp_tree);
      if(c==EOF) break;
    }
  tree = Read_Tree(line,data);
  Free(line);
  return tree;
}

/*********************************************************/

/* Reads the tree and loads it into a char* "line"*/
char* Get_Tree_Char(FILE *fp_tree)
{
  char *line;
  int i;
  char c;

  line = (char *)mCalloc(T_MAX_LINE,sizeof(char));

  do
    c=fgetc(fp_tree);
  while((c != '(') && (c != EOF));

  if(c==EOF) return NULL;

  i=0;
  for(;;)
    {
      if((c==' ') || (c=='\n'))
	{
	  c=fgetc(fp_tree); 
	  if(c==EOF) break;
	  else continue;
	}
      
      line[i]=c;
      i++;
      c=fgetc(fp_tree);
      if(c==EOF) break;
    }
  return line;
}

/*********************************************************/

void Check_Ambiguities(allseq *data, int datatype, int stepsize)
{
  int i,j;

  Fors(j,data->crunch_len,stepsize) For(i,data->n_otu)
    {

      if(Is_Ambigu(data->c_seq[i]->state+j,
		   datatype,
		   stepsize))
	{
	  data->ambigu[j] = 1;
	  break;
	}
    }
}

/*********************************************************/
allseq *Make_Seq(int n_otu, int len, char **sp_names)
{
  allseq *alldata;
  int j;

  alldata                        = (allseq *)mCalloc(1,sizeof(allseq));
  alldata->n_otu                 = n_otu;
  alldata->c_seq                 = (seq **)mCalloc(n_otu,sizeof(seq *));
  alldata->wght                  = (int *)mCalloc(len,sizeof(int));
  alldata->pos                   = (int *)mCalloc(len,sizeof(int)); //To store positions in codons for instance
  alldata->factw                 = (double *)mCalloc(len,sizeof(double));
  alldata->b_frq                 = (double *)mCalloc(4,sizeof(double));
  alldata->c_frq                 = (double *)mCalloc(T_MAX_ALPHABET,sizeof(double));
  alldata->ambigu                = (int *)mCalloc(len,sizeof(int));
  alldata->pospatt               = (int **)mCalloc(len,sizeof(int *));
  alldata->score_site            = (double *)mCalloc(len,sizeof(double));
  alldata->whole_seq             = (int *)mCalloc(len,sizeof(int));
  //  alldata->invar                 = (int *)mCalloc(len,sizeof(int));

  alldata->crunch_len = alldata->init_len = -1;

  For(j,n_otu)
    {
      alldata->c_seq[j]          = (seq *)mCalloc(1,sizeof(seq));
      alldata->c_seq[j]->name    = (char *)mCalloc(T_MAX_NAME,sizeof(char));
      strcpy(alldata->c_seq[j]->name,sp_names[j]);
     
      alldata->c_seq[j]->state    = (char *)mCalloc(len+1,sizeof(char));
    }
  return alldata;
}


/*********************************************************/

int Is_Ambigu(char *state, int datatype, int stepsize)
{
  int i;

  if(!datatype) 
    {
      For(i,stepsize)
	{

	  if(strchr("MRWSYKBDHVNXO?-.",state[i]))
	    return 1;
	}
    }
  else
    {
      if(strchr("X?-.",state[0])) return 1;       
    }
  
  return 0;
}

/*********************************************************/
void Copy_One_State(char *from, char *to, int state_size)
{
  int i;
  For(i,state_size) to[i] = from[i];
}
/*********************************************************/
int Assign_State(char *c, int stepsize)
{
  int state[3];
  int i;

  state[0] = -1;
  For(i,stepsize)
    {
      switch(c[i])
	{
	case 'A' : state[i]=0; break;
	case 'C' : state[i]=1; break;
	case 'G' : state[i]=2; break;
	case 'T' : state[i]=3; break;
	case 'U' : state[i]=3; break;
	  
	default  : state[i]=-1;
	  break;
	}
    }
  return (stepsize>1)?(state[0]*16+state[1]*4+state[2]):(state[0]);
  
}


/*********************************************************/
void *mRealloc(void *p,int nb, size_t size)
{

  if((p = realloc(p,(size_t)nb*size)) != NULL)
	return p;
  else
    Exit("\n. Err: low memory\n");
  
  return NULL;
}

/*********************************************************/

/*Builds the arbre structure from the char* s_tree*/
arbre *Read_Tree(char *s_tree, allseq *data)
{
  char **subs;
  int i,n_ext,n_int,n_otu, opar, cpar;
  char *sub_tp;
  arbre *tree;

  tree=(arbre *)mCalloc(1,sizeof(arbre));
   
  n_otu=0;  
  opar=cpar=0;


  //Checking the structure of the tree

 
  For(i,(int)strlen(s_tree)) { 
    if ((s_tree[i] == '(')||(s_tree[i] == '{')) opar++;
    else if ((s_tree[i] == ')')||(s_tree[i] == '}')) cpar++;
  }
  if (opar!=cpar) {
    Exit("\n\nErr : Number of opening parentheses does not match number of closing parentheses\n\n");
  }


  //Counting the number of evolutionary units

  For(i,(int)strlen(s_tree)) if(s_tree[i] == ',') n_otu++;
  n_otu+=1;
  if (n_otu==1) {Exit("\n\nNo need to build a tree with only one sequence\n\n");}


  //Starting to really read the tree

  tree->n_otu = n_otu;
  Init_Tree(tree);


  //The root node parameters
  tree->noeud[n_otu]=(node *)mCalloc(1,sizeof(node));
  tree->noeud[n_otu]->v = NULL; 
  
  Make_Node_Light(tree->noeud[n_otu], data);
  tree->noeud[n_otu]->num=n_otu;
  tree->noeud[n_otu]->tax=0;
  tree->noeud[n_otu]->agglo=0;
  tree->noeud[n_otu]->ni=0;

 
  //Builds two subtrees and stores them in the subs variable 

  subs=Sub_Trees_Root(s_tree);

  //  if(subs[2][0]) {Exit("Err: rooted tree is needed\n");} else {printf ("Rooted tree...");}
  sub_tp=(char *)mCalloc((int)strlen(s_tree),sizeof(char));

  For(i,2)
    {
      strcpy(sub_tp,subs[i]);
      strcat(sub_tp,":");
      if(strstr(s_tree,sub_tp))
	{
	  tree->noeud[n_otu]->l[i] = 
	    atof((char *)strstr(s_tree,sub_tp)+strlen(subs[i])+1);// Gets the length attached to the branch leading to the root
	  
	}
    }
  Free(sub_tp);

  n_int = n_ext = 0;
  
  R_rtree(subs[0],tree->noeud[n_otu],tree,&n_int,&n_ext, data); // noeud[n_otu] is the true root of the tree !
   
  R_rtree(subs[1],tree->noeud[n_otu],tree,&n_int,&n_ext, data);
  
 
  Make_All_Edges_Light(tree->noeud[n_otu],tree->noeud[n_otu]->v[0], n_otu);// tree->noeud[0]->v[0] ??
  Make_All_Edges_Light(tree->noeud[n_otu],tree->noeud[n_otu]->v[1], n_otu);// tree->noeud[0]->v[0] ??
  i = 0;
  // Need to make the initialisation in the two subtrees
   
  Init_Tree_Edges(tree->noeud[n_otu],tree->noeud[n_otu]->v[0],tree,&i);
  Init_Tree_Edges(tree->noeud[n_otu],tree->noeud[n_otu]->v[1],tree,&i);
   
  For(i,2)
  if(subs[i]) Free(subs[i]);
  Free(subs);

  return tree;
}

/*********************************************************/


void Make_Node_Light(node *n, allseq *data)
{
  int pos;
  if(n->v) return;

  n->v     = (node **)mCalloc(3,sizeof(node *));
  n->l     = (double *)mCalloc(3,sizeof(double));
  n->b     = (edge **)mCalloc(3,sizeof(edge *));
  n->name  = (char *)mCalloc(T_MAX_NAME,sizeof(char));
  n->score = (double *)mCalloc(3,sizeof(double));
  n->BigSubtree = (node*)mCalloc(1,sizeof(node*));
  n->SmallSubtree = (node*)mCalloc(1,sizeof(node*));
  n->qtheta = (double **)mCalloc(MAX_NPOS,sizeof(double));
  for (pos=0;pos<MAX_NPOS;pos++){
    n->qtheta[pos] =(double *)mCalloc(3,sizeof(double));
  }
 
  n->cpi = (double *)mCalloc(64,sizeof(double));
  //Equilibrium GC contents in the parent branch
  n->theta = (double *)mCalloc(MAX_NPOS,sizeof(double));
  n->theta_old = (double *)mCalloc(MAX_NPOS,sizeof(double));
  n->pi = (double **)mCalloc(MAX_NPOS,sizeof(double*));

  for (pos=0;pos<MAX_NPOS;pos++){
    n->pi[pos] =(double *)mCalloc(4,sizeof(double));
  }
 

  Init_Node_Light(n, data);
  
}

/*********************************************************/

void Init_Node_Light(node *n, allseq *data)
{
  int i, pos;

  n->check_branch  = 0;
  n->is_attach     = 0;
  n->is_free       = 0;
  n->ni = n->agglo = 0;
  n->rearr         = 1; //we allow rearrangements by default

  // Accounting for non-homogeneity

  for (pos=0;pos<MAX_NPOS;pos++){
    n->pi[pos][0]=data->b_frq[0];
    n->pi[pos][1]=data->b_frq[1];
    n->pi[pos][2]=data->b_frq[2];
    n->pi[pos][3]=data->b_frq[3];
  }
  // We account for all positions
  for (pos=0;pos<MAX_NPOS;pos++) {
    n->theta_old[pos]=n->theta[pos]=n->pi[pos][1]+n->pi[pos][2];
    n->qtheta[pos][0]= n->qtheta[pos][1] =  n->qtheta[pos][2] = n->theta[pos];
  }


  For(i,3)
    {
      n->v[i]=NULL;
      n->b[i]=NULL;
      n->l[i]=-1;
    }

}

/*********************************************************/


void Init_Tree(arbre *tree)
{
  int i;
  tree->noeud              = (node **)mCalloc(2*tree->n_otu-1,sizeof(node *));
  tree->t_edges            = (edge **)mCalloc(2*tree->n_otu-2,sizeof(edge *));
  
  For(i,2*tree->n_otu-2) {
      tree->t_edges[i]         = NULL;
  }
  tree->best_tree          = NULL;
  tree->old_tree           = NULL;

  tree->best_loglk         = UNLIKELY;
  tree->tot_loglk          = UNLIKELY;
  tree->n_swap             = 0;
  tree->frac_site          = 1;
  tree->min_diff_lk        = .0;

  tree->tot_dloglk         = NULL;
  tree->tot_d2loglk        = NULL;
  tree->t_best_l           = NULL;

  tree->p_lk               = NULL;
  
  tree->n_pattern          = -1;
}

/*********************************************************/
/*Begins to explore the tree from the root which is a binary node */
char **Sub_Trees_Root(char *tree)
{
  char **subs;
  int pos1,pos2,i,dec;
  //printf("%s\n",tree);
  subs=(char **)mCalloc(2,sizeof(char *)); /*Table for the subtrees */

  For(i,2) subs[i]=(char *)mCalloc((int)strlen(tree),sizeof(char)); /* Each subtree */
  
  pos1=1;
  if ((tree[pos1] != '(')  && (tree[pos1] != '{'))
    {
      while((tree[pos1] != ',' ) && 
	    (tree[pos1] != ':' ))
	{
	  pos1++;
	}
      pos1-=1;
    }
  else pos1=Next_Par(tree,pos1); /* Finds the corresponding closing parenthesis */
  
  Test_Tree(tree+pos1+1);  /* checks if the following character is not a , or a ), or ... */
  while(*(tree+pos1+1) != ',' && *(tree+pos1+1) != ':'){pos1++;}
  strncpy(subs[0],tree+1,pos1); /* First sequence name recorded, without the branch length leading to it */
  

  
  pos2=pos1;
  dec=0;                 // Shift between the two subtrees (distance)
  while(tree[pos2] != ',') // permits to skip the distance
    {
      pos2++; 
      dec++;
    }
  pos2+=1; dec+=1;

  if((tree[pos2] != '(')&&(tree[pos2] != '{'))  // New sequence name
    {
      while(tree[pos2] != ',' &&  
	    tree[pos2] != ')' &&  
	    tree[pos2] != '}' &&
	    tree[pos2] != ':') 
	pos2++;
     pos2-=1;
    }
  else pos2=Next_Par(tree,pos2); // else new subtree to consider

  Test_Tree(tree+pos2+1);   /* checks if the following character is not a , or a ), or ... */
  while((*(tree+pos2+1) != ',') 
	&& (*(tree+pos2+1) != ':')
	&& (*(tree+pos2+1) != ')')
	&& (*(tree+pos2+1) != '}')
	){pos2++;}
  strncpy(subs[1],tree+pos1+dec,pos2-pos1-dec+1); // Here is the new subtree to consider

  return subs;
}

/*********************************************************/
void R_rtree(char *s_tree, node *pere, arbre *tree, int *n_int, int *n_ext, allseq *data) 
{
  int i;
  char *sub_tp;
  node *fils;
  int n_otu = tree->n_otu;
  i=0;
  if(strstr(s_tree," ")) Exit("\n Err : tree must not contain a ' ' character\n");

  fils=(node *)mCalloc(1,sizeof(node));  
  fils->v = NULL;
  Make_Node_Light(fils, data);

  if ((s_tree[0] == '(')||(s_tree[0] == '{'))  // if the subtree begins with a parenthesis: current node is not a leaf.
    {
      char **subs;
      (*n_int)+=1;                                  
      tree->noeud[n_otu+*n_int]=fils;
      fils->num=n_otu+*n_int; // we give the node its number
      fils->tax=0;
      fils->agglo=0;
      fils->ni=0;
      
      if(s_tree[(int)strlen(s_tree)-1] == '*') 
	{
	  fils->check_branch = 1;
	  s_tree[(int)strlen(s_tree)-1] = '\0';
	}


      For(i,3) 
       {
	 if(!pere->v[i])
	   {
	     pere->v[i]=fils;
	     fils->l[0]=pere->l[i];
	     break;
	   }
       }
      
      fils->v[0]=pere;           //Father node will always be neighbor 0, except for the root

      if (s_tree[0] == '{') {
	fils->rearr = 0; //Rearrangements are forbidden between this subtree and the rest of the tree
      }

      subs=Sub_Trees_Root(s_tree);      
      sub_tp = (char *)mCalloc((int)strlen(s_tree),sizeof(char));
      strcpy(sub_tp,subs[0]);
      strcat(sub_tp,":");
      if(strstr(s_tree,sub_tp))
	{
	  fils->l[1] = atof((char *)strstr(s_tree,sub_tp)
			    +(int)strlen(subs[0])+1); //Records the length of the branch
	}
      
      strcpy(sub_tp,subs[1]);
      strcat(sub_tp,":");
      if(strstr(s_tree,sub_tp))
	{
	  fils->l[2] = atof((char *)strstr(s_tree,sub_tp)
			    +(int)strlen(subs[1])+1); //Records the length of the branch
	}
      
      Free(sub_tp);
      R_rtree(subs[0],fils,tree,n_int,n_ext, data);
      R_rtree(subs[1],fils,tree,n_int,n_ext, data);
      Free(subs[0]); Free(subs[1]);
      Free(subs);
    }

  else           // if the subtree does not begin with a parenthesis: current node is a leaf.         
    {
      tree->noeud[*n_ext]=fils;
      fils->tax=1;
      fils->agglo=1;
      fils->ni=1;
     
      

      For(i,3)
	{
	 if(!pere->v[i])
	   {
	     pere->v[i]=fils;
	     fils->l[0]=pere->l[i];
	     break;
	   }
	}
      
      if(s_tree[(int)strlen(s_tree)-1] == '*') 
	{
	  fils->check_branch = 1;
	  s_tree[(int)strlen(s_tree)-1] = '\0';
	}
      

      fils->v[0]=pere;  //Father node will always be neighbor 0, except for the root
      strcpy(fils->name,s_tree); 
      fils->num=*n_ext;
      (*n_ext)+=1;
    }
}

/*********************************************************/
void Make_All_Edges_Light(node *a, node *d, int n_otu)
{
  int i;

  Make_Edge_Light(a, d, n_otu);
 
  if(d->tax) return;
  else
    {
      For(i,3)
	{
	  if(d->v[i] != a)
	    Make_All_Edges_Light(d,d->v[i], n_otu);
	}
    }
}

/*********************************************************/
int Next_Par(char *s, int pos)
{
  int curr;
  curr=pos+1;
  
  while ((*(s+curr) != ')')&&(*(s+curr) != '}'))
    {
      if((*(s+curr) == '(')||(*(s+curr) == '{')) curr=Next_Par(s,curr);
      curr++;
    }

  return curr; 
}

/*********************************************************/


void Test_Tree(char *s)
{
  errno = 0;
  if(*s != ',' && *s != ')' && *s != ':' && *s != '*' && *s != '}' &&(!strtod(s,(char **)NULL) && errno))
    {
      printf("\n ! Tree bad format !\n");
      exit(1);
    }
  return;
}

/*********************************************************/
void Init_Tree_Edges(node *a, node *d, arbre *tree, int *cur)
{
  int i,dir_a_d;
  edge *br;
  dir_a_d = -1;
  if (a->num==tree->n_otu) {          //If we're at the root
    For(i,2) if(a->v[i] == d) {dir_a_d = i; break;}
  }
  else {
    For(i,3) if(a->v[i] == d) {dir_a_d = i; break;}
  }
  br=a->b[dir_a_d];
   tree->t_edges[*cur] = a->b[dir_a_d];
  tree->t_edges[*cur]->num = *cur; 
  tree->t_edges[*cur]->qmat_indice = *cur;
  
   *cur = *cur + 1;
  if(d->tax) {return;}
  else
    {
      For(i,3)
	{
	  if(d->v[i] != a)
	    Init_Tree_Edges(d,d->v[i],tree,cur);
	}
    }

}

/*********************************************************/
void Make_Edge_Light(node *a, node *d, int n_otu)
{
  edge *b;

  b = (edge *)mCalloc(1,sizeof(edge));

  
  b->left = a;  b->rght = d; 
  if(a->tax) {b->rght = a; b->left = d;} /* root */
  /* a tip is necessary on the right side of the edge */ 
  
   
  Make_Edge_Dirs(b,a,d, n_otu);
  b->l                    = a->l[b->l_r];
  if(a->tax) b->l         = a->l[b->r_l];
  if(b->l < BL_MIN)  b->l = BL_MIN;
  b->l_old                = b->l;
  Init_Edge_Light(b);
  b->rearr=b->rght->rearr;
}
/*********************************************************/
void Init_Edge_Light(edge *b)
{
  b->in_both         = 0;
  b->nj_score        = .0;


  b->site_dlk        = -1.;
  b->site_d2lk       = -1.;
  b->site_dlk_rr     = NULL;
  b->site_d2lk_rr    = NULL;
  b->p_lk_left       = NULL;
  b->p_lk_rght       = NULL;
  b->Pij_rr          = NULL;
  b->dPij_rr         = NULL;
  b->d2Pij_rr        = NULL;
 
}

/*********************************************************/
/*Computes the substitution probabilities along the edge b*/
/*whose ancestor is node left*/
void Computes_Substitution_Probabilities (edge *b, arbre *tree, int catg, int pos) {
  double theta, kappa,l, rr;
  l= b->l * tree->mod->l_coef[pos];
  theta = b->rght->theta[pos];
  kappa = tree->mod->kappa[pos];
  rr=tree->mod->rr[catg];

  /* Here we do take the gamma distribution coefficient in account, with the category integer*/
  /* A=0; C=1, G=2, T=3 */

  /*PAA = PTT  */
  b->Pij_rr[pos][catg][0][0] = b->Pij_rr[pos][catg][3][3] = (1.-theta)/2.*(1.+exp(- rr * l))+theta*exp(-(kappa+1.)/2.* rr *l);


  /*PCC = PGG  */
  b->Pij_rr[pos][catg][1][1] = b->Pij_rr[pos][catg][2][2] = (theta)/2.*(1.+exp(- rr *l))+(1.-theta)*exp(-(kappa+1.)/2.* rr *l);


  /*PGA = PCT  */
  b->Pij_rr[pos][catg][2][0] = b->Pij_rr[pos][catg][1][3] = (1.-theta)/2.*(1.+exp(- rr * l))-(1.-theta)*exp(-(kappa+1.)/2.* rr * l);


  /*PAG = PTC  */
  b->Pij_rr[pos][catg][0][2] = b->Pij_rr[pos][catg][3][1] = (theta)/2.*(1.+exp(- rr *l))-theta*exp(-(kappa+1.)/2.* rr *l);


  /*PAT = PTA = PGT = PCA  */
  b->Pij_rr[pos][catg][0][3] = b->Pij_rr[pos][catg][3][0] = b->Pij_rr[pos][catg][2][3] = b->Pij_rr[pos][catg][1][0] = (1.-theta)/2.*(1.-exp(- rr * l));


  /*PCG = PGC = PTG = PAC  */
  b->Pij_rr[pos][catg][1][2] = b->Pij_rr[pos][catg][2][1] = b->Pij_rr[pos][catg][3][2] = b->Pij_rr[pos][catg][0][1] = theta/2.*(1.-exp(- rr *l));

  // Computes_Derivative_Substitution_Probabilities (b, tree, catg);
  // Computes_Second_Derivative_Substitution_Probabilities (b, tree, catg);
}
/*********************************************************/
/* Computes the derivative of the substitution probabilities */
/* for length optimization */
void Computes_Derivative_Substitution_Probabilities (edge *b, arbre *tree, int catg, int pos) {
  double theta, kappa,l,rr;
  l= b->l * tree->mod->l_coef[pos];
  theta = b->rght->theta[pos];
  kappa = tree->mod->kappa[pos];
  rr=tree->mod->rr[catg];

/* Here we do take the gamma distribution coefficient in account, with the category integer*/
  /* A=0; C=1, G=2, T=3 */

  /*dPAA = dPTT  */
  b->dPij_rr[pos][catg][0][0] = b->dPij_rr[pos][catg][3][3] = -((1.-theta)/2.) * ( rr * exp(- rr * l))+ theta*(-(kappa+1.)/2.)*rr *exp(-(kappa+1.)/2.*rr*l);


  /*dPCC = dPGG  */
  b->dPij_rr[pos][catg][1][1] = b->dPij_rr[pos][catg][2][2] = -((theta)/2.) *  rr * exp(- rr * l) + (1.-theta)* (-(kappa+1.)/2.) * rr * exp(-(kappa+1.)/2.* rr * l);


  /*dPGA = dPCT  */
  b->dPij_rr[pos][catg][2][0] = b->dPij_rr[pos][catg][1][3] = -((1.-theta)/2.) *  rr * exp(- rr * l) + (1.-theta)* ((kappa+1.)/2.) * rr * exp(-((kappa+1.)/2.) * rr * l);


  /*dPAG = dPTC  */
 b->dPij_rr[pos][catg][0][2] = b->dPij_rr[pos][catg][3][1] = - ((theta)/2.) * rr * exp(- rr * l) + theta * ((kappa+1.)/2.) * rr * exp(-((kappa+1.)/2.) * rr * l);


  /*dPAT = dPTA = dPGT = dPCA  */
   b->dPij_rr[pos][catg][0][3] = b->dPij_rr[pos][catg][3][0] = b->dPij_rr[pos][catg][2][3] = b->dPij_rr[pos][catg][1][0] = ((1.-theta)/2.) * rr * exp(- rr * l);


  /*dPCG = dPGC = dPTG = dPAC  */
  b->dPij_rr[pos][catg][1][2] = b->dPij_rr[pos][catg][2][1] = b->dPij_rr[pos][catg][3][2] = b->dPij_rr[pos][catg][0][1] = (theta/2.) * rr * exp(- rr * l);

}

/*********************************************************/


/* Computes the second derivative of the substitution probabilities */
/* for length optimization */
void Computes_Second_Derivative_Substitution_Probabilities (edge *b, arbre *tree, int catg, int pos) {
  double theta, kappa, l, rr, rr2;

  l= b->l * tree->mod->l_coef[pos];
  if (l < BL_MIN) l = BL_MIN;
  theta = b->rght->theta[pos];
  kappa = tree->mod->kappa[pos];
  rr = tree->mod->rr[catg];
  rr2 = rr*rr;
 


/* Here we do take the gamma distribution coefficient in account, with the category integer*/
  /* A=0; C=1, G=2, T=3 */

  /*d2PAA = d2PTT  */
  b->d2Pij_rr[pos][catg][0][0] = b->d2Pij_rr[pos][catg][3][3] = ((1.-theta)/2.) * rr2 * exp(- rr * l) + theta*((kappa+1.)/2.) *((kappa+1.)/2.) * rr2 * exp(-((kappa+1.)/2.)* rr * l);


  /*d2PCC = d2PGG  */
 b->d2Pij_rr[pos][catg][1][1] = b->d2Pij_rr[pos][catg][2][2] = ((theta)/2.) * rr2 * exp(- rr * l) + (1.-theta)* ((kappa+1.)/2.) * ((kappa+1.)/2.) * rr2 * exp(-((kappa+1.)/2.) * rr * l);


  /*d2PGA = d2PCT  */
 b->d2Pij_rr[pos][catg][2][0] = b->d2Pij_rr[pos][catg][1][3] = ((1.-theta)/2.) * rr2 * exp(- rr * l) - (1.-theta) * ((kappa+1.)/2.) * ((kappa+1.)/2.) * rr2 * exp(-((kappa+1.)/2.)* rr * l);


  /*d2PAG = d2PTC  */
 b->d2Pij_rr[pos][catg][0][2] = b->d2Pij_rr[pos][catg][3][1] = ((theta)/2.) * (rr2 * exp(- rr * l)) - theta * ((kappa+1.)/2.) * ((kappa+1.)/2.) * rr2 * exp(-((kappa+1.)/2.) * rr * l);


  /*d2PAT = d2PTA = d2PGT = d2PCA  */
  b->d2Pij_rr[pos][catg][0][3] = b->d2Pij_rr[pos][catg][3][0] = b->d2Pij_rr[pos][catg][2][3] = b->d2Pij_rr[pos][catg][1][0] = -((1.-theta)/2.) * rr2 * exp(- rr * l);


  /*d2PCG = d2PGC = d2PTG = d2PAC  */
 b->d2Pij_rr[pos][catg][1][2] = b->d2Pij_rr[pos][catg][2][1] = b->d2Pij_rr[pos][catg][3][2] = b->d2Pij_rr[pos][catg][0][1] = - (theta/2.) * rr2 * exp(- rr * l);

}


/*********************************************************/

void Make_Edge_Dirs(edge *b, node *a, node *d, int n_otu)
{
  int i;

  b->l_r = b->r_l = -1;
  if (a->num==n_otu) {
    For(i,2)
      {
	if((a->v[i]) && (a->v[i] == d)) 
	  {
	    b->l_r  = i;
	    a->b[i] = b;
	  }
	if((d->v[i]) && (d->v[i] == a)) 
	  {
	    b->r_l  = i;
	    d->b[i] = b;
	  }
      }
    
  }
  else {
    For(i,3)
      {
	if((a->v[i]) && (a->v[i] == d)) 
	  {
	  b->l_r  = i;
	  a->b[i] = b;
	  }
	if((d->v[i]) && (d->v[i] == a)) 
	{
	  b->r_l  = i;
	  d->b[i] = b;
	}
      }
  }
  
  if(a->tax) {b->r_l = 0; For(i,3) if(d->v[i]==a) {b->l_r = i; break;}}

  
  b->l_v1 = b->l_v2 = b->r_v1 = b->r_v2 = -1;
  For(i,3)
    {
      if(b->left->v[i] != b->rght)
	{
	  if(b->l_v1 < 0) b->l_v1 = i;
	  else            b->l_v2 = i;
	}
     
      if(b->rght->v[i] != b->left)
	{
	  if(b->r_v1 < 0) b->r_v1 = i;
	  else            b->r_v2 = i;
	}
    }
}


/*********************************************************/
/*Orders sequences so that their numero corresponds to the one they have got in the tree*/
void Order_Tree_CSeq(arbre *tree, allseq *data)
{
  int i,j,k,n_otu_tree,n_otu_seq, pos;
  seq *buff;
  double fA, fC, fG, fT;

  n_otu_tree = tree->n_otu;
  n_otu_seq  = data->n_otu;
 
  if(n_otu_tree != n_otu_seq) Exit("\n. The number of tips in the tree is not the same as the number of sequences\n");
  
  printf("\n\n\t\t\tSequences GC contents\n");
  
  For(i,MAX(n_otu_tree,n_otu_seq))
    {
      For(j,MIN(n_otu_tree,n_otu_seq))
	{
	  if(!strcmp(tree->noeud[i]->name,data->c_seq[j]->name))
	    break;
	}

      if(j==MIN(n_otu_tree,n_otu_seq))
	{
	  printf("\n. Err: %s is not found in sequences data set\n",
		 tree->noeud[i]->name);
	  Exit("");
	}
      buff = data->c_seq[j];
      data->c_seq[j] = data->c_seq[i];
      data->c_seq[i] = buff;

     
      //Compute the ATCG ratios.

      fA=fC=fG=fT=0;
    
      for (k=0;k<data->c_seq[i]->len;k++) {
	if (data->c_seq[i]->state[k]=='A') fA+=data->wght[k];
	else if (data->c_seq[i]->state[k]=='C') fC+=data->wght[k];
	else if (data->c_seq[i]->state[k]=='G') fG+=data->wght[k];
	else if (data->c_seq[i]->state[k]=='T') fT+=data->wght[k];
      }
      fA/=data->init_len;
      fC/=data->init_len;
      fG/=data->init_len;
      fT/=data->init_len;
      
      //Assign the ratios to node pi, without distinguishing between positions

      /* printf("Before tree->mod->n_pos\n");
      printf("Before tree->mod->n_pos : %d\n",tree->mod->n_pos  );
      */

      for (pos=0;pos<tree->mod->n_pos;pos++) {
	tree->noeud[i]->pi[pos][0]=fA;
	tree->noeud[i]->pi[pos][1]=fC;
	tree->noeud[i]->pi[pos][2]=fG;
	tree->noeud[i]->pi[pos][3]=fT;
      }
   
      if (tree->mod->n_pos>=2){} else {
	printf ("%s sequence GC content :\t%f\n",tree->noeud[i]->name,tree->noeud[i]->pi[0][1]+tree->noeud[i]->pi[0][2]);
      }	
    }
}

/*********************************************************/
void Make_Edge_Lk(node *a, node *d, arbre *tree)
{
  int i,j, pos;
  int len;
  edge *b;
  b = NULL;

  For(i,3) if((a->v[i]) && (a->v[i] == d)) {b = a->b[i]; break;}// Find the edge we are interested in 

  len = (int)tree->data->crunch_len;

  b->diff_lk   = 0.0;
  b->l_old     = b->l;

  if(!b->Pij_rr)
    {
      b->best_conf = 1;
      b->ql        = (double *)mCalloc(3,sizeof(double));


      b->Pij_rr   = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      
      b->dPij_rr  = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      b->d2Pij_rr = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      b->dPij_rr_theta  = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      b->dPij_rr_frac  = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      b->d2Pij_rr_frac = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      b->dPij_rr_tot  = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      b->d2Pij_rr_tot = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      b->dPij_rr_kappa  = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      b->d2Pij_rr_kappa = (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));

      /*   b->dPij_rr  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
       b->d2Pij_rr = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
       
       b->dPij_rr_theta  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
       b->dPij_rr_frac  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
       b->d2Pij_rr_frac = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
       b->dPij_rr_tot  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
       b->d2Pij_rr_tot = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
       b->dPij_rr_kappa  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
       b->d2Pij_rr_kappa = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
      */

      For(pos,tree->mod->n_pos) 
	{
	  b->Pij_rr[pos]   = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  b->dPij_rr[pos]  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  b->d2Pij_rr[pos] = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  
	  b->dPij_rr_theta[pos]  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  b->dPij_rr_frac[pos]  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  b->d2Pij_rr_frac[pos] = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  b->dPij_rr_tot[pos]  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  b->d2Pij_rr_tot[pos] = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  b->dPij_rr_kappa[pos]  = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  b->d2Pij_rr_kappa[pos] = (double ***)mCalloc(tree->mod->n_catg,sizeof(double **));
	  
	  For(i,tree->mod->n_catg)
	    {
	      b->Pij_rr[pos][i]   = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      b->dPij_rr[pos][i]  = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      b->d2Pij_rr[pos][i] = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      
	      b->dPij_rr_theta[pos][i]  = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      b->dPij_rr_frac[pos][i]  = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      b->d2Pij_rr_frac[pos][i] = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      b->dPij_rr_tot[pos][i]  = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      b->d2Pij_rr_tot[pos][i] = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      b->dPij_rr_kappa[pos][i]  = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      b->d2Pij_rr_kappa[pos][i] = (double **)mCalloc(tree->mod->ns,sizeof(double *));
	      
	      For(j,tree->mod->ns)
		{
		  b->Pij_rr[pos][i][j]   = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  b->dPij_rr[pos][i][j]  = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  b->d2Pij_rr[pos][i][j] = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  
		  b->dPij_rr_theta[pos][i][j]  = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  b->dPij_rr_frac[pos][i][j]  = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  b->d2Pij_rr_frac[pos][i][j] = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  b->dPij_rr_tot[pos][i][j]  = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  b->d2Pij_rr_tot[pos][i][j] = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  b->dPij_rr_kappa[pos][i][j]  = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		  b->d2Pij_rr_kappa[pos][i][j] = (double *)mCalloc(tree->mod->ns,sizeof(double ));
		}
	    }
	}

      b->rght_nodes=(int*)mCalloc(tree->n_otu,sizeof(int ));
      b->left_nodes=(int*)mCalloc(tree->n_otu,sizeof(int ));
      /*    
      b->site_p_lk_left = (double **)mCalloc((int)tree->mod->n_catg,sizeof(double *));
      For(i,tree->mod->n_catg) b->site_p_lk_left[i] = (double *)mCalloc(tree->mod->ns,sizeof(double));
      b->site_p_lk_rght = (double **)mCalloc((int)tree->mod->n_catg,sizeof(double *));
      For(i,tree->mod->n_catg) b->site_p_lk_rght[i] = (double *)mCalloc(tree->mod->ns,sizeof(double));
      */

      /*  
      b->dp_lk_kappa_rght = (double **)mCalloc((int)tree->mod->n_catg,sizeof(double *));
      For(i,tree->mod->n_catg) b->site_dp_lk_kappa_rght[i] = (double *)mCalloc(tree->mod->ns,sizeof(double));
      b->d2p_lk_kappa_rght = (double **)mCalloc((int)tree->mod->n_catg,sizeof(double *));
      For(i,tree->mod->n_catg) b->site_d2p_lk_kappa_rght[i] = (double *)mCalloc(tree->mod->ns,sizeof(double));
      */

      b->site_dlk_rr  = (double *)mCalloc((int)tree->mod->n_catg,sizeof(double));
      b->site_d2lk_rr = (double *)mCalloc((int)tree->mod->n_catg,sizeof(double));
  
      b->scale_left = b->scale_rght = 0;

      b->p_lk_left   = NULL;
      b->p_lk_rght  = NULL;
      
      b->sum_scale_f_rght = NULL;
      b->sum_scale_f_left  = NULL;

      b->get_p_lk_left = 0;
      b->get_p_lk_rght = 0;
      b->ud_p_lk_left  = 0;
      b->ud_p_lk_rght  = 0;


      b->numrght=0;
      b->numleft=0;
   
    }
}

/*********************************************************/
void Make_Node_Lk(node *n)
{
  n->n_ex_nodes = (int *)mCalloc(2,sizeof(int));
  return;
}

/*********************************************************/
void Alloc_All_P_Lk(arbre *tree)
{
  int i,j,k, pos;

  For(i,2*tree->n_otu-2)  // The rooted tree has got one more branch than the unrooted one !
    {
      tree->t_edges[i]->get_p_lk_left = 1;
      tree->t_edges[i]->get_p_lk_rght = 1;

      tree->t_edges[i]->p_lk_left = 
      (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      
      tree->t_edges[i]->p_lk_rght = 
      (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));

      tree->t_edges[i]->dp_lk_kappa_rght = 
      (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));
      
      tree->t_edges[i]->d2p_lk_kappa_rght = 
      (double ****)mCalloc(tree->mod->n_pos,sizeof(double ***));

      For(pos, tree->mod->n_pos) {
	tree->t_edges[i]->p_lk_left[pos] =
	  (double ***)mCalloc(tree->data->crunch_len,sizeof(double **));
	tree->t_edges[i]->p_lk_rght[pos] = 
	  (double ***)mCalloc(tree->data->crunch_len,sizeof(double **));
	For(j,tree->data->crunch_len)
	  {
	    tree->t_edges[i]->p_lk_left[pos][j] = 
	      (double **)mCalloc(tree->mod->n_catg,sizeof(double *));
	    
	    For(k,tree->mod->n_catg)
	      tree->t_edges[i]->p_lk_left[pos][j][k] = 
	      (double *)mCalloc(tree->mod->ns,sizeof(double ));
	    
	    tree->t_edges[i]->p_lk_rght[pos][j] = 
	      (double **)mCalloc(tree->mod->n_catg,sizeof(double *));
	    
	    For(k,tree->mod->n_catg)
	      tree->t_edges[i]->p_lk_rght[pos][j][k] = 
	      (double *)mCalloc(tree->mod->ns,sizeof(double ));
	  }
      }

      For(pos, tree->mod->n_pos) {
	tree->t_edges[i]->dp_lk_kappa_rght[pos] = 
	  (double ***)mCalloc(tree->data->crunch_len,sizeof(double **));
	tree->t_edges[i]->d2p_lk_kappa_rght[pos] =
	  (double ***)mCalloc(tree->data->crunch_len,sizeof(double **));
	For(j,tree->data->crunch_len)
	  {
	    tree->t_edges[i]->dp_lk_kappa_rght[pos][j] = 
	      (double **)mCalloc(tree->mod->n_catg,sizeof(double *));
	    
	    For(k,tree->mod->n_catg)
	      tree->t_edges[i]->dp_lk_kappa_rght[pos][j][k] = 
	      (double *)mCalloc(tree->mod->ns,sizeof(double ));
	    
	    tree->t_edges[i]->d2p_lk_kappa_rght[pos][j] = 
	      (double **)mCalloc(tree->mod->n_catg,sizeof(double *));
	    
	    For(k,tree->mod->n_catg)
	      tree->t_edges[i]->d2p_lk_kappa_rght[pos][j][k] = 
	      (double *)mCalloc(tree->mod->ns,sizeof(double ));
	    
	  }
      }
      
      tree->t_edges[i]->sum_scale_f_left = 
      (double *)mCalloc(tree->data->crunch_len,sizeof(double ));
      
      tree->t_edges[i]->sum_scale_f_rght = 
      (double *)mCalloc(tree->data->crunch_len,sizeof(double ));
      
    }
}

/*********************************************************/
int DiscreteGamma (double freqK[], double rK[], 
    double alfa, double beta, int K, int median)
{
/* discretization of gamma distribution with equal proportions in each 
   category
*/
   int i;
   double gap05=1.0/(2.0*K), t, factor=alfa/beta*K, lnga1;

  
   if(K==1) 
     {
       rK[0] = 1.0;
       return 0;
     }

   if (median) {
      for (i=0; i<K; i++) rK[i]=PointGamma((i*2.0+1)*gap05, alfa, beta);
      for (i=0,t=0; i<K; i++) t+=rK[i];
      for (i=0; i<K; i++)     rK[i]*=factor/t;
   }
   else {
      lnga1=LnGamma(alfa+1);
      for (i=0; i<K-1; i++)
	 freqK[i]=PointGamma((i+1.0)/K, alfa, beta);
      for (i=0; i<K-1; i++)
	 freqK[i]=IncompleteGamma(freqK[i]*beta, alfa+1, lnga1);
      rK[0] = freqK[0]*factor;
      rK[K-1] = (1-freqK[K-2])*factor;
      for (i=1; i<K-1; i++)  rK[i] = (freqK[i]-freqK[i-1])*factor;
   }
   for (i=0; i<K; i++) freqK[i]=1.0/K;

   return (0);
}

/*********************************************************/
double LnGamma (double alpha)
{
/* returns ln(gamma(alpha)) for alpha>0, accurate to 10 decimal places.  
   Stirling's formula is used for the central polynomial part of the procedure.
   Pike MC & Hill ID (1966) Algorithm 291: Logarithm of the gamma function.
   Communications of the Association for Computing Machinery, 9:684
*/
   double x=alpha, f=0, z;

   if (x<7) {
      f=1;  z=x-1;
      while (++z<7)  f*=z;
      x=z;   f=-log(f);
   }
   z = 1/(x*x);
   return  f + (x-0.5)*log(x) - x + .918938533204673 
	  + (((-.000595238095238*z+.000793650793651)*z-.002777777777778)*z
	       +.083333333333333)/x;  
}

/*********************************************************/
double IncompleteGamma (double x, double alpha, double ln_gamma_alpha)
{
/* returns the incomplete gamma ratio I(x,alpha) where x is the upper 
	   limit of the integration and alpha is the shape parameter.
   returns (-1) if in error
   ln_gamma_alpha = ln(Gamma(alpha)), is almost redundant.
   (1) series expansion     if (alpha>x || x<=1)
   (2) continued fraction   otherwise
   RATNEST FORTRAN by
   Bhattacharjee GP (1970) The incomplete gamma integral.  Applied Statistics,
   19: 285-287 (AS32)
*/
   int i;
   double p=alpha, g=ln_gamma_alpha;
   double accurate=1e-8, overflow=1e30;
   double factor, gin=0, rn=0, a=0,b=0,an=0,dif=0, term=0, pn[6];

   if (x==0) return (0);
   if (x<0 || p<=0) return (-1);

   factor=exp(p*log(x)-x-g);   
   if (x>1 && x>=p) goto l30;
   /* (1) series expansion */
   gin=1;  term=1;  rn=p;
 l20:
   rn++;
   term*=x/rn;   gin+=term;

   if (term > accurate) goto l20;
   gin*=factor/p;
   goto l50;
 l30:
   /* (2) continued fraction */
   a=1-p;   b=a+x+1;  term=0;
   pn[0]=1;  pn[1]=x;  pn[2]=x+1;  pn[3]=x*b;
   gin=pn[2]/pn[3];
 l32:
   a++;  b+=2;  term++;   an=a*term;
   for (i=0; i<2; i++) pn[i+4]=b*pn[i+2]-an*pn[i];
   if (pn[5] == 0) goto l35;
   rn=pn[4]/pn[5];   dif=fabs(gin-rn);
   if (dif>accurate) goto l34;
   if (dif<=accurate*rn) goto l42;
 l34:
   gin=rn;
 l35:
   for (i=0; i<4; i++) pn[i]=pn[i+2];
   if (fabs(pn[4]) < overflow) goto l32;
   for (i=0; i<4; i++) pn[i]/=overflow;
   goto l32;
 l42:
   gin=1-factor*gin;

 l50:
   return (gin);
}


/*********************************************************/
double PointChi2 (double prob, double v)
{
/* returns z so that Prob{x<z}=prob where x is Chi2 distributed with df=v
   returns -1 if in error.   0.000002<prob<0.999998
   RATNEST FORTRAN by
       Best DJ & Roberts DE (1975) The percentage points of the 
       Chi2 distribution.  Applied Statistics 24: 385-388.  (AS91)
   Converted into C by Ziheng Yang, Oct. 1993.
*/
   double e=.5e-6, aa=.6931471805, p=prob, g;
   double xx, c, ch, a=0,q=0,p1=0,p2=0,t=0,x=0,b=0,s1,s2,s3,s4,s5,s6;

   if (p<.000002 || p>.999998 || v<=0) return (-1);

   g = LnGamma (v/2);
   xx=v/2;   c=xx-1;
   if (v >= -1.24*log(p)) goto l1;

   ch=pow((p*xx*exp(g+xx*aa)), 1/xx);
   if (ch-e<0) return (ch);
   goto l4;
l1:
   if (v>.32) goto l3;
   ch=0.4;   a=log(1-p);
l2:
   q=ch;  p1=1+ch*(4.67+ch);  p2=ch*(6.73+ch*(6.66+ch));
   t=-0.5+(4.67+2*ch)/p1 - (6.73+ch*(13.32+3*ch))/p2;
   ch-=(1-exp(a+g+.5*ch+c*aa)*p2/p1)/t;
   if (fabs(q/ch-1)-.01 <= 0) goto l4;
   else                       goto l2;
  
l3: 
   x=PointNormal (p);
   p1=0.222222/v;   ch=v*pow((x*sqrt(p1)+1-p1), 3.0);
   if (ch>2.2*v+6)  ch=-2*(log(1-p)-c*log(.5*ch)+g);
l4:
   q=ch;   p1=.5*ch;
   if ((t=IncompleteGamma (p1, xx, g))<0) {
      printf ("\nerr IncompleteGamma");
      return (-1);
   }
   p2=p-t;
   t=p2*exp(xx*aa+g+p1-c*log(ch));   
   b=t/ch;  a=0.5*t-b*c;

   s1=(210+a*(140+a*(105+a*(84+a*(70+60*a))))) / 420;
   s2=(420+a*(735+a*(966+a*(1141+1278*a))))/2520;
   s3=(210+a*(462+a*(707+932*a)))/2520;
   s4=(252+a*(672+1182*a)+c*(294+a*(889+1740*a)))/5040;
   s5=(84+264*a+c*(175+606*a))/2520;
   s6=(120+c*(346+127*c))/5040;
   ch+=t*(1+0.5*t*s1-b*c*(s1-b*(s2-b*(s3-b*(s4-b*(s5-b*s6))))));
   if (fabs(q/ch-1) > e) goto l4;

   return (ch);
}

/*********************************************************/
double PointNormal (double prob)
{
/* returns z so that Prob{x<z}=prob where x ~ N(0,1) and (1e-12)<prob<1-(1e-12)
   returns (-9999) if in error
   Odeh RE & Evans JO (1974) The percentage points of the normal distribution.
   Applied Statistics 22: 96-97 (AS70)

   Newer methods:
     Wichura MJ (1988) Algorithm AS 241: the percentage points of the
       normal distribution.  37: 477-484.
     Beasley JD & Springer SG  (1977).  Algorithm AS 111: the percentage 
       points of the normal distribution.  26: 118-121.

*/
   double a0=-.322232431088, a1=-1, a2=-.342242088547, a3=-.0204231210245;
   double a4=-.453642210148e-4, b0=.0993484626060, b1=.588581570495;
   double b2=.531103462366, b3=.103537752850, b4=.0038560700634;
   double y, z=0, p=prob, p1;

   p1 = (p<0.5 ? p : 1-p);
   if (p1<1e-20) return (-9999);

   y = sqrt (log(1/(p1*p1)));   
   z = y + ((((y*a4+a3)*y+a2)*y+a1)*y+a0) / ((((y*b4+b3)*y+b2)*y+b1)*y+b0);
   return (p<0.5 ? -z : z);
}
/*********************************************************/
void qksort(double* A, int ilo, int ihi)
{
    double pivot;	// pivot value for partitioning array
    int ulo, uhi;	// indices at ends of unpartitioned region
    int ieq;		// least index of array entry with value equal to pivot
    double tempEntry;	// temporary entry used for swapping

    if (ilo >= ihi) {
	return;
    }
    // Select a pivot value.
    pivot = A[(ilo + ihi)/2];
    // Initialize ends of unpartitioned region and least index of entry
    // with value equal to pivot.
    ieq = ulo = ilo;
    uhi = ihi;
    // While the unpartitioned region is not empty, try to reduce its size.
    while (ulo <= uhi) {
	if (A[uhi] > pivot) {
	    // Here, we can reduce the size of the unpartitioned region and
	    // try again.
	    uhi--;
	} else {
	    // Here, A[uhi] <= pivot, so swap entries at indices ulo and
	    // uhi.
	    tempEntry = A[ulo];
	    A[ulo] = A[uhi];
	    A[uhi] = tempEntry;
	    // After the swap, A[ulo] <= pivot.
	    if (A[ulo] < pivot) {
		// Swap entries at indices ieq and ulo.
		tempEntry = A[ieq];
		A[ieq] = A[ulo];
		A[ulo] = tempEntry;
		// After the swap, A[ieq] < pivot, so we need to change
		// ieq.
		ieq++;
		// We also need to change ulo, but we also need to do
		// that when A[ulo] = pivot, so we do it after this if
		// statement.
	    }
	    // Once again, we can reduce the size of the unpartitioned
	    // region and try again.
	    ulo++;
	}
    }
    // Now, all entries from index ilo to ieq - 1 are less than the pivot
    // and all entries from index uhi to ihi + 1 are greater than the
    // pivot.  So we have two regions of the array that can be sorted
    // recursively to put all of the entries in order.
    qksort(A, ilo, ieq - 1);
    qksort(A, uhi + 1, ihi);
}

/*********************************************************/
void Swap(node *a, node *b, node *c, node *d, arbre *tree)
{
  int ab, ba, cd, dc;
  int i;


  /* \             /d      \             /a
      \           /         \           / 
       \b__...__c/    ->     \b__...__c/  
       /         \	     /         \  
      /           \	    /           \ 
     /a            \  	   /d            \ */


 
  ab = ba = cd = dc = -1;

  For(i,3) if(a->v[i] == b) { ab = i; break; }
  For(i,3) if(b->v[i] == a) { ba = i; break; }
  For(i,3) if(c->v[i] == d) { cd = i; break; }
  For(i,3) if(d->v[i] == c) { dc = i; break; }
 
  a->v[ab] = c;
  d->v[dc] = b;
  b->v[ba] = d;
  c->v[cd] = a;
  b->b[ba] = d->b[dc];
  c->b[cd] = a->b[ab];

  (a->b[ab]->left == b)?
    (a->b[ab]->left = c):
    (a->b[ab]->rght = c);
  
  (d->b[dc]->left == c)?
    (d->b[dc]->left = b):
    (d->b[dc]->rght = b);
 
  For(i,3)
    {
      if(a->b[ab]->left->v[i] == a->b[ab]->rght) a->b[ab]->l_r = i;
      if(a->b[ab]->rght->v[i] == a->b[ab]->left) a->b[ab]->r_l = i;
      if(d->b[dc]->left->v[i] == d->b[dc]->rght) d->b[dc]->l_r = i;
      if(d->b[dc]->rght->v[i] == d->b[dc]->left) d->b[dc]->r_l = i;
    }
  

  a->b[ab]->l_v1 = a->b[ab]->l_v2 = 
  a->b[ab]->r_v1 = a->b[ab]->r_v2 = 
  d->b[dc]->l_v1 = d->b[dc]->l_v2 = 
  d->b[dc]->r_v1 = d->b[dc]->r_v2 = -1;
  
  For(i,3)
    {
      if(i != a->b[ab]->l_r)
	{
	  if(a->b[ab]->l_v1 < 0) a->b[ab]->l_v1 = i;
	  else a->b[ab]->l_v2 = i;
	}
      if(i != a->b[ab]->r_l)
	{
	  if(a->b[ab]->r_v1 < 0) a->b[ab]->r_v1 = i;
	  else a->b[ab]->r_v2 = i;
	}
      if(i != d->b[dc]->l_r)
	{
	  if(d->b[dc]->l_v1 < 0) d->b[dc]->l_v1 = i;
	  else d->b[dc]->l_v2 = i;
	}
      if(i != d->b[dc]->r_l)
	{
	  if(d->b[dc]->r_v1 < 0) d->b[dc]->r_v1 = i;
	  else d->b[dc]->r_v2 = i;
	}
    }
}

/*********************************************************/

void NNI(arbre *tree, edge *b_fcus, int do_swap)
{
  int l_r, r_l, l_v1, l_v2, r_v3, r_v4, i, pos;
  node *v1,*v2,*v3,*v4;
  double lk1, lk2, lk3;
  double bl_init;
  double l1,l2,l3, theta1[tree->mod->n_pos], theta2[tree->mod->n_pos], theta3[tree->mod->n_pos];
  double l_infa, l_infb, l_max;
  double lk_infa, lk_infb, lk_max, lk_init;
  int starting_Opt_Bl;
  // char *arbre;
  // printf("IN NNI\n");
  bl_init           = b_fcus->l;
  lk_init           = tree->tot_loglk;

  starting_Opt_Bl = tree->mod->s_opt->opt_bl;

  b_fcus->best_conf = 1;
  b_fcus->diff_lk   = .0;

  lk1 = lk2 = lk3   = UNLIKELY;
  v1 = v2 = v3 = v4 = NULL;
  
  l_r = r_l = l_v1 = l_v2 = r_v3 = r_v4 = -1;

  // The root of the subtree shall always be in v1 since it doesn't move : 
  // we can then keep the direction in the tree.
  // cases to distinguish : 
  // ancestor in l_v1
  // ancestor in l_v2 (ancestor is always at the left).
  l_r               = b_fcus->l_r;
  r_l               = b_fcus->r_l;
 
  // If the ancestor is in l_v1.


  // No NNI if the edge under focus is linked to the root : we only optimize its length
  if(b_fcus->left==tree->noeud[tree->n_otu]) 
    {//printf("IN NNI2\n");
      b_fcus->ql[0] = b_fcus->l;
      b_fcus->ql[1] = b_fcus->l;
      b_fcus->ql[2] = b_fcus->l;
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	b_fcus->rght->qtheta[pos][0] = b_fcus->rght->theta[pos];
	b_fcus->rght->qtheta[pos][1] = b_fcus->rght->theta[pos];
	b_fcus->rght->qtheta[pos][2] = b_fcus->rght->theta[pos];
      }
      b_fcus->diff_lk = 1000000.;

    
      tree->mod->s_opt->opt_bl  = 0;
      
      lk1 = Update_Lk_At_Given_Edge(b_fcus,tree);
      tree->mod->s_opt->opt_bl = starting_Opt_Bl;

      l_infa = BL_MIN;
      b_fcus->l_old=b_fcus->l;
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	b_fcus->rght->theta_old[pos]=b_fcus->rght->theta[pos];
      }
      l_max  = b_fcus->l;
      l_infb = 4.*b_fcus->l;
      if(b_fcus->l <= BL_MIN) {l_infa = -BL_START; l_max = .0; l_infb = BL_START;}
      lk_infa = lk_max = lk_infb = .0;
      //printf("FROM NNI2\n");
      //lk1=Br_Len_NR(tree,tree->data, b_fcus, 10000,b_fcus->l);
      //printf("\nFirst call\n");
      
      //If the number of equilibrium frequencies is limited :
        if (tree->mod->EqFreq) { 
	lk1=Try_Thetas(tree, b_fcus, 1); 
      } else {
	lk1=Br_Len_Theta_NR(tree,tree->data, b_fcus, 10000,b_fcus->l,b_fcus->rght->theta);
	}


/*      Br_Len_Brak(&l_infa, &l_max, &l_infb,
		  &lk_infa,&lk_max,&lk_infb,
		  b_fcus,tree);
      lk1 = Br_Len_Brent(l_infa,l_max,l_infb,
			 1.e-6,
			 &(b_fcus->l),
			 b_fcus,tree,1000);
*/  
      return;
    }
  //printf("IN NNI3\n");
  for (pos=0;pos<tree->mod->n_pos;pos++) {
	b_fcus->rght->theta_old[pos]=b_fcus->rght->theta[pos];
      }
  v1 = b_fcus->left->b[0]->left;
  if (b_fcus->left->b[1]==b_fcus) {v2 = b_fcus->left->b[2]->rght;} else {v2 = b_fcus->left->b[1]->rght;}
  v3 = b_fcus->rght->v[b_fcus->r_v1];
  v4 = b_fcus->rght->v[b_fcus->r_v2];
    


  l1 = l2 = l3 = -1.;
  /* for (pos=0;pos<tree->mod->n_pos;pos++) {
    theta1[pos] = theta2[pos] = theta3[pos] = -1.;
    }*/

  // first alternate topology
  /***********/
  Swap(v2,b_fcus->left,b_fcus->rght,v3,tree);
  
  

  
  tree->mod->s_opt->opt_bl = 0;
  tree->both_sides    = 1;
  lk2 = Update_Lk_At_Given_Edge(b_fcus,tree);
  tree->mod->s_opt->opt_bl = starting_Opt_Bl;
  
  l_infa = BL_MIN;
  l_max  = b_fcus->l;
  l_infb = 4.*b_fcus->l;
  if(b_fcus->l <= BL_MIN) {l_infa = -BL_START; l_max = .0; l_infb = BL_START;}
  lk_infa = lk_max = lk_infb = .0;
  //lk2=Br_Len_NR(tree,tree->data, b_fcus, 10000,b_fcus->l);
  //printf("Second call b_fcus->l : %f, b_fcus->rght->theta : %f, b_fcus->rght->theta_old : %f\n", b_fcus->l,b_fcus->rght->theta[0],b_fcus->rght->theta_old[0] );

  //If the number of equilibrium frequencies is limited :
  if (tree->mod->EqFreq) { 
    lk2=Try_Thetas(tree, b_fcus, 1); }
  else {
    lk2=Br_Len_Theta_NR(tree,tree->data, b_fcus, 10000,b_fcus->l,b_fcus->rght->theta );
  }	

  /*Br_Len_Brak(&l_infa, &l_max, &l_infb,
	       &lk_infa,&lk_max,&lk_infb,
	       b_fcus,tree);
  lk2 = Br_Len_Brent(l_infa,l_max,l_infb,
		     1.e-6,
		     &(b_fcus->l),
		     b_fcus,tree,1000);
  */
  l2  = b_fcus->l;
  for (pos=0;pos<tree->mod->n_pos;pos++) {
    theta2[pos]=b_fcus->rght->theta[pos];
  }
  //arbre = Write_Tree(tree);
  // printf("TREE : %s\n",arbre);
  //printf("Alternate LKH2 : %f\n",  lk2);
  Swap(v3,b_fcus->left,b_fcus->rght,v2,tree); // back to origin position
  /***********/


  // second alternate topology
  /***********/
  /* for (pos=0;pos<tree->mod->n_pos;pos++) {
	b_fcus->rght->theta[pos]=b_fcus->rght->theta_old[pos];
	}*/

  Swap(v2,b_fcus->left,b_fcus->rght,v4,tree);
  tree->mod->s_opt->opt_bl = 0;
  tree->both_sides = 1;
  lk3 = Update_Lk_At_Given_Edge(b_fcus,tree);
  tree->mod->s_opt->opt_bl = starting_Opt_Bl;

  l_infa = BL_MIN;
  l_max  = b_fcus->l;
  l_infb = 4.*b_fcus->l;
  if(b_fcus->l <= BL_MIN) {l_infa = -BL_START; l_max = .0; l_infb = BL_START;}
  lk_infa = lk_max = lk_infb = .0;
  //lk3=Br_Len_NR(tree,tree->data, b_fcus, 10000,b_fcus->l);
  //printf("Third call b_fcus->l : %f, b_fcus->rght->theta : %f, b_fcus->rght->theta_old : %f\n", b_fcus->l,b_fcus->rght->theta[0],b_fcus->rght->theta_old[0] );

  //If the number of equilibrium frequencies is limited :
  if (tree->mod->EqFreq) { 
    lk3=Try_Thetas(tree, b_fcus, 1); }
  else {
    lk3=Br_Len_Theta_NR(tree,tree->data, b_fcus, 10000,b_fcus->l, b_fcus->rght->theta);
  }
 
  /*Br_Len_Brak(&l_infa, &l_max, &l_infb,
    &lk_infa,&lk_max,&lk_infb,
	       b_fcus,tree);
  lk3 = Br_Len_Brent(l_infa,l_max,l_infb,
		      1.e-6,
		      &(b_fcus->l),
		      b_fcus,tree,1000);
  */
  l3  = b_fcus->l;
  for (pos=0;pos<tree->mod->n_pos;pos++) {
    theta3[pos]=b_fcus->rght->theta[pos];
  }
  //arbre = Write_Tree(tree);
  //printf("TREE : %s\n",arbre);
  //printf("Alternate LKH3 : %f\n",  lk3);
  Swap(v4,b_fcus->left,b_fcus->rght,v2,tree); // back to origin position
  /***********/
   


  /***********/
  /*for (pos=0;pos<tree->mod->n_pos;pos++) {
	b_fcus->rght->theta[pos]=b_fcus->rght->theta_old[pos];
	}*/
   b_fcus->l = bl_init;
   
   tree->mod->s_opt->opt_bl  = 0;
   tree->both_sides = 1;

   lk1 = Update_Lk_At_Given_Edge(b_fcus,tree);
   tree->mod->s_opt->opt_bl = starting_Opt_Bl;

   l_infa = BL_MIN;
   l_max  = b_fcus->l;
   l_infb = 4.*b_fcus->l;
   if(b_fcus->l <= BL_MIN) {l_infa = -BL_START; l_max = .0; l_infb = BL_START;}
   lk_infa = lk_max = lk_infb = .0;
   //lk1=Br_Len_NR(tree,tree->data, b_fcus, 10000,b_fcus->l);
   //printf("Fourth call b_fcus->l : %f, b_fcus->rght->theta : %f, b_fcus->rght->theta_old : %f\n", b_fcus->l,b_fcus->rght->theta[0],b_fcus->rght->theta_old[0] );

   //If the number of equilibrium frequencies is limited :
   if (tree->mod->EqFreq) { 
     lk1=Try_Thetas(tree, b_fcus, 1); 
   } else {
     lk1=Br_Len_Theta_NR(tree,tree->data, b_fcus, 10000,b_fcus->l,b_fcus->rght->theta);
   }
   //	printf("lk1 again :%f\n", lk1);
   
   
   /*  Br_Len_Brak(&l_infa, &l_max, &l_infb,
	       &lk_infa,&lk_max,&lk_infb,
	       b_fcus,tree);
   lk1 = Br_Len_Brent(l_infa,l_max,l_infb,
		       1.e-6,
		       &(b_fcus->l),
		       b_fcus,tree,1000);
   */
   l1  = b_fcus->l;
   for (pos=0;pos<tree->mod->n_pos;pos++) {
     theta1[pos]=b_fcus->rght->theta[pos];
    
   }
   // arbre = Write_Tree(tree);
   //printf("TREE : %s\n",arbre);
   /* printf("Origin LKH : %f\n",  lk1);
   printf("LKH2 : %f\n",  lk1);
   printf("LKH3 : %f\n",  lk1);*/

  /***********/

  

   b_fcus->ql[0] = l1;
   b_fcus->ql[1] = l2;
   b_fcus->ql[2] = l3;
   for (pos=0;pos<tree->mod->n_pos;pos++) {
     b_fcus->rght->qtheta[pos][0] = theta1[pos];
     b_fcus->rght->qtheta[pos][1] = theta2[pos];
     b_fcus->rght->qtheta[pos][2] = theta3[pos];
   }
   b_fcus->diff_lk = lk1 - MAX(lk2,lk3);
   
   //printf("lk1 : %f, lk2 : %f, lk3 : %f, diff_lk :%f\n", lk1, lk2, lk3, b_fcus->diff_lk);
   //  printf("bnum :%d, diff_lk :%f\n",b_fcus->num, b_fcus->diff_lk);
   
   if(lk2 > lk3) b_fcus->best_conf = 2;
   else          b_fcus->best_conf = 3;


   if((do_swap) && ((lk2 > lk1+MDBL_MIN) || (lk3 > lk1+MDBL_MIN)))
     {
      tree->n_swap++;
      fflush(stdout);
      if(lk2 > lk3)
	 {
	   tree->best_loglk = lk2;
	   Swap(v2,b_fcus->left,b_fcus->rght,v3,tree);
	   b_fcus->l = l2;
	   for (pos=0;pos<tree->mod->n_pos;pos++) {
	     b_fcus->rght->theta[pos]=theta2[pos];
	  
	   }
	   tree->both_sides = 1;
	   Lk(tree,tree->data);
	 }
       else
	 {
	   tree->best_loglk = lk3;
	   Swap(v2,b_fcus->left,b_fcus->rght,v4,tree);
	   b_fcus->l = l3;
	   for (pos=0;pos<tree->mod->n_pos;pos++) {
	   b_fcus->rght->theta[pos]=theta3[pos];
	   }
	   tree->both_sides = 1;
	   Lk(tree,tree->data);
	 }
     }
   else 
     {
       b_fcus->l = bl_init;
       For(pos,tree->mod->n_pos){
	 For(i,tree->mod->n_catg)
	   {
	     Computes_Substitution_Probabilities(b_fcus,tree, i, pos);
	   }
       }
       tree->tot_loglk = lk_init;
     }
}

/*********************************************************/
double Update_Lk_At_Given_Edge(edge *b_fcus, arbre *tree)
{


  /* Updating partial likelihood after branch swapping */
   Update_P_Lk(tree,b_fcus,b_fcus->left);
  Update_P_Lk(tree,b_fcus,b_fcus->rght);
  tree->tot_loglk = Lk_At_Given_Edge(tree,b_fcus);
  return tree->tot_loglk;
}

/*********************************************************/

int Sort_Edges_Diff_Lk(arbre *tree, edge **sorted_edges, int n_elem)
{
  int i,j;
  edge *buff;

  For(i,n_elem-1)
    {
      for(j=i+1;j<n_elem;j++)
	{
	  if(sorted_edges[j]->diff_lk < sorted_edges[i]->diff_lk)
	    {
	      buff = sorted_edges[j];
	      sorted_edges[j] = sorted_edges[i];
	      sorted_edges[i] = buff;
	    }
	}
    }
  return 1;
}
/*********************************************************/
void Update_BrLen_Invar(arbre *tree)
{
  /* int i;
     For(i,2*tree->n_otu-2) tree->t_edges[i]->l*=(1.0-tree->mod->pinvar);*/
}

/*********************************************************/
char *Write_Tree(arbre *tree)
{

  char *s;
  int i;

  s=(char *)mCalloc(T_MAX_LINE,sizeof(char));
  
  s[0]='(';
  
  i = 0;
  while((!tree->noeud[tree->n_otu+i]->v[0]) ||
	(!tree->noeud[tree->n_otu+i]->v[1]) /*||
	(!tree->noeud[tree->n_otu+i]->v[2])*/) i++;
  
  R_wtree(tree->noeud[tree->n_otu+i],tree->noeud[tree->n_otu+i]->v[0],s,tree);
  R_wtree(tree->noeud[tree->n_otu+i],tree->noeud[tree->n_otu+i]->v[1],s,tree);
 

  s[(int)strlen(s)-1]=')';
  s[(int)strlen(s)]=';';
  
  
  return s;
}

/*********************************************************/

void R_wtree(node *pere, node *fils, char *s_tree,arbre *tree)
{
  int i,p, pos;
  char s[10];
  double tl, gc;
  p = -1;
  gc=0.;
  if (fils->b[0]->num==1) {
    tl=true_length(fils,tree->mod->kappa,tree);
    //    printf("fils->b[0]->l : %f, tl : %f\n",fils->b[0]->l, tl); 
  }
  if(fils->tax) 
    {
      strcat(s_tree,fils->name);
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	gc=fils->pi[pos][1]+fils->pi[pos][2];
      }
      gc/=tree->mod->n_pos;
      sprintf(s, " %.2f", 100*gc);
      strcat(s_tree,s);
      if((fils->b[0]) && (fils->b[0]->l != -1))
	{
	  
	  strcat(s_tree,":");
	  tl=true_length(fils,tree->mod->kappa,tree);
	  sprintf(s_tree+(int)strlen(s_tree),"%.15f",tl);
	  fflush(stdout);
	}
      sprintf(s_tree+(int)strlen(s_tree),",");
   }

  else
    {
      s_tree[(int)strlen(s_tree)]='(';
      For(i,3)
	{
	  if(fils->v[i] != pere)
	    R_wtree(fils,fils->v[i],s_tree,tree);
	  else p=i;
	}	
      s_tree[(int)strlen(s_tree)-1]=')';
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	gc=fils->pi[pos][1]+fils->pi[pos][2];
      }
      gc/=tree->mod->n_pos;
      sprintf(s, "%.2f", 100*gc);
      strcat(s_tree,s);
      if(fils->b[0]->l != -1)
	{
	  strcat(s_tree,":");
	  tl=true_length(fils,tree->mod->kappa,tree);
	  sprintf(s_tree+(int)strlen(s_tree),"%.15f,",tl);
	  fflush(stdout);
	}
    }
}

/*********************************************************/
void Print_Fp_Out(FILE *fp_out, time_t t_beg, time_t t_end, arbre *tree, option *input, int n_data_set)
{
  char *s;
  div_t hour,min;
  int i;
  
  fprintf(fp_out,". Sequence file : [%s]\n\n", input->seqfile);  
  


 

  /*was after Sequence file ; moved here FLT*/
  s = (char *)mCalloc(T_MAX_LINE,sizeof(char));
  fprintf(fp_out,". Initial tree : [%s]\n\n",
	  (strcat(strcat(strcat(s,"user tree ("),input->inputtreefile),")")));  
  Free(s);
  
  fprintf(fp_out,". Number of taxons : %d\n\n",tree->n_otu);/*added FLT*/

  fprintf(fp_out,"\n");

  fprintf(fp_out,". Likelihood : loglk = %.5f\n\n",tree->tot_loglk);/*was last ; moved here FLT*/

  fprintf(fp_out,". Discrete gamma model : %s\n",
	  (tree->mod->n_catg>1)?("Yes"):("No\n"));
  if(tree->mod->n_catg > 1)
    {
      fprintf(fp_out,"  - Number of categories : %d\n",tree->mod->n_catg);
      fprintf(fp_out,"  - Gamma shape parameter : %.3f\n\n",tree->mod->alpha);
    }
  
  if (tree->mod->n_pos==1) {
    fprintf(fp_out,"Transition/transversion ratio : %f\n", tree->mod->kappa[0]);
  }
  else {
    for (i=0;i<tree->mod->n_pos;i++) {
      fprintf(fp_out,"Sites %d : Transition/transversion ratio : %f\n", i, tree->mod->kappa[i]);
    }
  }
  if (tree->mod->EqFreq==1) {
    fprintf(fp_out,"Limited set of %d equilibrium frequencies\n", tree->mod->NumEqFreq);
    
  }else {
    fprintf(fp_out,"Each branch has its own optimized equilibrium frequency\n");
  }


  /*  if(tree->mod->invar)
    fprintf(fp_out,". Proportion of invariant : %.3f\n\n",tree->mod->pinvar);
  */
  /*was before Discrete gamma model ; moved here FLT*/
  
  



  
  

  hour = div(t_end-t_beg,3600);
  min  = div(t_end-t_beg,60  );
  
  min.quot -= hour.quot*60;
  
  fprintf(fp_out,". Time used %dh%dm%ds\n", hour.quot,min.quot,(int)(t_end-t_beg)%60);
  if(t_end-t_beg > 60)
    fprintf(fp_out,". -> %d seconds\n",(int)(t_end-t_beg));
  
  fprintf(fp_out,"\n%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n");

  fflush(fp_out);

}

/*********************************************************/
void Print_Fp_Out_Lines(FILE *fp_out, time_t t_beg, time_t t_end, arbre *tree, option *input, int n_data_set)
{
  char *s;
  /*div_t hour,min;*/

  if (n_data_set==1) {
    
    fprintf(fp_out,". Sequence file : [%s]\n\n", input->seqfile);  

  

    s = (char *)mCalloc(T_MAX_LINE,sizeof(char));
    fprintf(fp_out,". Initial tree : [%s]\n\n",
	    (strcat(strcat(strcat(s,"user tree ("),input->inputtreefile),")")));
    Free(s);

    fprintf(fp_out,"\n");

    /*headline 1*/
    fprintf(fp_out, ". Data\t");

    fprintf(fp_out,"Nb of \t");

    fprintf(fp_out,"Likelihood\t");

    fprintf(fp_out, "Discrete   \t");

    if(tree->mod->n_catg > 1)
      fprintf(fp_out, "Number of \tGamma shape\t");

    fprintf(fp_out,"Proportion of\t");

  
    fprintf(fp_out,"Nucleotides frequencies               \t");

  

    fprintf(fp_out, "\n");


    /*headline 2*/
    fprintf(fp_out, "  set\t");

    fprintf(fp_out,"taxons\t");

    fprintf(fp_out,"loglk     \t");

    fprintf(fp_out, "gamma model\t");

    if(tree->mod->n_catg > 1)
      fprintf(fp_out, "categories\tparameter  \t");

    fprintf(fp_out,"invariant    \t");

 

    fprintf(fp_out,"f(A)      f(C)      f(G)      f(T)    \t");




    fprintf(fp_out, "\n");


    /*headline 3*/
    

    fprintf(fp_out, "\n");
  }


  /*line items*/

  fprintf(fp_out,"  #%d\t",n_data_set);  

  fprintf(fp_out,"%d   \t",tree->n_otu);

  fprintf(fp_out,"%.5f\t",tree->tot_loglk);

  fprintf(fp_out,"%s        \t",
	  (tree->mod->n_catg>1)?("Yes"):("No "));
  if(tree->mod->n_catg > 1)
    {
      fprintf(fp_out,"%d        \t",tree->mod->n_catg);
      fprintf(fp_out,"%.3f    \t",tree->mod->alpha);
    }
  

  




  
  fprintf(fp_out, "\n\n");


  fflush(fp_out);

}

/*********************************************************/
void Update_Pis (arbre *tree) {
  Update_Pis_Subtree(tree->noeud[tree->n_otu]->b[0], tree);
  Update_Pis_Subtree(tree->noeud[tree->n_otu]->b[1], tree);
  tree->mod->omega_changed=0;
}


/*********************************************************/

void Update_Pis_Subtree (edge *b, arbre *tree) {

 if(b->rght->tax) {Compute_Pis(b, tree);return;}
  else {
    Compute_Pis(b, tree);
    Update_Pis_Subtree(b->rght->b[1], tree);
    Update_Pis_Subtree(b->rght->b[2], tree);
    }
}


/*********************************************************/
/*For the moment we do not distinguish between positions in the sequence*/
void Compute_Pis (edge *b, arbre *tree) {
  double pi[4];
  int i, j, catg, pos;

  //if (b->rght->tax) {} else { //If we want the real G+C proportions at leaves
  for (pos=0;pos<tree->mod->n_pos;pos++) {
    For(i,tree->mod->ns) { 
      pi[i]=0.0;
      For(j,tree->mod->ns) {
	For(catg, tree->mod->n_catg){
	  pi[i] += tree->mod->r_proba[catg] *
	    b->left->pi[pos][j] *
	    b->Pij_rr[pos][catg][j][i];
	}
      }
    }
    For(i,tree->mod->ns) {
      b->rght->pi[pos][i]=pi[i];
    }
  }
   
}
/*********************************************************/

char *Write_Tree_Equilibrium(arbre *tree)
{

  char *s;
  int i;

  s=(char *)mCalloc(T_MAX_LINE,sizeof(char));
  
  s[0]='(';
  
  i = 0;
  while((!tree->noeud[tree->n_otu+i]->v[0]) ||
	(!tree->noeud[tree->n_otu+i]->v[1]) ) i++;
  
  R_wtree_Equilibrium(tree->noeud[tree->n_otu+i],tree->noeud[tree->n_otu+i]->v[0],s, tree);
  R_wtree_Equilibrium(tree->noeud[tree->n_otu+i],tree->noeud[tree->n_otu+i]->v[1],s, tree);
 

  s[(int)strlen(s)-1]=')';
  s[(int)strlen(s)]=';';
  
  
  return s;
}

/*********************************************************/

void R_wtree_Equilibrium(node *pere, node *fils, char *s_tree, arbre *tree)
{
  int i,p, pos;
  char s[10];
  double tl, theta;
  p = -1;
  if(fils->tax) 
    {

      strcat(s_tree,fils->name);
      theta=0.;
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	theta+=fils->theta[pos];
      }
      sprintf(s, " %.2f", 100*theta/tree->mod->n_pos);
    
      strcat(s_tree,s);
      if((fils->b[0]) && (fils->b[0]->l != -1))
	{
	  strcat(s_tree,":");
	  tl=true_length(fils,tree->mod->kappa,tree);
	  sprintf(s_tree+(int)strlen(s_tree),"%.15f",tl);
	  fflush(stdout);
	}
      sprintf(s_tree+(int)strlen(s_tree),",");
   }

  else
    {
      s_tree[(int)strlen(s_tree)]='(';
      For(i,3)
	{
	  if(fils->v[i] != pere)
	    R_wtree_Equilibrium(fils,fils->v[i],s_tree, tree);
	  else p=i;
	}	
      s_tree[(int)strlen(s_tree)-1]=')';
      theta=0.;
      for (pos=0;pos<tree->mod->n_pos;pos++) {
	theta+=fils->theta[pos];
      }
      sprintf(s, "%.2f", 100*theta/tree->mod->n_pos);
      strcat(s_tree,s);
      if(fils->b[0]->l != -1)
	{
	  strcat(s_tree,":");
	  tl=true_length(fils,tree->mod->kappa,tree);
	  sprintf(s_tree+(int)strlen(s_tree),"%.15f,",tl);
	  fflush(stdout);
	}
    }
}

/*********************************************************/
/*Prints a tree in text format displaying node numbers*/

void Print_Tree_Text (arbre *tree, FILE *fp_seqs_out)
{
  int i, d, j, h, nameslength, sum, site, k, X;
  char  *map;
  char *temp;
  char *temp2;
  char **seqs;
  char num[20];
  //  FILE *fp_seqs_out;
  X=0;
  
  sum=1;
    nameslength=0;
  //Recursive calls
  Count_Nodes(tree->noeud[tree->n_otu]->b[0],tree);
  Count_Nodes(tree->noeud[tree->n_otu]->b[1],tree);
  tree->noeud[tree->n_otu]->numNodes=tree->noeud[tree->n_otu]->b[0]->rght->numNodesNormalized+tree->noeud[tree->n_otu]->b[1]->rght->numNodesNormalized;

 i=1;
 while (i<tree->noeud[tree->n_otu]->numNodes) {
   i=i*2;
 }
 tree->noeud[tree->n_otu]->numNodesNormalized=i;



 
  if ( tree->noeud[tree->n_otu]->b[0]->rght->numNodesNormalized >= tree->noeud[tree->n_otu]->b[1]->rght->numNodesNormalized) {
    tree->noeud[tree->n_otu]->MaxSubtree=tree->noeud[tree->n_otu]->b[0]->rght->numNodesNormalized;
    tree->noeud[tree->n_otu]->BigSubtree=tree->noeud[tree->n_otu]->b[0]->rght;
    tree->noeud[tree->n_otu]->SmallSubtree=tree->noeud[tree->n_otu]->b[1]->rght;
  }
  else {
    tree->noeud[tree->n_otu]->MaxSubtree=tree->noeud[tree->n_otu]->b[1]->rght->numNodesNormalized;
    tree->noeud[tree->n_otu]->BigSubtree=tree->noeud[tree->n_otu]->b[1]->rght;
    tree->noeud[tree->n_otu]->SmallSubtree=tree->noeud[tree->n_otu]->b[0]->rght;
  }

  i=1;

  if (tree->noeud[tree->n_otu]->MaxSubtree<2) {  }
  else {
    while (i<tree->noeud[tree->n_otu]->MaxSubtree) {
      i=i*2;
    }

    tree->noeud[tree->n_otu]->height=4*i;
    tree->noeud[tree->n_otu]->depth=5;
  }


  // There we consider the size of the numbers at the nodes and the size of the sequence names (we add 20 for this)
  // This yelds an intrinsic limit in the number of sequences : 10000.
  if (tree->noeud[tree->n_otu]->numNodes<10) {
    nameslength=20;
  }
  else if (tree->noeud[tree->n_otu]->numNodes<100) {
    nameslength=20+90*2;
  }
  else if (tree->noeud[tree->n_otu]->numNodes<1000) {
    nameslength=20+90*2+900*3;
  }
  else if (tree->noeud[tree->n_otu]->numNodes<10000) {
    nameslength=20+90*2+900*3+9000*4;
  }

  //Defines the max depth of the tree
  d=tree->noeud[tree->n_otu]->MaxSubtree*5+nameslength;
  //Total height is (tree->noeud[tree->n_otu]->height*2+1)
  h=(tree->noeud[tree->n_otu]->height)*2+1;


  temp2 =  (char*)mCalloc(T_MAX_LINE,sizeof(char));
if (d*h> MEM_LIM) {
  printf("Too many sequences to represent the tree in the sequence file.\nYou 'll still get the ancestral sequence though\n");  
}
else {

  map  = (char*)mCalloc(h*d,sizeof(char));

  temp =  (char*)mCalloc(d+1,sizeof(char));


  
  for (j=0;j<=(h*d);j++) {
    sprintf(&map[j],"%c",' ');
  }
  
 
  
  sprintf(temp2,"%d",tree->noeud[tree->n_otu]->num);
  for (j=0;j<=strlen(temp);j++) {
    map[tree->noeud[tree->n_otu]->height*d+5+j]=temp2[j];
  }
  for (j=1;j<5;j++) {
    map[tree->noeud[tree->n_otu]->height*d+5-j]='-';
  }
  
  tree->noeud[tree->n_otu]->BigSubtree->height=tree->noeud[tree->n_otu]->height-i;
  tree->noeud[tree->n_otu]->SmallSubtree->height=tree->noeud[tree->n_otu]->height+i;
  tree->noeud[tree->n_otu]->BigSubtree->depth=tree->noeud[tree->n_otu]->depth+5;
  tree->noeud[tree->n_otu]->SmallSubtree->depth=tree->noeud[tree->n_otu]->depth+5;
  
  for (j=tree->noeud[tree->n_otu]->BigSubtree->height;j<tree->noeud[tree->n_otu]->height;j++) {
    map[j*d+5]='|';
  }
  for (j=tree->noeud[tree->n_otu]->height+1;j<tree->noeud[tree->n_otu]->SmallSubtree->height;j++) {
    map[j*d+5]='|';
  }
  
  Fill_Table(tree->noeud[tree->n_otu]->BigSubtree,tree->noeud[tree->n_otu]->BigSubtree->height,10,map, h, d);
  Fill_Table(tree->noeud[tree->n_otu]->SmallSubtree,tree->noeud[tree->n_otu]->SmallSubtree->height,10,map, h, d);
  
 

  for (j=0;j<=h;j++) {
    for(i=j*d;i<(j+1)*d;i++) {
      temp[i-j*d]=map[i];
    }
    fprintf(fp_seqs_out,"%s\n",temp);
    
  }
  fprintf(fp_seqs_out,"\n\n\n\n");
}

  
  /* The text tree has been output.
     Now we begin to write down the sequences and the ancestral sequences.
  */



//Allocation

 seqs=(char**)mCalloc(2*tree->n_otu-1,sizeof(char*));
 for(i=0;i<2*tree->n_otu-1;i++) {
   seqs[i]=(char*)mCalloc(T_MAX_LINE,sizeof(char));//tree->data->init_len,sizeof(char));
 }

 //Filling in the sequences
 for(site=0;site<tree->data->init_len;site++) {
   for(i=0;i<tree->n_otu;i++) {
     if(tree->data->c_seq[i]->state[tree->data->whole_seq[site]]!='X') {
     seqs[i][site]=tree->data->c_seq[i]->state[tree->data->whole_seq[site]];
     }
     else {seqs[i][site]='-';}
   }
 }
 strcpy(temp2,">");

 
 

  // Printing sequences...
   for(i=0;i<tree->n_otu;i++) { 
     strcpy(temp2,">");
     strcat(temp2,tree->data->c_seq[i]->name);
     fprintf(fp_seqs_out,"%s\n",temp2);
     fprintf(fp_seqs_out,"%s\n\n",seqs[i]);
   }

   // Building ancestral sequences
   seqs=Estimate_Anc_Seqs(tree, seqs);
 
   for(i=tree->n_otu;i<tree->n_otu+1;i++) {
     strcpy(temp2,">");

     sprintf(num,"%d",tree->noeud[i]->num);
     strcat(temp2,num);
     fprintf(fp_seqs_out,"%s\n",temp2);
    
      for (j=0;j<strlen(seqs[tree->n_otu]);j++) {
	for(k=0;k<tree->n_otu;k++) {
	  X=0;
	  if (seqs[k][j]=='-') {
	    X=1;
	    break;
	  }
	}
	if(X) {fprintf(fp_seqs_out,"%c",'-');}
	else {fprintf(fp_seqs_out,"%c",seqs[i][j]);}fflush(NULL);
      }
      fprintf(fp_seqs_out,"\n\n");
      fflush(NULL);
   }
}

/*********************************************************/
/*Counts the number of underlying nodes : postfixe depth search*/
void Count_Nodes (edge *b, arbre *tree) {
  int i;
 
 if (b->rght->tax) {b->rght->numNodes=1;b->rght->numNodesNormalized=1;return;}
  else {
    Count_Nodes(b->rght->b[1], tree);
    Count_Nodes(b->rght->b[2], tree);
    
    b->rght->numNodes=b->rght->b[1]->rght->numNodesNormalized+b->rght->b[2]->rght->numNodesNormalized;
    
    i=1;
    while (i<b->rght->numNodes) {
      i=i*2;
    }
    b->rght->numNodesNormalized=i;
  


    if (b->rght->b[1]->rght->numNodesNormalized > b->rght->b[2]->rght->numNodesNormalized) {
      b->rght->MaxSubtree=b->rght->b[1]->rght->numNodesNormalized;
      b->rght->BigSubtree=b->rght->b[1]->rght;
      b->rght->SmallSubtree=b->rght->b[2]->rght;
    }
    else {
      b->rght->MaxSubtree=b->rght->b[2]->rght->numNodesNormalized;
      b->rght->BigSubtree=b->rght->b[2]->rght;
      b->rght->SmallSubtree=b->rght->b[1]->rght;
    }
  }
}

/*********************************************************/
/*Fills a character table so that we can then print out the tree in text format.
The table is a 1-D one, of height times the maximal depth*/

void Fill_Table (node *n, int h, int d, char *map, int sizeh, int sized) {
  int i,j;
  char temp[20];

  if (n->tax) {
    for (i=0;i<strlen(n->name);i++) {
       map[h*sized+d+i]=n->name[i];
     }
   
     for (i=1;i<=5;i++) {
    
       map[h*sized+d-i]='-';
     }
     return;
    
  }
  else {
    sprintf(temp,"%d",n->num);
    for (i=0;i<strlen(temp);i++) {
       map[h*sized+d+i]=temp[i];
     }
    for (i=1;i<=5;i++) {
      map[h*sized+d-i]='-';
    }

  }  

  

  i=1;
  if (n->MaxSubtree<2) {  
    n->BigSubtree->height=(n->height)-1;
    n->SmallSubtree->height=(n->height)+1;
  }
  else {
    while (i<n->MaxSubtree) {
      i=i*2;
    }
  
   


    n->BigSubtree->height=(n->height)-i;
    n->SmallSubtree->height=(n->height)+i;
  }
  n->BigSubtree->depth=d+5;
  n->SmallSubtree->depth=d+5;
 
  if (!n->tax) {
    Fill_Table(n->BigSubtree,n->BigSubtree->height,n->BigSubtree->depth,map, sizeh, sized);
    Fill_Table(n->SmallSubtree,n->SmallSubtree->height,n->SmallSubtree->depth,map, sizeh, sized);
  }
 for (j=n->BigSubtree->height;j<n->height;j++) {
    map[j*sized+n->depth]='|';
  }
  for (j=n->height+1;j<n->SmallSubtree->height;j++) {
    map[j*sized+n->depth]='|';
  }

}

/*********************************************************/

char**      Estimate_Anc_Seqs(arbre *tree, char **seqs) {
  int site, i, catg, k, l, sitefin, pos;
  double site_lk[tree->mod->ns];
  char alpha[4], temp[tree->n_otu*2-1];
  double max, **parp;
  /* A=0; C=1, G=2, T=3 */
  alpha[0]='A';
  alpha[1]='C';
  alpha[2]='G';
  alpha[3]='T';

  k=0;
  parp=(double**) mCalloc (tree->mod->n_pos, sizeof(double*));
  
  for(pos=0;pos<tree->mod->n_pos;pos++)
    {
      parp[pos]=(double*) mCalloc (tree->mod->ns, sizeof(double));
    }

  for(sitefin=0;sitefin<tree->data->init_len;sitefin++) {
    site=tree->data->whole_seq[sitefin];
 
    site_lk[0]=site_lk[1]=site_lk[2]=site_lk[3]=0;
      For(catg,tree->mod->n_catg)
	{
	  For(k,tree->mod->ns) //sort sum terms ? No global effect
	    {
	      For(l,tree->mod->ns)
		{
		  For(i,tree->mod->ns)
		    {
		      site_lk[k] += 
			tree->mod->r_proba[catg] * // Rate probability
			tree->noeud[tree->n_otu]->pi[tree->data->pos[site]][k] *  // Proportion of nucleotide k at root
			tree->noeud[tree->n_otu]->b[0]->p_lk_rght[tree->data->pos[site]][site][catg][l] *   // Inferior Partial Likelihood
			tree->noeud[tree->n_otu]->b[0]->Pij_rr[tree->data->pos[site]][catg][k][l] *         // Substitution probability
			tree->noeud[tree->n_otu]->b[1]->p_lk_rght[tree->data->pos[site]][site][catg][i] *   // Inferior Partial Likelihood
			tree->noeud[tree->n_otu]->b[1]->Pij_rr[tree->data->pos[site]][catg][k][i]    ;     // Substitution probability
					    
		    }
		}
	    }
	}
      max=UNLIKELY;
      for(pos=0;pos<tree->mod->n_pos;pos++)
	{
	  for(k=0;k<tree->mod->ns;k++) {
	    if (site_lk[k]>max) {
	      max=site_lk[k];
	      seqs[tree->noeud[tree->n_otu]->num][sitefin]=alpha[k];
	      parp[pos][k]=1.; // We use the most likely base at the root to estimate the node bases
	      for (i=0;i<k;i++) {
		parp[pos][i]=0.;
	      }
	      
	    }
	  }
	}
      Estimate_Seqs(tree,parp,temp); 
     
      for (k=tree->n_otu+1;k<tree->n_otu*2-1;k++)
	{
	  seqs[k][sitefin]=temp[k];
	  fflush(NULL);

	}
      for(pos=0;pos<tree->mod->n_pos;pos++)
	{
	  for(k=0;k<tree->mod->ns;k++) {
	    
	    parp[pos][k]=0.;
	    
	  }
	}

  }
  for(pos=0;pos<tree->mod->n_pos;pos++)
    {
      free (parp[pos]);
    }
  free (parp);
  
  return (seqs);

}





/*********************************************************/
/*TO CHECK : Might be not right
 */

void Estimate_Seqs (arbre *tree,double **parp, char *align) {

  Estimate_Seqs_Subtree(tree->noeud[tree->n_otu]->b[0], tree, parp, align);
  Estimate_Seqs_Subtree(tree->noeud[tree->n_otu]->b[1], tree, parp, align);
  
}


/*********************************************************/

void Estimate_Seqs_Subtree (edge *b, arbre *tree, double **parp, char *align) {

  double **p;
  int pos;

  if(b->rght->tax) {}
  else {
    p=(double**) mCalloc (tree->mod->n_pos, sizeof(double*));
    For(pos, tree->mod->n_pos) {
      p[pos]     = (double*)mCalloc(tree->mod->ns,sizeof(double));
    }
    p=Compute_Seqs(b, tree, parp, align, p);

    Estimate_Seqs_Subtree(b->rght->b[1], tree, p, align);
    Estimate_Seqs_Subtree(b->rght->b[2], tree, p, align);
    Free(p);
    fflush(NULL);
  }
}


/*********************************************************/

double** Compute_Seqs (edge *b, arbre *tree, double **parp, char *align, double **p) {
  double max, mean[tree->mod->ns];//, *p;
  int i, j, catg, pos;
  char alpha[4];
  alpha[0]='A';
  alpha[1]='C';
  alpha[2]='G';
  alpha[3]='T';



    For(i,tree->mod->ns) { 

      For(j,tree->mod->ns) {
	For(pos, tree->mod->n_pos)
	  { p[pos][i]=0.0;
	    For(catg, tree->mod->n_catg){
	      p[pos][i] += tree->mod->r_proba[catg] *
		parp[pos][j] *
		b->Pij_rr[pos][catg][j][i];
	    }
	  }
      }
    }

    max=0;

    
    For (pos, tree->mod->n_pos)
      {	
	For(i,tree->mod->ns) {
	  mean[i]+=p[pos][i];
	}
      }
    
	For(i,tree->mod->ns) {
	 mean[i]/=tree->mod->n_pos;
	  if (mean[i]>max)
	    {
	      max=mean[i];
	      
	      align[b->rght->num]=alpha[i];
	    }
	}
      
    
    return(p);
  
 
}


/*********************************************************/

double true_length(node *nd, double *kappa, arbre *tree){

  double mylg, teta, lat, lcg, glat, glcg, coeff;
  //  node v3;
  int i, pos;
  if(nd==tree->noeud[tree->n_otu]) {Exit("Root node has no ancestral branch\n");}
  lat =lcg=0.;
  coeff=0.;
  //  v3=nd->b[0];
  mylg=nd->b[0]->l;
  
  for (pos=0;pos<tree->mod->n_pos;pos++) {
    teta=nd->theta[pos];
    lat=(1+kappa[pos]*teta)/2.;
    lcg=(1+kappa[pos]*(1-teta))/2.;
  
  //  lat/=tree->mod->n_pos;
  //lcg/=tree->mod->n_pos;
  
    for(i=0;i<tree->mod->n_catg;i++){
      glat=tree->mod->rr[i]*lat;
      glcg=tree->mod->rr[i]*lcg;
      coeff+= nd->pi[pos][0]*glat + nd->pi[pos][1]*glcg + nd->pi[pos][2]*glcg + nd->pi[pos][3]*glat;
    }
  }
  coeff/=tree->mod->n_pos;

  return mylg*coeff/tree->mod->n_catg;
}

/*********************************************************/
/* This function is to be rebuilt if we want to reuse it */
void Adapt_Length(arbre *tree) {
  int i,j,pos;
  double lat, lcg, glat, glcg,coeff;
  lat =lcg=0.;
  for (i=0;i<2*tree->n_otu-2;i++) {
    coeff=0;
    for(pos=0;pos<tree->mod->n_pos;pos++) {
      lat=tree->mod->kappa[pos]*tree->t_edges[i]->rght->theta[pos];
      lcg=tree->mod->kappa[pos]*(1-tree->t_edges[i]->rght->theta[pos]);
      
      lat=(1+lat)/2;
      lcg=(1+lcg)/2;
      
      for(j=0;j<tree->mod->n_catg;j++){
	glat=tree->mod->rr[j]*lat;
	glcg=tree->mod->rr[j]*lcg;
	coeff+= tree->t_edges[i]->rght->pi[pos][0]*glat + tree->t_edges[i]->rght->pi[pos][1]*glcg + tree->t_edges[i]->rght->pi[pos][2]*glcg + tree->t_edges[i]->rght->pi[pos][3]*glat;
	
      }
    }
    coeff/=tree->mod->n_pos;
    tree->t_edges[i]->l=tree->t_edges[i]->l*tree->mod->n_catg/coeff;
  }
}
/*********************************************************/

/* gg95 */
/* Galtier & Gouy distance */

double gg95(seq* seq1, seq* seq2, allseq *data, double a){

  double k11, k12, k21, k22, t0, rt, *freq, Q, t1, t2, e;
  int i;


  freq=calloc(17, sizeof(double));
  if(!freq_obs(seq1, seq2, data->wght, freq, data)) return -1.;
 
  Q=freq[1]+freq[3]+freq[4]+freq[6]+freq[9]+freq[11]+freq[12]+freq[14];
  t1=0.;
  for(i=4;i<=11;i++) t1+=freq[i];
  t2=freq[1]+freq[2]+freq[5]+freq[6]+freq[9]+freq[10]+freq[13]+freq[14];
			
  if(2*Q>=1.) return -1.;
  rt=-0.5*log(1-Q*2);
	
  t0=(t1+t2)/2;

  e=1-exp(-rt*(a+1.)/2);
  k11=(0.5+a*t1*(1-t1))*rt ; 
  k12=(a/(a+1.)) * ((t0-t1)*(1-2*t1)) * e;	
  k21=(0.5+a*t2*(1-t2))*rt;
  k22=(a/(a+1.)) * ((t0-t2)*(1-2*t2)) * e;
 

  free(freq);

  return(k11+k12+k21+k22);

}



/*********************************************************/

/* freq_obs */
/* Write at address freq observed frequencies of 16 di-nucleotides XY */
/* (X= A,C,G,T  ,  Y=A,C,G,T) X and Y being homologous nucleotides of sequences */
/* seq1 and seq2. Alphabetic order is used : freq[0]=AA frequency, freq[1]=AC, */
/* ..., freq[15]=TT. */

int freq_obs(seq* seq1, seq* seq2, int *seqw, double* freq, allseq* data){

  int i, lgseq, lgseqvrai,tot;

  lgseq=data->crunch_len;
 
  tot=0;
  /* for (i=0;i<lgseq;i++){
    tot=tot+seqw[i];
  }
  */
  for(i=0;i<16;i++) freq[i]=0;
  lgseqvrai=lgseq;
  for(i=0;i<lgseq;i++){
    // printf("i : %d\n",i);
    switch(seq1->state[i]){
    case 'A':
      switch(seq2->state[i]){
      case 'A' :   freq[0]+=seqw[i]; break;
      case 'C' :   freq[1]+=seqw[i]; break;
      case 'G' :   freq[2]+=seqw[i]; break;
      case 'T' :   freq[3]+=seqw[i]; break;
      default : lgseqvrai --; break;
      }
      break;
    case 'C':
      switch(seq2->state[i]){
      case 'A' :   freq[4]+=seqw[i]; break;
      case 'C' :   freq[5]+=seqw[i]; break;
      case 'G' :   freq[6]+=seqw[i]; break;
      case 'T' :   freq[7]+=seqw[i]; break;
      default : lgseqvrai --; break;
      }
      break;
    case 'G':
      switch(seq2->state[i]){
      case 'A' :   freq[8]+=seqw[i]; break;
      case 'C' :   freq[9]+=seqw[i]; break;
      case 'G' :   freq[10]+=seqw[i];break;
      case 'T' :   freq[11]+=seqw[i];break;
      default : lgseqvrai --; break;
      }
      break;
    case 'T':
      switch(seq2->state[i]){
      case 'A' :   freq[12]+=seqw[i]; break;
      case 'C' :   freq[13]+=seqw[i]; break;
      case 'G' :   freq[14]+=seqw[i]; break;
      case 'T' :   freq[15]+=seqw[i]; break;
      default : lgseqvrai --; break;
      }
      break;
    default :
      lgseqvrai --;
    }
  }
 
  if(lgseqvrai!=0){
    for(i=0;i<16;i++)  {
      tot+=freq[i];
    }
    for(i=0;i<16;i++)  {
      freq[i]/=tot; 
    }
 
    return 1;
    }
      else return 0;
}




/*********************************************************/



/* alsurbet_kim */
/* Return the quotient between transition substitution rate (alpha) and */
/* transversion substitution rate (beta) according to Kimura 1980 */
/* 2-parameter model between sequence seq1 and seq2. */

double alsurbet_kim(seq* seq1, seq* seq2, double* freq, allseq* data){

  double alphakim, betakim, P, Q;

  if(!freq_obs(seq1, seq2, data->wght, freq, data)) return -1.;

  /*
  printf("seq1 : %s, ",seq1->name);
  printf("seq2 : %s\n",seq2->name);
  printf("freq[0] : %f, total : %f\n", freq[0],freq[0]+freq[1]+freq[2]+freq[3]+freq[4]+freq[5]+freq[6]+freq[7]+freq[8]+freq[9]+freq[10]+freq[11]+freq[12]+freq[13]+freq[14]+freq[15] );
  */

  Q=freq[1]+freq[3]+freq[4]+freq[6]+freq[9]+freq[11]+freq[12]+freq[14];
  P=1-Q-freq[0]-freq[5]-freq[10]-freq[15];

  if (2*Q>=1.) return -1.;
  if (2*P+Q>=1.) return -1.;
  /*  printf ("1.-2*Q : %.75f\n",(double)(1.-2*Q));
  printf ("1.-2*P-Q : %.75f\n",(double)(1.-2*P-Q));
  */

  betakim=-0.125*log(1.-2*Q);
  alphakim=-0.25*log(1.-2*P-Q)+0.125*log(1.-2*Q);
  /*  printf ("betakim : %.75f\n",(double)betakim);
  printf ("alphakim : %.75f\n",(double)alphakim);
  */
  if (betakim==0.) return -1.;
 	
  return(alphakim/betakim);
}

/*********************************************************/
/*Computes the likelihood derivatives in respect to omega*/

void Compute_Derivatives_Domega(arbre *tree, int pos) {
  int site;
  Lk(tree, tree->data);
  For(site,tree->n_pattern)
    {
      //We only compute the derivative if the position of the site is the one we consider : no time lost
      if (tree->data->pos[site]==pos) {
	tree->curr_site=site;
	Site_Derivatives_Domega(tree, pos);
      }
    }
 qksort(tree->tot_dloglk_domega_sorted, 0, tree->n_pattern-1);
 qksort(tree->tot_d2loglk_domega2_sorted, 0, tree->n_pattern-1);

 tree->tot_dloglk_domega = .0;
 tree->tot_d2loglk_domega2 = .0;

 For(site, tree->n_pattern){
   tree->tot_dloglk_domega += tree->tot_dloglk_domega_sorted[site];
   tree->tot_d2loglk_domega2 += tree->tot_d2loglk_domega2_sorted[site];
 }
}

/*********************************************************/

void Site_Derivatives_Domega(arbre *tree, int pos) {

  int  catg, k, l, i, site = tree->curr_site;
  double dlog_site_lk,dsite_lk, d2log_site_lk;//, realSiteLk;
  double dpi[4];
  dpi[1]=dpi[2]=1./2.;
  dpi[0]=dpi[3]=-1./2.;
 
  dlog_site_lk =  d2log_site_lk = dsite_lk = .0;      
  For(catg,tree->mod->n_catg)
    {     
      For(k,tree->mod->ns) //sort sum terms ? No global effect
	{
	  For(l,tree->mod->ns)
	    {
	      For(i,tree->mod->ns)
		{
		  dsite_lk += 
		    tree->mod->r_proba[catg] * // Rate probability
		    dpi[k] *  // derivative of a function of omega
		    tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][l] *   // Inferior Partial Likelihood
		    tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][l] *         // Substitution probability
		    tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][i] *   // Inferior Partial Likelihood
		    tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][i]    ;     // Substitution probability
		  
	
		}
	    }
	}
    }
  /* if (site==0) {
    printf("k : %d, truelike : %f,  like : %f, dlike : %f\n", 0,  tree->true_site_lk[site], tree->site_lk[site], dsite_lk);
    }*/

  //  realSiteLk=exp(tree->site_lk[site]);
  dlog_site_lk = dsite_lk/tree->true_site_lk[site];//dsite_lk/realSiteLk;//tree->site_lk[site]; 
 
 
  d2log_site_lk = - ((dsite_lk/tree->true_site_lk[site])*(dsite_lk/tree->true_site_lk[site]));//- ((dsite_lk/realSiteLk)*(dsite_lk/realSiteLk));
 
  tree->tot_dloglk_domega_sorted[site] = tree->data->wght[site]*dlog_site_lk;
  tree->tot_d2loglk_domega2_sorted[site] = tree->data->wght[site]*d2log_site_lk;

}


/*********************************************************/
/*Computes the likelihood derivatives in respect to theta
  on edge b_fcus */
void Compute_Derivatives_Dtheta(arbre *tree, edge *b_fcus, int pos){
 int site; 


 Site_Derivatives_Dtheta(tree,b_fcus, pos);
   

 qksort(tree->tot_dloglk_dtheta_sorted, 0, tree->n_pattern-1);
 qksort(tree->tot_d2loglk_dtheta2_sorted, 0, tree->n_pattern-1);

 tree->tot_dloglk_dtheta[b_fcus->num] = .0;
 tree->tot_d2loglk_dtheta2[b_fcus->num] = .0;

 For(site, tree->n_pattern){

     tree->tot_dloglk_dtheta[b_fcus->num] += tree->tot_dloglk_dtheta_sorted[site];
     tree->tot_d2loglk_dtheta2[b_fcus->num] += tree->tot_d2loglk_dtheta2_sorted[site];

 }

}

/*********************************************************/

void Site_Derivatives_Dtheta(arbre *tree, edge *b_fcus, int pos) {
  int site, catg, k, l, ns;
  double dsite_lk, dlog_site_lk, d2log_site_lk;//, realSiteLk;
  ns=tree->mod->ns;
 
    For(catg,tree->mod->n_catg)
      {
	Computes_Dtheta_Substitution_Probabilities (b_fcus, tree, catg, pos) ;
      }
 

  For(site,tree->n_pattern)
    {
      if(tree->data->wght[site])
      	{
	  dlog_site_lk = d2log_site_lk = dsite_lk = .0;
	  /* see equation (2) in phyml_tech_doc.pdf */
	  	
	    For(catg,tree->mod->n_catg)
	      {
		For(k,ns) /*sort sum terms ? No global effect*/
		  {
		    For(l,ns)
		      {
			dsite_lk += 
			  tree->mod->r_proba[catg] * // Rate probability
			  b_fcus->p_lk_left[pos][site][catg][k] * // Left Partial Likelihood
			  b_fcus->dPij_rr_theta[pos][catg][k][l] *  // Substitution probability
			  b_fcus->p_lk_rght[pos][site][catg][l];
		      }
		  }
	      }
	    /*   printf("THETA\n");
	    printf("dsite_lk : %f\n", dsite_lk);
	    printf("tree->site_lk[site] : %f\n", tree->site_lk[site]);*/
	    //	    realSiteLk=exp(tree->site_lk[site]);
	    dlog_site_lk += dsite_lk/tree->true_site_lk[site];//dsite_lk/realSiteLk;//tree->site_lk[site];
	    d2log_site_lk+= -(dsite_lk/tree->true_site_lk[site])*(dsite_lk/tree->true_site_lk[site]);//- (dsite_lk/realSiteLk)*(dsite_lk/realSiteLk);
	  //	  printf("Theta : sitelk : %f, dsitelk : %f, d2sitelk : %f, branche : %d, theta : %f\n",realSiteLk, dsite_lk, 0.0, b_fcus->num, b_fcus->rght->theta);  
	  tree->tot_dloglk_dtheta_sorted[site] = tree->data->wght[site]*dlog_site_lk;
	  tree->tot_d2loglk_dtheta2_sorted[site] = tree->data->wght[site]*d2log_site_lk; 
	}
    }
}
/*********************************************************/

/*Computes the substitution probabilities derivatives in respect to theta 
along the edge b whose ancestor is node left*/
void Computes_Dtheta_Substitution_Probabilities (edge *b, arbre *tree, int catg, int pos) {
  double theta, kappa,l, rr;
  l= b->l * tree->mod->l_coef[pos];
  theta = b->rght->theta[pos];
  kappa = tree->mod->kappa[pos];
  rr=tree->mod->rr[catg];

  /* Here we do take the gamma distribution coefficient in account, with the category integer*/
  /* A=0; C=1, G=2, T=3 */

  /*dPAA = dPTT  */
  b->dPij_rr_theta[pos][catg][0][0] = b->dPij_rr_theta[pos][catg][3][3] = -1./2.*(1.+exp(- rr * l))+exp(-(kappa+1.)/2.* rr *l);

  /*dPCC = dPGG  */
  b->dPij_rr_theta[pos][catg][1][1] = b->dPij_rr_theta[pos][catg][2][2] = 1./2.*(1.+exp(- rr *l))- exp(-(kappa+1.)/2.* rr *l);

  /*dPGA = dPCT  */
  b->dPij_rr_theta[pos][catg][2][0] = b->dPij_rr_theta[pos][catg][1][3] = -1./2.*(1.+exp(- rr * l))+ exp(-(kappa+1.)/2.* rr * l);

  /*dPAG = dPTC  */
  b->dPij_rr_theta[pos][catg][0][2] = b->dPij_rr_theta[pos][catg][3][1] = 1./2.*(1.+exp(- rr *l))- exp(-(kappa+1.)/2.* rr *l);

  /*dPAT = dPTA = dPGT = dPCA  */
  b->dPij_rr_theta[pos][catg][0][3] = b->dPij_rr_theta[pos][catg][3][0] = b->dPij_rr_theta[pos][catg][2][3] = b->dPij_rr_theta[pos][catg][1][0] = -1./2.*(1.-exp(- rr * l));

  /*dPCG = dPGC = dPTG = dPAC  */
  b->dPij_rr_theta[pos][catg][1][2] = b->dPij_rr_theta[pos][catg][2][1] = b->dPij_rr_theta[pos][catg][3][2] = b->dPij_rr_theta[pos][catg][0][1] = 1./2.*(1.-exp(- rr *l));

}
/*********************************************************/

void NJ_Tree_Fit (node *pere, node *fils, arbre *tree, double **dist, int *n_Non_Agglo) {

  int i, dir1, dir2, dir3;
  double S_a, S_b; // Notations according to Guindon's thesis
  S_a=S_b=0.0;
 
   if(fils->tax) {
      return ;
   }
   else {
     dir1=dir2=-1;
     For(i,3)
       {
	 if(fils->v[i] != pere)
	   {
	     if(dir1 < 0) {
		dir1 = i; // find the two branches to study
		NJ_Tree_Fit(fils,fils->v[i],tree, dist, n_Non_Agglo); 
	      }
	      else {
		dir2 = i;
		NJ_Tree_Fit(fils,fils->v[i],tree, dist, n_Non_Agglo); 
	      }
	    }
	  else dir3 = i; // v[dir3] is the father node
       }

     //Computation of S_a and S_b
     for (i=0;i<2*tree->n_otu-1;i++) {
       S_a += dist[fils->v[dir1]->num][i];
       S_b += dist[fils->v[dir2]->num][i];
     }

     //Computation of the distance between the son nodes that have been agglomerated and their father
     fils->b[dir1]->l=1./2.*(dist[fils->v[dir1]->num][fils->v[dir2]->num] + (S_a-S_b)/(double)(*n_Non_Agglo -2));
     fils->b[dir2]->l=1./2.*(dist[fils->v[dir2]->num][fils->v[dir1]->num] + (S_b-S_a)/(double)(*n_Non_Agglo -2));

   

     //Computation of the distances between the new father nodes and the other nodes
     for (i=0;i<2*tree->n_otu-1;i++) {
       dist[pere->num][i]=dist[i][pere->num]=1./2.*(dist[fils->v[dir1]->num][i]-fils->b[dir1]->l+dist[fils->v[dir2]->num][i]-fils->b[dir2]->l);
     }
  //"removing" of the nodes that have been agglomerated : their distances value 0
     for (i=0;i<2*tree->n_otu-1;i++) {
       dist[fils->v[dir1]->num][i]=0.;
       dist[fils->v[dir2]->num][i]=0.;
       dist[i][fils->v[dir1]->num]=0.;
       dist[i][fils->v[dir2]->num]=0.;
     }
     n_Non_Agglo--;
     return;
   }
}


/*********************************************************/
/*Computes the derivatives of substitution probabilities along the edge b*/
/*in respect to kappa whose ancestor is node left*/
void Computes_Derivative_Substitution_Probabilities_Kappa (edge *b, arbre *tree, int catg, int pos) {
  double theta, kappa,l, rr;
  l= b->l * tree->mod->l_coef[pos];
  theta = b->rght->theta[pos];
  kappa = tree->mod->kappa[pos];
  rr=tree->mod->rr[catg];

  /* Here we do take the gamma distribution coefficient in account, with the category integer*/
  /* A=0; C=1, G=2, T=3 */

  /*dPAA = dPTT  */
  b->dPij_rr_kappa[pos][catg][0][0] = b->dPij_rr_kappa[pos][catg][3][3] = -rr*l*1./2.*theta*exp(-(kappa+1.)/2.* rr *l);


  /*dPCC = dPGG  */
  b->dPij_rr_kappa[pos][catg][1][1] = b->dPij_rr_kappa[pos][catg][2][2] = -rr*l*1./2.*(1.-theta)*exp(-(kappa+1.)/2.* rr *l);


  /*dPGA = dPCT  */
  b->dPij_rr_kappa[pos][catg][2][0] = b->dPij_rr_kappa[pos][catg][1][3] = rr*l*1./2.*(1.-theta)*exp(-(kappa+1.)/2.* rr * l);


  /*dPAG = dPTC  */
  b->dPij_rr_kappa[pos][catg][0][2] = b->dPij_rr_kappa[pos][catg][3][1] = rr*l*1./2.*theta*exp(-(kappa+1.)/2.* rr *l);


  /*dPAT = dPTA = dPGT = dPCA  */
  b->dPij_rr_kappa[pos][catg][0][3] = b->dPij_rr_kappa[pos][catg][3][0] = b->dPij_rr_kappa[pos][catg][2][3] = b->dPij_rr_kappa[pos][catg][1][0] = 0.0;


  /*dPCG = dPGC = dPTG = dPAC  */
  b->dPij_rr_kappa[pos][catg][1][2] = b->dPij_rr_kappa[pos][catg][2][1] = b->dPij_rr_kappa[pos][catg][3][2] = b->dPij_rr_kappa[pos][catg][0][1] = 0.0;

}
/*********************************************************/
/*Computes the second derivatives of substitution probabilities along the edge b*/
/*in repect to kappa whose ancestor is node left*/
void Computes_Second_Derivative_Substitution_Probabilities_Kappa (edge *b, arbre *tree, int catg, int pos) {
  double theta, kappa,l, rr;
  l= b->l * tree->mod->l_coef[pos];
  theta = b->rght->theta[pos];
  kappa = tree->mod->kappa[pos];
  rr=tree->mod->rr[catg];

  /* Here we do take the gamma distribution coefficient in account, with the category integer*/
  /* A=0; C=1, G=2, T=3 */

  /*d2PAA = d2PTT  */
  b->d2Pij_rr_kappa[pos][catg][0][0] = b->d2Pij_rr_kappa[pos][catg][3][3] = rr*l*rr*l*1./4.*theta*exp(-(kappa+1.)/2.* rr *l);


  /*d2PCC = d2PGG  */
  b->d2Pij_rr_kappa[pos][catg][1][1] = b->d2Pij_rr_kappa[pos][catg][2][2] = rr*l*rr*l*1./4.*(1.-theta)*exp(-(kappa+1.)/2.* rr *l);


  /*d2PGA = d2PCT  */
  b->d2Pij_rr_kappa[pos][catg][2][0] = b->d2Pij_rr_kappa[pos][catg][1][3] = -rr*l*rr*l*1./4.*(1.-theta)*exp(-(kappa+1.)/2.* rr * l);


  /*d2PAG = d2PTC  */
  b->d2Pij_rr_kappa[pos][catg][0][2] = b->d2Pij_rr_kappa[pos][catg][3][1] = -rr*l*rr*l*1./4.*theta*exp(-(kappa+1.)/2.* rr *l);


  /*d2PAT = d2PTA = d2PGT = d2PCA  */
  b->d2Pij_rr_kappa[pos][catg][0][3] = b->d2Pij_rr_kappa[pos][catg][3][0] = b->d2Pij_rr_kappa[pos][catg][2][3] = b->d2Pij_rr_kappa[pos][catg][1][0] = 0.0;


  /*d2PCG = d2PGC = d2PTG = d2PAC  */
  b->d2Pij_rr_kappa[pos][catg][1][2] = b->d2Pij_rr_kappa[pos][catg][2][1] = b->d2Pij_rr_kappa[pos][catg][3][2] = b->d2Pij_rr_kappa[pos][catg][0][1] = 0.0;

}


/*********************************************************/

void Compute_Derivatives_Dkappa(arbre *tree, int pos){
  int site, catg, i;

  for (i=0;i<2*tree->n_otu-2;i++){
 
    // For(pos,tree->mod->n_pos){
      For(catg,tree->mod->n_catg)
	{
	  Computes_Derivative_Substitution_Probabilities_Kappa (tree->t_edges[i], tree, catg, pos);
	  Computes_Second_Derivative_Substitution_Probabilities_Kappa (tree->t_edges[i], tree, catg, pos);
	}
      // }
  }
  
  
  Post_Order_Dkappa(tree->noeud[tree->n_otu],
		    tree->noeud[tree->n_otu]->v[0],
		    tree, pos); // Rooted case!
  Post_Order_Dkappa(tree->noeud[tree->n_otu],
		    tree->noeud[tree->n_otu]->v[1],
		    tree, pos); // Rooted case!
  
    Post_Order_Dkappa(tree->noeud[tree->n_otu],
		      tree->noeud[tree->n_otu]->v[1],
		      tree, pos); // Rooted case!
    /*     Post_Order_Dkappa(tree->noeud[tree->n_otu],
		      tree->noeud[tree->n_otu]->v[0],
		      tree); // Rooted case!
    */
  
  For(site,tree->n_pattern)
    {
      if (tree->data->pos[site]==pos){
	tree->curr_site=site;
	Site_Derivatives_Dkappa(tree, pos);
      }
    }
 qksort(tree->tot_dloglk_dkappa_sorted, 0, tree->n_pattern-1);
 qksort(tree->tot_d2loglk_dkappa2_sorted, 0, tree->n_pattern-1);

 tree->tot_dloglk_dkappa = .0;
 tree->tot_d2loglk_dkappa2 = .0;

 For(site, tree->n_pattern){
   tree->tot_dloglk_dkappa += tree->tot_dloglk_dkappa_sorted[site];
   tree->tot_d2loglk_dkappa2 += tree->tot_d2loglk_dkappa2_sorted[site];
 }

  
}
/*********************************************************/

void Post_Order_Dkappa(node *pere, node *fils, arbre *tree, int pos){

  int i,dir1,dir2,dir3, site, catg, k, j;
  
  dir1 = dir2 = dir3 = -1;
  
  if(fils->tax) 
    {
      For(site,tree->n_pattern)
      {
	For (catg, tree->mod->n_catg){
	  For(k,tree->mod->ns){
	  fils->b[0]->dp_lk_kappa_rght[pos][site][catg][k]=0.0;
	  fils->b[0]->d2p_lk_kappa_rght[pos][site][catg][k]=0.0;
	  }
	}
      }
      return;
    }
  else
    {
      For(i,3)
	{
	  if(fils->v[i] != pere)
	    {
	     
	      Post_Order_Dkappa(fils,fils->v[i],tree, pos); // Compute the two derivatives of conditional likelihoods of L and R subtrees
	
	      if(dir1 < 0) dir1 = i; // find the two branches to study
	      else dir2 = i;
	    }
	  else dir3 = i; // v[dir3] is the father node
	}



 
      For(site,tree->n_pattern)
      {
	if (tree->data->pos[site]==pos) {
	For(catg,tree->mod->n_catg)
	{
	  For(k,tree->mod->ns) {//For all the sites possible at inferior node of edge under focus
	    fils->b[0]->dp_lk_kappa_rght[pos][site][catg][k]=0.;
	    For(i,tree->mod->ns)//For all the sites possible at root of first subtree
	    {              
	      For(j,tree->mod->ns) {//For all the sites possible at root of second subtree
		fils->b[0]->dp_lk_kappa_rght[pos][site][catg][k]+=
		  fils->b[1]->dPij_rr_kappa[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +
		  
		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->dp_lk_kappa_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->dPij_rr_kappa[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->dp_lk_kappa_rght[pos][site][catg][j] 
		  ;
		
	
	      }
	    }
	  }
	}
      }
      }
      For(site,tree->n_pattern)
      {
	if (tree->data->pos[site]==pos) {
	For(catg,tree->mod->n_catg)
	{
	  For(k,tree->mod->ns) {//For all the sites possible at inferior node of edge under focus
	    fils->b[0]->d2p_lk_kappa_rght[pos][site][catg][k]=0.;
	    For(i,tree->mod->ns)//For all the sites possible at root of first subtree
	    {              
	      For(j,tree->mod->ns) {//For all the sites possible at root of second subtree
		fils->b[0]->d2p_lk_kappa_rght[pos][site][catg][k]+=
		  fils->b[1]->d2Pij_rr_kappa[pos][catg][k][i] * //Derivative of the first group above
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +
		  
		  fils->b[1]->dPij_rr_kappa[pos][catg][k][i] *
		  fils->b[1]->dp_lk_kappa_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->dPij_rr_kappa[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->dPij_rr_kappa[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->dPij_rr_kappa[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->dp_lk_kappa_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] * //Derivative of the second  group above
		  fils->b[1]->d2p_lk_kappa_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->dPij_rr_kappa[pos][catg][k][i] *
		  fils->b[1]->dp_lk_kappa_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->dp_lk_kappa_rght[pos][site][catg][i] *
		  fils->b[2]->dPij_rr_kappa[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->dp_lk_kappa_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->dp_lk_kappa_rght[pos][site][catg][j] +

		  fils->b[1]->dPij_rr_kappa[pos][catg][k][i] *    //Derivative of the third  group above
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->dPij_rr_kappa[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->dp_lk_kappa_rght[pos][site][catg][i] *
		  fils->b[2]->dPij_rr_kappa[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->d2Pij_rr_kappa[pos][catg][k][j] *
		  fils->b[2]->p_lk_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->dPij_rr_kappa[pos][catg][k][j] *
		  fils->b[2]->dp_lk_kappa_rght[pos][site][catg][j] +

		  fils->b[1]->dPij_rr_kappa[pos][catg][k][i] *   //Derivative of the fourth  group above
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->dp_lk_kappa_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->dp_lk_kappa_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->dp_lk_kappa_rght[pos][site][catg][j] +
		  
		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->dPij_rr_kappa[pos][catg][k][j] *
		  fils->b[2]->dp_lk_kappa_rght[pos][site][catg][j] +

		  fils->b[1]->Pij_rr[pos][catg][k][i] *
		  fils->b[1]->p_lk_rght[pos][site][catg][i] *
		  fils->b[2]->Pij_rr[pos][catg][k][j] *
		  fils->b[2]->d2p_lk_kappa_rght[pos][site][catg][j] 
		  ;
	      }
	    }
	  }
	}
	}else {}
      }      
    }
  
}

/*********************************************************/
void  Site_Derivatives_Dkappa(arbre *tree, int pos){
  int catg, k, i, j, site=tree->curr_site;
  double dsite_lk, d2site_lk/*, realSiteLk*/, dlog_site_lk, d2log_site_lk;
  dsite_lk =0., d2site_lk=0.;


	For(catg,tree->mod->n_catg)
	{
	  For(k,tree->mod->ns) {//For all the sites possible at inferior node of edge under focus
	    For(i,tree->mod->ns)//For all the sites possible at root of first subtree
	    {              
	      For(j,tree->mod->ns) {//For all the sites possible at root of second subtree
		dsite_lk += tree->mod->r_proba[catg] * 
		  tree->noeud[tree->n_otu]->pi[pos][k]  *(
		  tree->noeud[tree->n_otu]->b[0]->dPij_rr_kappa[pos][catg][k][i] *
		  tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		  tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		  tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		  
		  tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		  tree->noeud[tree->n_otu]->b[0]->dp_lk_kappa_rght[pos][site][catg][i] *
		  tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		  tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +

		  tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		  tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		  tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[pos][catg][k][j] *
		  tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +

		  tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		  tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		  tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		  tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[pos][site][catg][j] )
		  ;
		/*		printf("dPij_rr_kappa[catg][k][j] : %f\n",tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[catg][k][j] );
		printf("dp_lk_kappa_rght[site][catg][j] : %f\n",  tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[site][catg][j]);
		*/
	      }
	    }
	  }
	}
	
	For(catg,tree->mod->n_catg)
	{
	  For(k,tree->mod->ns) {//For all the sites possible at inferior node of edge under focus
	    For(i,tree->mod->ns)//For all the sites possible at root of first subtree
	    {              
	      For(j,tree->mod->ns) {//For all the sites possible at root of second subtree
		d2site_lk+= tree->mod->r_proba[catg] * 
		  tree->noeud[tree->n_otu]->pi[pos][k]  *
		  (
		   tree->noeud[tree->n_otu]->b[0]->d2Pij_rr_kappa[pos][catg][k][i] * //Derivative of the first group above
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->dPij_rr_kappa[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->dp_lk_kappa_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->dPij_rr_kappa[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->dPij_rr_kappa[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] * //Derivative of the second  group above
		   tree->noeud[tree->n_otu]->b[0]->d2p_lk_kappa_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->dPij_rr_kappa[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->dp_lk_kappa_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->dp_lk_kappa_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->dp_lk_kappa_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->dPij_rr_kappa[pos][catg][k][i] *    //Derivative of the third  group above
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->dp_lk_kappa_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->d2Pij_rr_kappa[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->p_lk_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->dPij_rr_kappa[pos][catg][k][i] *   //Derivative of the fourth  group above
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->dp_lk_kappa_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[pos][site][catg][j] +
		   
		   tree->noeud[tree->n_otu]->b[0]->Pij_rr[pos][catg][k][i] *
		   tree->noeud[tree->n_otu]->b[0]->p_lk_rght[pos][site][catg][i] *
		   tree->noeud[tree->n_otu]->b[1]->Pij_rr[pos][catg][k][j] *
		   tree->noeud[tree->n_otu]->b[1]->d2p_lk_kappa_rght[pos][site][catg][j] )
		  ;
	      }
	    }
	  }
	}


	//	realSiteLk=exp(tree->site_lk[site]); 
	dlog_site_lk = dsite_lk/tree->true_site_lk[site];//dsite_lk/realSiteLk;//tree->site_lk[site]; 
	d2log_site_lk = d2site_lk/tree->true_site_lk[site] - ((dsite_lk/tree->true_site_lk[site])*(dsite_lk/tree->true_site_lk[site]));//d2site_lk/realSiteLk - ((dsite_lk/realSiteLk)*(dsite_lk/realSiteLk));

	/*		printf("Kappa : %f\n", tree->mod->kappa);
	  printf("RealSiteLk : %f, dsite_lk : %f, d2site_lk : %f\n", realSiteLk, dsite_lk, d2site_lk);
	  printf("b[1]->d2Pij_rr_kappa : %f, b[1]->d2p_lk_kappa_rght :%f\n", tree->noeud[tree->n_otu]->b[1]->d2Pij_rr_kappa[0][2][1], tree->noeud[tree->n_otu]->b[1]->d2p_lk_kappa_rght[site][0][1]);
	  printf("b[1]->dPij_rr_kappa : %f, b[1]->dp_lk_kappa_rght :%f\n", tree->noeud[tree->n_otu]->b[1]->dPij_rr_kappa[0][2][1], tree->noeud[tree->n_otu]->b[1]->dp_lk_kappa_rght[site][0][1]);*/
	tree->tot_dloglk_dkappa_sorted[site] = tree->data->wght[site]*dlog_site_lk;
	tree->tot_d2loglk_dkappa2_sorted[site] = tree->data->wght[site]*d2log_site_lk;



}


/*********************************************************/

int Compare_Two_States(char *state1, char *state2, int state_size)
{
  /* 1 the two states are identical */
  /* 0 the two states are different */
  int i;

  For(i,state_size) if(state1[i] != state2[i]) break;
  
  return (i==state_size)?(1):(0);
}

/*********************************************************/

model *Make_Model_Basic()
{
  /* Start the allocation of memory for the substitution model */

  model *mod;

  mod                 = (model *)mCalloc(1,sizeof(model)); 
  mod->pi = (double*) mCalloc(T_MAX_ALPHABET, sizeof(double)); 
  mod->kappa               = (double *)mCalloc(1, sizeof(double));
  mod->omega               = (double *)mCalloc(1, sizeof(double));
  mod->kappa[0]=4.; 
  mod->EqFreq            = 0; /*by default, each branch has its own equilibrium frequency that is optimized */
  mod->NumEqFreq         = 1; /*by default, the number of equilibrium frequencies would be 1 if EqFreq turned out to be ==1 */
 
  return mod;
}

/*********************************************************/


code *Make_Code(int num_code)
{
  /* Genetic codes 
     http://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi?mode=c#thetop
  */

  int i,j;
  code *c_code;


/*
Genetic codes : -1 if a nonsense codon
Codon rank in the list is related to its composition :
We pose :
A : 0
C : 1
G : 2
T : 3
The relation is the following one :
codon_base[1]=(codon number/16)%4
codon_base[2]=(codon number/4)%4
codon_base[3]=(codon number)%4
Ex :
Codon number 14 :
codon_base[1]=(14/16)%4=0 -> A
codon_base[2]=(14/4)%4=3 -> T
codon_base[3]=(14)%4=2 -> G
Should code for M, which is amino-acid number 12, as we see in the list "universal".
*/
  int genetc[17][64] ={
 
     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,
      8,14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,
      0,7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,-1,
      4,17,4,10,13,10,13},    /* 0: universal */

     {11,2,11,2,16,16,16,16,-1,15,-1,15,12,9,12,9,5,8,
      5,8,14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,
      0,0,7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,
      17,4,17,4,10,13,10,13}, /* 1:vertebrate mt.*/
     
     {11,2,11,2,16,16,16,16,1,15,1,15,12,9,12,9,5,8,5,
      8,14,14,14,14,1,1,1,1,16,16,16,16,6,3,6,3,0,0,0,
      0,7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,17,
      4,17,4,10,13,10,13},    /* 2: yeast mt. */

     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,7,
      7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,17,4,17,
      4,10,13,10,13},         /* 3: mold mt. */

     {11,2,11,2,16,16,16,16,15,15,15,15,12,9,12,9,5,8,
      5,8,14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,
      0,0,7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,
      17,4,17,4,10,13,10,13}, /* 4: invertebrate mt. */
     
     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,
      7,7,7,7,19,19,19,19,5,18,5,18,15,15,15,15,-1,4,
      17,4,10,13,10,13},      /* 5: ciliate nu.*/
     
     {2,2,11,2,16,16,16,16,15,15,15,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,7,
      7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,17,4,17,
      4,10,13,10,13},         /* 6: echinoderm mt.*/
     
     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,7,
      7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,4,4,17,
      4,10,13,10,13},         /* 7: euplotid mt. */
     
     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,
      7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,-1,4,
      17,4,10,13,10,13},      /* 8: bacterial and plant plastid */

     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,15,10,6,3,6,3,0,0,0,0,
      7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,-1,
      4,17,4,10,13,10,13},    /* 9: alternative yeast nu.*/

     {11,2,11,2,16,16,16,16,7,15,7,15,12,9,12,9,5,8,5,
      8,14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,
      0,7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,17,
      4,17,4,10,13,10,13},    /* 10: ascidian mt. */
     
     {2,2,11,2,16,16,16,16,15,15,15,15,9,9,12,9,5,8,5,
      8,14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,
      0,7,7,7,7,19,19,19,19,18,18,-1,18,15,15,15,15,17,
      4,17,4,10,13,10,13},    /* 11: alternative flatworm mt. */

     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,
      7,7,7,7,19,19,19,19,-1,18,5,18,15,15,15,15,-1,4,
      17,4,10,13,10,13},      /* 12: blepharisma nu. */
     
     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,
      7,7,7,7,19,19,19,19,-1,18,10,18,15,15,15,15,-1,4,
      17,4,10,13,10,13},      /* 13: chlorophycean mt. */
     
     {2,2,11,2,16,16,16,16,15,15,15,15,12,9,12,9,5,8,5,
      8,14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,
      0,7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,17,
      4,17,4,10,13,10,13},    /* 14: trematode mt. */

     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,7,
      7,7,7,19,19,19,19,-1,18,10,18,-1,15,15,15,-1,4,17,
      4,10,13,10,13},         /* 15: scenedesmus obliquus mt. */

     {11,2,11,2,16,16,16,16,1,15,1,15,9,9,12,9,5,8,5,8,
      14,14,14,14,1,1,1,1,10,10,10,10,6,3,6,3,0,0,0,0,
      7,7,7,7,19,19,19,19,-1,18,-1,18,15,15,15,15,-1,4,
      17,4,-1,13,10,13}       /* 16: thraustochytrium mt. */
    };

  c_code = (code *)mCalloc(1,sizeof(code));

  c_code->n_diff_b_2_codons = (int *)mCalloc(64*64,sizeof(int));
  c_code->tstvtable         = (int *)mCalloc(64*64,sizeof(int));
  c_code->gg_98_codon       = (double **)mCalloc(64*64,sizeof(double *));
  c_code->sense_c           = (int *)mCalloc(64,sizeof(int));
  c_code->from_64_2_61      = (int *)mCalloc(64,sizeof(int));
  c_code->name              = (char *)mCalloc(100,sizeof(char));


  /*Here we set the genetic code*/
  switch(num_code)
    {
    case 0 :
      {
	c_code->num_curr_code = 0;
	strcpy(c_code->name,"universal");
	c_code->n_sense_c = 61;
	break;
      }
    case 1 :
      {
	c_code->num_curr_code = 1;
	strcpy(c_code->name,"vertebrate mt.");
	c_code->n_sense_c = 60;
	break;
      }
    case 2 :
      {
	c_code->num_curr_code = 2;
	strcpy(c_code->name,"yeast mt.");
	c_code->n_sense_c = 62;
	break;
      }
    case 3 :
      {
	c_code->num_curr_code = 3;
	strcpy(c_code->name,"mold mt.");
	c_code->n_sense_c = 62;
	break;
      }
    case 4 :
      {
	c_code->num_curr_code = 4;
	strcpy(c_code->name,"invertebrate mt.");
	c_code->n_sense_c = 62;
	break;
      }
    case 5 :
      {
	c_code->num_curr_code = 5;
	strcpy(c_code->name,"ciliate nu.");
	c_code->n_sense_c = 63;
	break;
      }
    case 6 :
      {
	c_code->num_curr_code = 6;
	strcpy(c_code->name,"echinoderm nu.");
	c_code->n_sense_c = 62;
	break;
      }
    case 7 :
      {
	c_code->num_curr_code = 7;
	strcpy(c_code->name,"euplotid nu.");
	c_code->n_sense_c = 62;
	break;
      }
    case 8 :
      {
	c_code->num_curr_code = 8;
	strcpy(c_code->name,"bacterial & plant plastid");
	c_code->n_sense_c = 61;
	break;
      }
    case 9 :
      {
	c_code->num_curr_code = 9;
	strcpy(c_code->name,"alternative yeast nu.");
	c_code->n_sense_c = 61;
	break;
      }
    case 10 :
      {
	c_code->num_curr_code = 10;
	strcpy(c_code->name,"ascidian mt.");
	c_code->n_sense_c = 62;
	break;
      }
    case 11 :
      {
	c_code->num_curr_code = 11;
	strcpy(c_code->name,"alternative flatworm mt.");
	c_code->n_sense_c = 63;
	break;
      }
    case 12 :
      {
	c_code->num_curr_code = 12;
	strcpy(c_code->name,"blepharisma nu.");
	c_code->n_sense_c = 62;
	break;
      }
    case 13 :
      {
	c_code->num_curr_code = 13;
	strcpy(c_code->name,"chlorophycean mt.");
	c_code->n_sense_c = 62;
	break;
      }
    case 14 :
      {
	c_code->num_curr_code = 14;
	strcpy(c_code->name,"trematode mt.");
	c_code->n_sense_c = 62;
	break;
      }
    case 15 :
      {
	c_code->num_curr_code = 15;
	strcpy(c_code->name,"scenedesmus obliquus mt.");
	c_code->n_sense_c = 61;
	break;
      }
    case 16 :
      {
	c_code->num_curr_code = 16;
	strcpy(c_code->name,"thraustochytrium mt.");
	c_code->n_sense_c = 60;
	break;
      }
    default : {Exit("\n. Err. in Make_Code (unknown genetic code)\n");}
    }


  c_code->aa = (char *)mCalloc(23,sizeof(char));
  c_code->genetc = (int **)mCalloc(17,sizeof(int *));
  For(i,17) c_code->genetc[i] = (int *)mCalloc(64,sizeof(int));

  /*The amino-acids*/
  strcpy(c_code->aa,"ARNDCQEGHILKMFPSTWYV?-");

  /*We copy the genetic codes into the appropriate code recipients*/
  For(i,17) For(j,64) c_code->genetc[i][j] = genetc[i][j];
  
  /*Maps sense codons from nonsense codons by filling c_code->from_64_2_61 and c_code->sense_c[n] 
   c_code->sense_c contains all the sense codons composition numbers without a gap (without nonsense codons*/
  Init_Codons(c_code);
  Init_Nb_Diff_Between_Two_Codons(c_code,c_code->n_sense_c);
  Init_Tstv_Table(c_code,c_code->n_sense_c);
 
  return c_code;
}

/*********************************************************/

void Init_Codons(code *c_code)
/*
Goes through all the genetic code to map sense codons from nonsense codons.
*/
{
  int i,n;

  n = 0;
  For(i,64) 
    {
      if(c_code->genetc[c_code->num_curr_code][i] > -1) 
	{
	  c_code->from_64_2_61[i] = n; //numbers the sense codons
	  c_code->sense_c[n] = i; //builds an array of sense codons indices.
	  n++;
	}
      else
	  c_code->from_64_2_61[i] = -1;
    }
  if(n != c_code->n_sense_c) Exit("\n. Err. in Init_Codons !\n");
}

/*********************************************************/


void Init_Nb_Diff_Between_Two_Codons(code *c_code, int ns)
{
  /* Get the number of differences for each pair of codon */

  char *codon1,*codon2;
  int i,j,k;
  int diff;
  int num_codon1, num_codon2;

  codon1 = (char *)mCalloc(3,sizeof(char));
  codon2 = (char *)mCalloc(3,sizeof(char));

  num_codon1 = num_codon2 = -1;
  For(i,ns)
    {      
      num_codon1 = c_code->sense_c[i];

      codon1[0] = (num_codon1/16)%4; 
      codon1[1] = (num_codon1/4)%4; 
      codon1[2] = num_codon1%4;

      
      For(j,ns)
	{	  
	  num_codon2 = c_code->sense_c[j];

	  codon2[0] = (num_codon2/16)%4; 
	  codon2[1] = (num_codon2/4)%4; 
	  codon2[2] = num_codon2%4;	  
	  
	  diff = 0;
	  For(k,3) if(codon1[k] != codon2[k]) diff++;
	  c_code->n_diff_b_2_codons[i*ns+j] = diff;

	}      
    }

  Free(codon1);
  Free(codon2);

  return;
}

/*********************************************************/


void Init_Tstv_Table(code *c_code, int ns)
{
  /* If two codons differ by only one change, determine
     if this change is a transversion or a transition */


  int *codon1,*codon2;
  int i,j,k;
  int diff,pos;
  int num_codon1, num_codon2;

  codon1 = (int *)mCalloc(3,sizeof(int));
  codon2 = (int *)mCalloc(3,sizeof(int));

  num_codon1 = num_codon2 = -1;
  For(i,ns)
    {
      num_codon1 = c_code->sense_c[i];
            
      codon1[0] = (num_codon1/16)%4; 
      codon1[1] = (num_codon1/4)%4; 
      codon1[2] = num_codon1%4;

      
      For(j,ns)
	{
	  num_codon2 = c_code->sense_c[j];

	  codon2[0] = (num_codon2/16)%4; 
	  codon2[1] = (num_codon2/4)%4; 
	  codon2[2] = num_codon2%4;	  
	  
	  diff = 0;
	  pos = -1;
	  For(k,3) if(codon1[k] != codon2[k]) 
	    {
	      diff++;
	      pos = k;
	    }
	  if(diff > 1) c_code->tstvtable[i*ns+j] = -1;
	  else 
	    {
	      if(diff == 1)
		{
		  if((((codon1[pos] == 0) || (codon1[pos] == 2))  &&
		      ((codon2[pos] == 1) || (codon2[pos] == 3))) ||
		     (((codon1[pos] == 1) || (codon1[pos] == 3))  &&
		      ((codon2[pos] == 0) || (codon2[pos] == 2))))
		    c_code->tstvtable[i*ns+j] = 2; /* tv */
		  else
		    c_code->tstvtable[i*ns+j] = 1; /* ts */
		}
	      else  c_code->tstvtable[i*ns+j] = 0;	  
	    }
	}
    }

  Free(codon1);
  Free(codon2);
  return;
}

/*********************************************************/

/*This function fills gg98_codons with relative substitution rates (only kappa in the gg98 model), as Init_GTR_4_Codon_Table did in fitmodel.
We use only one tstv ratio here, kappa[0]
 */
void Init_GG98_Codon_Table(model *mod)
{
  int *codon1,*codon2;
  int i,j,k;
  int diff,pos;
  int num_codon1, num_codon2;
  int ns;

  codon1 = (int *)mCalloc(3,sizeof(int));
  codon2 = (int *)mCalloc(3,sizeof(int));

  ns = mod->c_code->n_sense_c;

  For(i,ns*ns) mod->c_code->gg_98_codon[i] = NULL;

  num_codon1 = num_codon2 = -1;
  For(i,ns)
    {
      num_codon1 = mod->c_code->sense_c[i];
            
      codon1[0] = (num_codon1/16)%4; 
      codon1[1] = (num_codon1/4)%4; 
      codon1[2] = num_codon1%4;

      
      For(j,ns)
	{
	  num_codon2 = mod->c_code->sense_c[j];

	  codon2[0] = (num_codon2/16)%4; 
	  codon2[1] = (num_codon2/4)%4; 
	  codon2[2] = num_codon2%4;	  
	  



	  diff = 0;
	  pos = -1;
	//Why computing again the differences already stocked in c_code->n_diff_b_2_codons ?
	  //Anyway, this is not an important problem as we do not compute this very often...
	  For(k,3) if(codon1[k] != codon2[k]) 
	    {
	      diff++;
	      pos = k;
	    }
	  if(diff > 1) mod->c_code->gg_98_codon[i*ns+j] = NULL;
	  else 
	    {
                if(diff == 1)
                    {
                        switch(codon1[pos])
                            {
                            case 0 : /* A */
                                {
                                    switch(codon2[pos])
                                        {
                                        case 1 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->unity); break;}/* A->C */
                                        case 2 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->kappa[0]); break;}/* A->G */
                                        case 3 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->unity); break;}/* A->T */
                                        default : {Exit("\n. Err. in Init_GG98_codon_Table\n");}
                                        }
                                    break;
                                }
                                
                            case 1 : /* C */
                                {
                                    switch(codon2[pos])
                                        {
                                        case 0 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->unity); break;}/* C->A */
                                        case 2 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->unity); break;}/* C->G */
                                        case 3 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->kappa[0]); break;}/* C->T */
                                        default : {Exit("\n. Err. in Init_GG98_codon_Table\n");}
                                        }
                                    break;
                                }
                            case 2 : /* G */
                                {
                                    switch(codon2[pos])
                                        {
                                        case 0 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->kappa[0]); break;}/* G->A */
                                        case 1 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->unity); break;}/* G->C */
                                        case 3 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->unity); break;}/* G->T */
                                        default : {Exit("\n. Err. in Init_gg98_codon_Table\n");}
                                        }
                                    break;
                                }
                            case 3 : /* T */
                                {
                                    switch(codon2[pos])
                                        {
                                        case 0 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->unity); break;}/* T->A */
                                        case 1 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->kappa[0]); break;}/* T->C */
                                        case 2 : {mod->c_code->gg_98_codon[i*ns+j] = &(mod->unity); break;}/* T->G */
                                        default : {Exit("\n. Err. in Init_GG98_codon_Table\n");}
                                        }
                                    break;
                                }
                            default : Exit("\n. Err. in Init_GG98_codon_Table\n");
                            }
                    }
	    }
	}
    }

  Free(codon1);
  Free(codon2);
  return;
}

/*********************************************************/
void Make_All_Qmat_Struct(model *mod, int n_otu)
{

  /* Allocate memory of the instantaneous rate matrices structure */
  /* One matrice per branch */
  int i;

  mod->qmat_struct = (qmat **)mCalloc(2*n_otu-2,sizeof(qmat*));
  
  for(i=1;i<2*n_otu-2;i++){
    mod->qmat_struct[i] = Make_Qmat_Struct(mod->ns,mod);
  } 
  
  
}

/*********************************************************/
qmat *Make_Qmat_Struct(int ns,  model *mod)
{
  /* Make up the instantaneous rate matrix */

  int i;
  qmat *qmat_struct;

  
  qmat_struct  = (qmat *)mCalloc(1,sizeof(qmat));

 
  qmat_struct->u_mat             = (double *) mCalloc(ns*ns,  sizeof(double ));
  qmat_struct->v_mat             = (double *) mCalloc(ns*ns,  sizeof(double ));
  qmat_struct->root_vct          = (double *) mCalloc(ns,     sizeof(double ));
  qmat_struct->expD_mr_vct       = (double *) mCalloc(ns,     sizeof(double ));
  qmat_struct->qmat              = (double *) mCalloc(ns*ns,  sizeof(double )); 
  
  for (i=0;i<ns*ns;i++) {
    qmat_struct->qmat[i]=0.0;
    //printf("qmat_struct->qmat[i] : %f\n",qmat_struct->qmat[i]); 
  }
  
  return qmat_struct;
}

/*********************************************************/
/* this function computes codon equilibrium frequencies from an equilibrium GC content*/
/*using the formula from Schadt and Lange 2002 :

     pi(XaXbXc)=pi(Xa)*pi(Xb)*pi(Xc)/(sum_over_a,b,c (pi(Xa)*pi(Xb)*pi(Xc)))

*/
void Compute_Codon_Equilibrium_Frequencies (double *cpi, double gc, model *mod) {
  double pi[4];
  double sum;
  int num_codon, i, j, k;

  /* A=0; C=1, G=2, T=3 */
  pi[0]=pi[3]=(1-gc)/2.;
  pi[1]=pi[2]=gc/2.;
 
  
  sum=0;
  For(i,4)
    {
      For(j,4)
	{
	  For(k,4)
	    {
	      num_codon = i*16+j*4+k;
	      if(mod->c_code->from_64_2_61[num_codon] > -1) //If not nonsense
		{
		  cpi[mod->c_code->from_64_2_61[num_codon]] = 
		    pi[i] *
		    pi[j] *
		    pi[k];
		  /*If we want to check the equilibrium codon frequencies
		    printf("Compute_Codon_Equilibrium_Frequencies\n");
		    printf("gc : %f\n", gc);
		    if (i==0) {
		    printf("A");
		  }
		  if (i==1) {
		    printf("C");
		  }
		  if (i==2) { //optimizations  
  if (input->changetree) {
    NR_Tot_First(tree);
    Simu(tree, 1000);
    if (!input->quick) {
      printf("\n\n. Last optimization step...");
      NR_Tot(tree); 
    }
    else {
      printf("\n\n. No Final optimization.");
    }
  }
  else NR_Tot(tree);

  Fill_lnf_File(tree,fp_lnf_out);
		    printf("G");
		  }
		  if (i==3) {
		    printf("T");
		  }
		  if (j==0) {
		    printf("A");
		  }
		  if (j==1) {
		    printf("C");
		  }
		  if (j==2) {
		    printf("G");
		  }
		  if (j==3) {
		    printf("T");
		  }
		  if (k==0) {
		    printf("A");
		  }
		  if (k==1) {
		    printf("C");
		  }
		  if (k==2) {
		    printf("G");
		  }
		  if (k==3) {
		    printf("T");
		  }
		  printf("  %f\n",cpi[mod->c_code->from_64_2_61[num_codon]]);*/
		  sum += cpi[mod->c_code->from_64_2_61[num_codon]];
		}
	    }
	}
    }
  
  For(i,mod->c_code->n_sense_c) cpi[i] /= sum;
  
  


}

/*********************************************************/

int Get_AA(code *c_code, int num_codon)
{
  /* return the amino acid number that correspond to codon 'num_codon' */
  return c_code->genetc[c_code->num_curr_code][num_codon];

}

/*********************************************************/
void      Fill_lnf_File (arbre *tree, FILE *fp_lnf_out) 
{
  int i;

  fprintf(fp_lnf_out,"    1  %d  %d\n\n\n",tree->data->init_len, tree->data->crunch_len);
  
  fprintf(fp_lnf_out," 1 \n \n");
    
  for(i=0;i<tree->data->crunch_len;i++) {
    fprintf(fp_lnf_out,"     %d     %d   %f\n",i+1, tree->data->wght[i],tree->site_lk[i]); 
    
  }
      

}


/*********************************************************/
/*
This function creates the right amount of thetas and sets their values uniformly accross the range 
0-1
*/
int Make_Thetas (model *mod) {
  int i;
  double StepSize;
  StepSize=(1./(double)mod->NumEqFreq);
  for (i=0;i<mod->NumEqFreq; i++ ) {
    mod->ThetaValues[i]=((float)i+1)/(mod->NumEqFreq+1);
    printf ("Thetavalue %d : %f\n",i, mod->ThetaValues[i]);
  }

  return 0;
}
/*********************************************************/
/*For a given branch, tries all the available thetas and choses the one that maximizes the likelihood*/
double Try_Thetas (arbre *tree, edge *b, int ChangeValues) {
  int i,j, n_iter;
  double prop_bl, old_bl, lk_new, lk_old, div;
  double starting_theta, starting_bl, starting_lk;
  double best_theta, best_bl, best_lk;
  
  //We record the starting parameters
  starting_theta=b->rght->theta[0];  
  starting_bl=b->l;
  starting_lk=Lk_At_Given_Edge(tree, b);

  best_theta=starting_theta;
  best_bl=starting_bl;
  best_lk=starting_lk; 


  lk_new=tree->tot_loglk;
  lk_old=UNLIKELY;

  // printf("For branch %d:", b->num);

  for (i=0;i<tree->mod->NumEqFreq;i++) {
    // printf("\nFor Equilibrium Frequency %d:", i);
    for (j=0;j<tree->mod->n_pos;j++) { 
      // There we could incorporate a different theta value for 
      // each position in the sequence
      b->rght->theta[j]=tree->mod->ThetaValues[i];
    }
    //We reset the branch length value
    b->l=starting_bl;
    lk_old=UNLIKELY; 
    lk_new=Lk_At_Given_Edge(tree, b);
    //   printf("lknew : %f\n", lk_new);
    //We optimize the branch length in agreement with this new theta
    while (lk_new>lk_old+MIN_DIFF_LK*100) {
      lk_old=lk_new=tree->tot_loglk;
      old_bl=b->l;

     
      //printf ("First Likelihood : %f\n", tree->tot_loglk);
      prop_bl=NR_Bl_Br(b,tree);

      n_iter=0;
      lk_new=UNLIKELY;
      //Churchill-Felsenstein like technique : 
      //if the likelihood does not increase
      //when applying the whole improvement, make smaller improvement.
      while ((lk_new<lk_old) && (n_iter<100)){
	div=pow(2,n_iter);
	  b->l =  old_bl+(prop_bl)/div;
	  
	  if (tree->t_edges[i]->l<=0.) {
	    tree->t_edges[i]->l=0.000001;
	  } 
	  n_iter++;

	  //	  lk_new=Return_Lk(tree);
	  // if (isnan(lk_new))   lk_new=Return_Lk(tree);
	  lk_new=Lk_At_Given_Edge(tree, b);
	  if (isnan(lk_new))   lk_new=Lk_At_Given_Edge(tree, b);
	  //  printf("\n\tl5298 Likelihood : %f, n_iter : %d", lk_new, n_iter);
      }

      if (n_iter==100) {
	b->l =  old_bl;
	lk_new=lk_old;
      }

    }
    //Now we have an approximate ML score for this theta on this branch; 
    //if superior to the best theta score, we save the new best values
    if (lk_new>best_lk) {
      best_theta=b->rght->theta[0];
      // printf("best_theta : %f\n", best_theta);
      best_bl=b->l;
      best_lk=lk_new;
    }

  }
  if (ChangeValues) {
    for (j=0;j<tree->mod->n_pos;j++) {
      b->rght->theta[j] = best_theta;
    }
    b->l = best_bl;
    tree->tot_loglk=best_lk;    
  } 
  else {
    //Back to the starting situation
    b->l=starting_bl;
    for (j=0;j<tree->mod->n_pos;j++) {
      b->rght->theta[j]=starting_theta;
    }
    tree->tot_loglk=starting_lk;
  }
  //We need to record the best values found during this search

  b->best_theta = best_theta;
  b->best_l = best_bl;
  b->diff_lk = starting_lk - best_lk;
  return (best_lk);
}
/*********************************************************/
void Select_Edges_Theta_To_Change(arbre *tree, edge **sorted_b, int *n_neg)
{
  int i;
  edge *b;
  int min;


  *n_neg = 0;
  tree->min_diff_lk = .0;
  min = 0;
  For(i,2*tree->n_otu-2)
    {
      b = tree->t_edges[i];

      if(b->diff_lk < 0.0-MDBL_MIN) 
	{

	  if(b->diff_lk < tree->min_diff_lk) 
	    {
	      tree->min_diff_lk = b->diff_lk;
	      min = i;
	    }

	  sorted_b[*n_neg] = b;
	  (*n_neg)++;
	}
    }
}

/*********************************************************/
/*
  In the case where we limit the number of thetas, 
  and where each branch choses its preferred theta among the available ones,
  this function makes all the changes that resulted in the best increases 
  in likelihood value.
*/
void Make_Theta_Changes(arbre *tree,edge **b,int beg,int end) {
  int i, j;
  for(i=beg;i<end;i++)
    {
      //Saving the old values :
      b[i]->rght->theta_old[0]=b[i]->rght->theta[0];
      b[i]->l_old=b[i]->l;
      for(j=0;j<tree->mod->n_pos;j++){
	b[i]->rght->theta[j]=b[i]->best_theta;
      }
      b[i]->l=b[i]->best_l;
      
    }

}
/*********************************************************/
/*
  In the case where we limit the number of thetas, 
  and where each branch choses its preferred theta among the available ones,
  this function makes all the changes that resulted in the best increases 
  in likelihood value.
*/
void UnMake_Theta_Changes(arbre *tree,edge **b,int beg,int end) {
  int i, j;
  for(i=beg;i<end;i++)
    {
      for(j=0;j<tree->mod->n_pos;j++){
	b[i]->rght->theta[j]=b[i]->rght->theta_old[0];
      }
      b[i]->l=b[i]->l_old;
      
    }

}

/*********************************************************/
void      Fill_lnf_File_iter (arbre *tree, FILE *fp_lnf_out, int num) 
{
  int i;

 
 
  fprintf(fp_lnf_out," %d \n \n",num);
    
  for(i=0;i<tree->data->crunch_len;i++) {
    fprintf(fp_lnf_out,"     %d     %d   %f\n",i+1, tree->data->wght[i],tree->site_lk[i]); 
    
  }
      

}


/*********************************************************/
//Here we marginally reconstruct ancestral distributions in an empirical bayesian way.
//I have not taken care of the site order, so they are not the same as in the input sequences
void Reconstruct_Ancestral_Distributions(arbre *tree, FILE *fout, FILE *fout2) {
  //k : node number
  //catg : gamma distribution category
  //m : state at father node
  //n : state at one son node
  //o : state at the other son node
  //i : site number
  //a : state at node of interest
  //j : number of sites having this pattern

  int k, catg, m, n, o, i, a, j;
  double ***probas;
  double **means;



 
  //For each node, we allocate memory for a table containing for each site pattern the probability distribution of bases
  probas=(double***)mCalloc(2*tree->n_otu-1,sizeof(double**));
  for(k=0;k<2*tree->n_otu-1;k++) {
    probas[k]=(double**)mCalloc(tree->data->crunch_len,sizeof(double*));
    for(i=0;i<tree->data->crunch_len;i++) {
      probas[k][i]=(double*)mCalloc(tree->mod->ns,sizeof(double));
    }
  }
  means=(double**)mCalloc(2*tree->n_otu-1,sizeof(double*));
  for(k=0;k<2*tree->n_otu-1;k++) {
    means[k]=(double*)mCalloc(tree->mod->ns,sizeof(double));
  }

  //Computation
  for (k=0; k<2*tree->n_otu-1;k++) {
    if (k==tree->n_otu) { // We are at the root
      for(i=0;i<tree->data->crunch_len;i++) {
	for(a=0;a<tree->mod->ns;a++) {
	  probas[k][i][a]=0.0;
	  for(n=0;n<tree->mod->ns;n++) {
	    for(o=0;o<tree->mod->ns;o++) {
	      for(catg=0;catg<tree->mod->n_catg;catg++) {
		probas[k][i][a]+=
		  tree->mod->r_proba[catg] * // Rate probability
		  tree->noeud[k]->pi[tree->data->pos[i]][a] *  // Proportion at root of nucleotide a
		  tree->noeud[k]->b[0]->p_lk_rght[tree->data->pos[i]][i][catg][n] *   // Inferior Partial Likelihood
		  tree->noeud[k]->b[0]->Pij_rr[tree->data->pos[i]][catg][a][n] *         // Substitution probability
		  tree->noeud[k]->b[1]->p_lk_rght[tree->data->pos[i]][i][catg][o] *   // Inferior Partial Likelihood
		  tree->noeud[k]->b[1]->Pij_rr[tree->data->pos[i]][catg][a][o]    ;     // Substitution probability
	      }
	    }
	  }
	}
      }
    }
    //else if not at a root...
    else if (!tree->noeud[k]->tax){ //...but not at a leaf
      for(i=0;i<tree->data->crunch_len;i++) {
	for(a=0;a<tree->mod->ns;a++) {
	  probas[k][i][a]=0.0;
	  for(m=0;m<tree->mod->ns;m++) {
	    for(n=0;n<tree->mod->ns;n++) {
	      for(o=0;o<tree->mod->ns;o++) {
		for(catg=0;catg<tree->mod->n_catg;catg++) {
		  probas[k][i][a]+=
		    tree->mod->r_proba[catg] * // Rate probability
		    tree->noeud[k]->b[0]->p_lk_left[tree->data->pos[i]][i][catg][m]*
		    tree->noeud[k]->b[0]->Pij_rr[tree->data->pos[i]][catg][m][a] *
		    tree->noeud[k]->b[1]->Pij_rr[tree->data->pos[i]][catg][a][n] *
		    tree->noeud[k]->b[1]->p_lk_rght[tree->data->pos[i]][i][catg][n]*
		    tree->noeud[k]->b[2]->Pij_rr[tree->data->pos[i]][catg][a][o] *
		    tree->noeud[k]->b[2]->p_lk_rght[tree->data->pos[i]][i][catg][o];					  
		}
	      }
	    }
	  }
	}
      }
    }
  }
  
  //We should have by now all the base likelihoods at internal nodes.
  //We therefore output them, after building the distributions rather than the likelihoods.
  
  //We compute the probability distributions
  for(i=0;i<tree->data->crunch_len;i++) {
    for (k=0; k<2*tree->n_otu-1;k++) {
      if (!tree->noeud[k]->tax){
	for(a=0;a<tree->mod->ns;a++) {
	  probas[k][i][a]/=tree->true_site_lk[i];
	}
      }
    }
  }

  /*
  if (Filexists("BaseDistributions.txt")) {
    fout = Openfile("BaseDistributions.txt",2); //We append
    fprintf(fout, "************************New Tree********************");
  }
  else {
    fout = Openfile("BaseDistributions.txt",1);
  }
  */
  for (k=0; k<2*tree->n_otu-1;k++) {
    if(!tree->noeud[k]->tax){
      fprintf(fout,">Node %d, son of Node %d by branch of length %f\n", tree->noeud[k]->num, tree->noeud[k]->b[0]->left->num, tree->noeud[k]->b[0]->l);
      for(a=0;a<tree->mod->ns;a++) {
	for(i=0;i<tree->data->crunch_len;i++) {
	  for (j=0;j<tree->data->wght[i];j++){
	    fprintf(fout, "%f ",probas[k][i][a]);
	  }
	}
	fprintf(fout,"\n");
      }
    }
  }

  //  fclose(fout);
  /* 
  if (Filexists("BaseContents.txt")) {
    fout2 = Openfile("BaseContents.txt",2);
    fprintf(fout, "************************New Tree********************");
  }
  else {
    fout2 = Openfile("BaseContents.txt",1);
  }
  */
  for (k=0; k<2*tree->n_otu-1;k++) {
    if(!tree->noeud[k]->tax){
      fprintf(fout2,">Node %d, son of Node %d by branch of length %f\n", tree->noeud[k]->num, tree->noeud[k]->b[0]->left->num, tree->noeud[k]->b[0]->l);
      for(a=0;a<tree->mod->ns;a++) {
	means[k][a]=0;
	for(i=0;i<tree->data->crunch_len;i++) {
	  for (j=0;j<tree->data->wght[i];j++){
	    means[k][a]+=probas[k][i][a];
	  }
	}
	means[k][a]/=tree->data->init_len;
	fprintf(fout2,"%f\n",means[k][a]);
      } 
    }
  }
  
  // fclose(fout2);
  
  
  
  
}
/*********************************************************/


